import maya.cmds as cmds

import rigging_framework.templates.common.utilities.transform as trans_mod
from rigging_framework.templates.common.utilities import utils as utils
from rigging_framework.templates.common.utilities import rig_controls as rig_ctrls
from rigging_framework.templates.common.utilities import constraints as cnst


def build(list_of_objects, name_or_none="FK_Chain", replace_in_name=(), add_prefix="", add_suffix="", scale=1,
          color="blue", point_constraint=True, orient_constraint=True, scale_constraint=False,
          virtual_constraints=False):
    """
    Builds an fk chain based on given parameters
    :param list[str] list_of_objects: list of objects to build chain grom
    :param str or None name_or_none: if name is None then chain elements will be named after objects
    :param tuple replace_in_name: tuple of pairs (tuples) - it can replace substrings in the name
    :param str add_prefix: will add prefix to FK controls
    :param str add_suffix: will add suffix to FK controls
    :param float scale: scale of controls
    :param str color: string name for color (blue, red, yellow, etc...)
    :param bool point_constraint: does this chain control translation
    :param bool orient_constraint: does this chaing control rotation
    :param bool scale_constraint: does this chain control scale
    :param bool virtual_constraints: whether constraints will be built during the build or saved for later
	
    :return dict:
	    result = {"Controls": list of controls
                  "Control_Groups": list of control groups
                  "Module_Group": main module group name}
    """
    
    if not virtual_constraints:
        cnst.initialize_constraints()

    controls = []

    last_parent = None
    for i, obj in enumerate(list_of_objects):
        if name_or_none:
            control_name = add_prefix + name_or_none + "_" + str(i).zfill(2) + add_suffix
        else:
            control_name = obj
            if replace_in_name:
                for original, replacement in replace_in_name:
                    control_name = control_name.replace(original, replacement)
            control_name = add_prefix + control_name + add_suffix

        control = rig_ctrls.make_control("circle_X", size=scale, color=color, name=control_name)
        control_grp = trans_mod.make_space_for(control)
        utils.match_sec_to_first(obj, control_grp)
        if last_parent:
            cmds.parent(control_grp, last_parent)
        last_parent = control
        controls.append(control)
        if point_constraint and orient_constraint:
            cnst.parent_constraint(control, obj, maintain_offset=True)
        elif point_constraint:
            cnst.point_constraint(control, obj, maintain_offset=True)
        elif orient_constraint:
            cnst.orient_constraint(control, obj, maintain_offset=True)
        if scale_constraint:
            cnst.scale_constraint(control, obj)

    main_group_name = name_or_none
    if not main_group_name:
        main_group_name = list_of_objects[0]
    else:
        main_group_name = add_prefix + main_group_name
    main_group_name += "_Module_Group"

    # create main group for the module at first object:
    module_group = cmds.createNode("transform", name=main_group_name)

    utils.match_sec_to_first(trans_mod.get_parent(controls[0]),
                             module_group, rotation=False)  # snap group to first obj

    cmds.parent(trans_mod.get_parent(controls[0]), main_group_name)

    result = {"Controls": controls,
              "Control_Groups": [trans_mod.get_parent(x) for x in controls],
              "Module_Group": module_group}

    # build constraints now unless they are virtual
    if not virtual_constraints:
        cnst.build_constraints()

    return result
