import maya.cmds as cmds
from rigging_framework.templates.common.utilities import utils as utils
from rigging_framework.templates.common.utilities import rigging_math as rmath
from rigging_framework.templates.common.utilities import follicles
from rigging_framework.templates.common.utilities import constraints as cnst
from rigging_framework.templates.common.utilities import rig_controls
import rigging_framework.templates.common.utilities.transform as trans_mod
import maya.api.OpenMaya as om


def find_closest_uv_on_surface(surface, transform):
    """
    finds closest u and v values on given surface
    :param surface:
    :param transform:
    :return: u and v
    """
    surface_shape = cmds.listRelatives(surface, shapes=True, noIntermediate=True)[0]
    transform_position = cmds.xform(transform, query=True, translation=True, worldSpace=True)
    cps = cmds.createNode("closestPointOnSurface")
    cmds.setAttr(cps + ".inPosition", *transform_position)
    cmds.connectAttr(surface_shape + ".worldSpace[0]", cps + ".inputSurface")
    u = cmds.getAttr(cps + ".result.parameterU")
    v = cmds.getAttr(cps + ".result.parameterV")
    cmds.delete(cps)
    return u, v


def build(list_of_objects, name="Ribbon", attach_objects=True, sticky=False, make_controls=False, scale=1,
          surface_up=(0, 1, 0), virtual_constraints=False):
    """
    Builds a ribbon based on selected objects positions, sticky or non sticky
	
    :param list[str] list_of_objects: list of objects or joints to build ribbon along
    :param str name: name of the ribbon module
    :param bool attach_objects: whether selected objects will be attached to the ribbon - (overriden by make controls)
    :param bool sticky: whether or not we want this ribbon to be sticky or sliding on the surface
    :param bool make_controls: whether or not we want to make offset controls on this ribbon (overrides attach objects)
    :param float scale: scale of the ribbon
    :param float[3] surface_up: we need this to determine up from the surface - perpendicular vector to it
    :param bool virtual_constraints: whether or not constraints are virtual (will be build after the tool executes)
	
    :return dict:     
	result = {"Controls": list of controls
              "Module_Group": main ribbon group name
              "Control_Groups": list of control offset groups
              "Ribbon_Surface": ribbon surface object}
    """
    
    if not virtual_constraints:
        cnst.initialize_constraints()

    main_grp = cmds.createNode("transform", name="{0}_Module_Group".format(name))

    ribbon_crv = utils.build_linear_curve_with_cvs(name="{0}_Ribbon_Crv".format(name), num_cvs=len(list_of_objects))

    ribbon_crv_cvs = cmds.ls("{0}.cv[*]".format(ribbon_crv), flatten=True)

    for cv, obj in zip(ribbon_crv_cvs, list_of_objects):
        pos = cmds.xform(obj, query=True, translation=True, worldSpace=True)
        cmds.xform(cv, translation=pos, worldSpace=True)

    # if there are only three objects ribbon does not make much sense and rebuild will fail:
    if len(list_of_objects) > 3:
        cmds.rebuildCurve(ribbon_crv, ch=1, rpo=1, rt=0, end=1, kr=1, kcp=1, kep=1, kt=0, s=4, d=3,
                          tol=0)

    start_pos = cmds.xform(list_of_objects[0], query=True, worldSpace=True, translation=True)
    end_pos = cmds.xform(list_of_objects[-1], query=True, worldSpace=True, translation=True)
    curve_vector = om.MVector(om.MVector(end_pos) - om.MVector(start_pos))
    up_vector = om.MVector(surface_up)
    curve_vector.normalize()
    cross_vector = curve_vector ^ up_vector * scale * 0.2

    outer_curve = cmds.duplicate(ribbon_crv, name="{0}_Outer_Ribbon_Crv".format(name))
    inner_curve = cmds.duplicate(ribbon_crv, name="{0}_Inner_Ribbon_Crv".format(name))

    cmds.move(cross_vector.x, cross_vector.y, cross_vector.z, outer_curve, relative=True, worldSpace=True)
    cmds.move(-cross_vector.x, -cross_vector.y, -cross_vector.z, inner_curve, relative=True, worldSpace=True)

    # now loft a surface between two curves:
    ribbon_sfc = cmds.loft(inner_curve, outer_curve, constructionHistory=1, uniform=1, close=0, autoReverse=1, degree=3,
                           sectionSpans=1, range=1, reverseSurfaceNormals=True, name="{0}_Ribbon_Sfc".format(name))[0]

    cmds.delete(inner_curve, outer_curve, ribbon_crv)
    cmds.rebuildSurface(ribbon_sfc, constructionHistory=1, rpo=1, rt=0, end=1, keepRange=0, keepControlPoints=0, kc=0,
                        spansU=3, degreeU=3, spansV=1, degreeV=1, tol=0, fitRebuild=0, direction=2)

    controls = []
    control_groups = []

    if not sticky:
        # now we are going to convert isoparm into a live curve on the ribbon
        live_ribbon_crv = cmds.duplicateCurve("{0}.v[0.5]".format(ribbon_sfc), ch=True, rn=0, local=0,
                                              name="{0}_Live_Ribbon_Crv".format(name))

        ribbon_sfc_shape = cmds.listRelatives(ribbon_sfc, shapes=True, noIntermediate=True)[0]

        cmds.parent(ribbon_sfc, main_grp)
        cmds.parent(live_ribbon_crv, main_grp)
        cmds.setAttr(live_ribbon_crv[0] + ".inheritsTransform", False)

        # now we are going to create points on isoparm curve matching the positions of our objects
        for obj in list_of_objects:
            # find closest points on curves for every object and make a transform on that curve
            mpath = cmds.createNode("motionPath", name="{0}_MPath".format(obj))
            cmds.setAttr(mpath + ".fractionMode", True)  # this determines that joints will be sliding on surface evenly
            live_ribbon_crv_shape = cmds.listRelatives(live_ribbon_crv, shapes=True, noIntermediate=True)[0]
            cmds.connectAttr(live_ribbon_crv_shape + ".worldSpace[0]", mpath + ".geometryPath")

            driver_grp = cmds.createNode("transform", name="{0}_Ribbon_Crv_Driver_Grp".format(obj))
            cmds.connectAttr(mpath + ".allCoordinates", driver_grp + ".translate")

            # using this as parametric length does not work with closest u using nearest point on curve
            u = rmath.get_best_value_for_nearest_distance(mpath + ".uValue", driver_grp, obj)
            # cmds.warning("U={0}".format(u))
            cmds.setAttr(mpath + ".uValue", u)

            # finding closest points on surface to the points on curves
            closest_on_surface = cmds.createNode("closestPointOnSurface", name="{0}_CPS".format(obj))
            cmds.connectAttr(ribbon_sfc_shape + ".worldSpace[0]", closest_on_surface + ".inputSurface")
            cmds.connectAttr(driver_grp + ".translate", closest_on_surface + ".inPosition")

            # create a follicle on surface and set its u and v to to closestPointOnSurface which makes it a projection
            # of a transform on a curve
            follicle_transform = follicles.make_follicle(ribbon_sfc, [0, 0], name="{0}_Ribbon_Driver_Fol".format(obj))
            follicle_shape = cmds.listRelatives(follicle_transform, shapes=True)[0]
            cmds.connectAttr(closest_on_surface + ".result.parameterU", follicle_shape + ".parameterU")
            cmds.connectAttr(closest_on_surface + ".result.parameterV", follicle_shape + ".parameterV")

            cmds.setAttr(follicle_transform + ".inheritsTransform", False)  # making sure the follicle doesnt double

            # those follicles will be driving our objects
            if attach_objects and not make_controls:
                cnst.parent_constraint(follicle_transform, obj, maintain_offset=True)
            elif make_controls:
                # here we call function to make offset controls for each object
                tweak_control = rig_controls.make_control(control_type="double_pointer", size=scale * 0.1,
                                                          name=follicle_transform.replace("Driver_Fol", "Tweak_Ctrl"),
                                                          color="teal")
                tweak_control_grp = trans_mod.make_space_for(tweak_control)
                cmds.delete(cmds.parentConstraint(follicle_transform, tweak_control_grp))
                cmds.parent(tweak_control_grp, follicle_transform)
                if attach_objects:
                    cnst.point_constraint(tweak_control, obj, maintain_offset=True)
                    cnst.orient_constraint(tweak_control, obj, maintain_offset=True)

                controls.append(tweak_control)
                control_groups.append(tweak_control_grp)

            # cleanup
            cmds.parent(driver_grp, main_grp)
            cmds.parent(follicle_transform, main_grp)
            cmds.scaleConstraint(main_grp, follicle_transform)
    else:
        cmds.parent(ribbon_sfc, main_grp)

        for obj in list_of_objects:
            # find closest points on surface for every object and make a follicle on the surface
            u, v = find_closest_uv_on_surface(ribbon_sfc, obj)
            follicle_transform = follicles.make_follicle(ribbon_sfc, [0, 0], name="{0}_Ribbon_Driver_Fol".format(obj))
            follicle_shape = cmds.listRelatives(follicle_transform, shapes=True)[0]
            cmds.setAttr(follicle_shape + ".parameterU", u)
            cmds.setAttr(follicle_shape + ".parameterV", v)

            cmds.setAttr(follicle_transform + ".inheritsTransform", False)  # making sure the follicle doesnt double

            # those follicles will be driving our objects
            if attach_objects and not make_controls:
                cnst.parent_constraint(follicle_transform, obj, maintain_offset=True)
            elif make_controls:
                # here we call function to make offset controls for each object
                tweak_control = rig_controls.make_control(control_type="double_pointer", size=scale * 0.1,
                                                          name=follicle_transform.replace("Driver_Fol", "Tweak_Ctrl"),
                                                          color="teal")
                tweak_control_grp = trans_mod.make_space_for(tweak_control)
                cmds.delete(cmds.parentConstraint(follicle_transform, tweak_control_grp))
                cmds.parent(tweak_control_grp, follicle_transform)
                if attach_objects:
                    cnst.point_constraint(tweak_control, obj, maintain_offset=True)
                    cnst.orient_constraint(tweak_control, obj, maintain_offset=True)

                controls.append(tweak_control)
                control_groups.append(tweak_control_grp)

            # cleanup
            cmds.parent(follicle_transform, main_grp)
            cmds.scaleConstraint(main_grp, follicle_transform)

    if not virtual_constraints:
        cnst.build_constraints()

    result = {"Controls": controls,
              "Module_Group": main_grp,
              "Control_Groups": control_groups,
              "Ribbon_Surface": ribbon_sfc}

    return result


