"""
Use this to import model to the rig scene
"""

import json
import os
import sys
from functools import partial
import maya.cmds as cmds

from rigging_framework import global_variables as gvar
from rigging_framework import log
from rigging_framework.templates.common.utilities import sorting

cwd = os.path.dirname(os.path.realpath(__file__))


def build_angle_between(name="Sample"):
    """
    :param name:
    :return: returns a list: main group, result, joint, direction
    """

    main_grp = cmds.createNode("transform", name="{0}_Angle_Between_Grp".format(name))

    result = cmds.spaceLocator(name="{0}_Angle_Between_Result".format(name))[0]
    direction = cmds.spaceLocator(name="{0}_Angle_Between_Direction".format(name))[0]
    root = cmds.spaceLocator(name="{0}_Angle_Between_Root".format(name))[0]
    joint = cmds.spaceLocator(name="{0}_Angle_Between_Joint".format(name))[0]

    cmds.parent(result, direction, root, joint, main_grp)

    angle = cmds.createNode("angleBetween", name="{0}_Angle_Between_Angle")
    direction_dm = cmds.createNode("decomposeMatrix", name="{0}_Direction_Decomp".format(name))
    joint_dm = cmds.createNode("decomposeMatrix", name="{0}_Joint_Decomp".format(name))
    root_dm = cmds.createNode("decomposeMatrix", name="{0}_Root_Decomp".format(name))

    direction_vec_pma = cmds.createNode("plusMinusAverage", name="{0}_Direction_Vec_Pma".format(name))
    joint_vec_pma = cmds.createNode("plusMinusAverage", name="{0}_Joint_Vec_Pma".format(name))

    cmds.connectAttr(direction + ".xformMatrix", direction_dm + ".inputMatrix")
    cmds.connectAttr(root + ".xformMatrix", root_dm + ".inputMatrix")
    cmds.connectAttr(joint + ".xformMatrix", joint_dm + ".inputMatrix")

    cmds.connectAttr(direction_dm + ".outputTranslate", direction_vec_pma + ".input3D[0]")
    cmds.connectAttr(root_dm + ".outputTranslate", direction_vec_pma + ".input3D[1]")
    cmds.setAttr(direction_vec_pma + ".operation", 2)

    cmds.connectAttr(joint_dm + ".outputTranslate", joint_vec_pma + ".input3D[0]")
    cmds.connectAttr(root_dm + ".outputTranslate", joint_vec_pma + ".input3D[1]")
    cmds.setAttr(joint_vec_pma + ".operation", 2)

    cmds.connectAttr(direction_vec_pma + ".output3D", angle + ".vector1")
    cmds.connectAttr(joint_vec_pma + ".output3D", angle + ".vector2")

    cmds.connectAttr(angle + ".euler", result + ".rotate")

    cmds.select(main_grp, result, root, joint, direction, angle, direction_dm,
                joint_dm, root_dm, direction_vec_pma, joint_vec_pma)

    cmds.sets(name="{0}_Angle_Between_Set".format(name))

    return [main_grp, result, joint, direction]


def build_twist_system(twist_joints, build_countertwist=False):
    """
    Build twists and countertwists on selected joints
    :param twist_joints:
    :param build_countertwist:
    :return:
    """
    ab_grp, ab_result, ab_jnt, ab_dir = build_angle_between(name=twist_joints[0])

    if build_countertwist:
        first_joint = cmds.listRelatives(twist_joints[0], parent=True)[0]
        second_joint = [x for x in cmds.listRelatives(first_joint, children=True) if x not in twist_joints][0]

        space = cmds.listRelatives(first_joint, parent=True)[0]

        cmds.matchTransform(ab_grp, first_joint)
        cmds.parentConstraint(space, ab_grp, maintainOffset=True)
        cmds.pointConstraint(second_joint, ab_jnt)
        cmds.delete(cmds.pointConstraint(second_joint, ab_dir))

        cmds.orientConstraint(ab_result, twist_joints[0], maintainOffset=True)

        if len(twist_joints) == 2:
            o_cns = cmds.orientConstraint(first_joint, twist_joints[0], twist_joints[1], maintainOffset=True)[0]
            cmds.setAttr(o_cns + ".interpType", 2)

        if len(twist_joints) == 3:
            o_cns = cmds.orientConstraint(twist_joints[0], twist_joints[2], twist_joints[1], maintainOffset=True)[0]
            cmds.setAttr(o_cns + ".interpType", 2)

        sys.stdout.write("\nCounter twist system built successfully!\n")

        cmds.addAttr(ab_grp, longName="countertwist", attributeType="bool", defaultValue=True, keyable=True)
    else:
        first_joint = cmds.listRelatives(twist_joints[0], parent=True)[0]
        second_joint = [x for x in cmds.listRelatives(first_joint, children=True) if x not in twist_joints][0]

        cmds.matchTransform(ab_grp, second_joint)
        cmds.parentConstraint(second_joint, ab_grp, maintainOffset=True)

        cmds.pointConstraint(first_joint, ab_jnt)
        cmds.delete(cmds.pointConstraint(first_joint, ab_dir))

        cmds.orientConstraint(ab_result, twist_joints[0], maintainOffset=True)

        if len(twist_joints) == 2:
            o_cns = cmds.orientConstraint(first_joint, twist_joints[0], twist_joints[1], maintainOffset=True)[0]
            cmds.setAttr(o_cns + ".interpType", 2)

        if len(twist_joints) == 3:
            o_cns = cmds.orientConstraint(twist_joints[0], twist_joints[2], twist_joints[1], maintainOffset=True)[0]
            cmds.setAttr(o_cns + ".interpType", 2)

        sys.stdout.write("\nTwist system built successfully!\n")

    for twist in twist_joints:
        cmds.setAttr(twist + ".displayLocalAxis", True)


def build_twisters(cls, countertwist=False):
    """
    Builds twisters under selected joint
    :return:
    """
    jnt = cmds.ls(sl=True, type="joint")
    if not jnt:
        raise RuntimeError("No joint selected")

    twisters = [x for x in cmds.listRelatives(jnt, children=True) if cmds.listRelatives(x, children=True) is None]
    twisters = sorting.sort_by_closest_to(twisters, jnt)

    build_twist_system(twisters, build_countertwist=countertwist)


twister_config = [
    {"joint": "upperarm_twist_01_l", "countertwist": True},
    {"joint": "upperarm_twist_01_r", "countertwist": True},
    {"joint": "lowerarm_twist_01_l", "countertwist": False},
    {"joint": "lowerarm_twist_01_r", "countertwist": False},
    {"joint": "thigh_twist_01_l", "countertwist": True},
    {"joint": "thigh_twist_01_r", "countertwist": True},
    {"joint": "calf_twist_01_l", "countertwist": False},
    {"joint": "calf_twist_01_r", "countertwist": False}
]

for twist_sys in twister_config:
    cmds.select(cmds.listRelatives(twist_sys["joint"], parent=True))
    countertwist = twist_sys["countertwist"]
    build_twisters(countertwist)

sys.stdout.write("\nTwister config loaded!\n")