import maya.cmds as cmds

from rigging_framework.templates.common.utilities import driven_keys
from rigging_framework.templates.common.utilities import utils
from rigging_framework.templates.common.utilities import colors
from rigging_framework.templates.common.utilities import constraints as cnst
from rigging_framework.templates.common.modules import fk_chain


def build(finger_joints, name_or_none, add_prefix=None, add_suffix="_Fk_Ctrl",
          replace_in_name=(("FK_", ""), ("_Jnt", "")),
          stretch_amount=10.0,
          scale=1.0, virtual_constraints=False):
          
    """
    builds a simple fk finger based on fk chain, 
    
    Expects certain hard-coded conventions that will affect finger behaviour (will be updated)

    "thumb", "index", "middle", "ring", "pinky"
    "left", "right" or "_l", "_r"

    :param list[str] finger_joints: list of finger joints
    :param str or None name_or_none: name of the module or it will adopt name of the object for control names
    :param str add_prefix: adds prefix to control names
    :param str add_suffix: adds suffix to control names
    :param tuple replace_in_name: tuple of two strings - replace substring with another substring from object to control
    :param float scale: scale of controls
    :param float stretch_amount: allowed amount to stretch (10.0 - default)
    :param bool virtual_constraints: build constraints or leave them for later...

    result = {"Module_Group": Module group name
              "Main_Node": Main node - first finger control
              "Controls": List of finger controls
              "Control_Groups": List of finger control groups
              "Finger_Pmas": dictionary of finger plus minus nodes with keys: bend, sway, curl etc... }
    """

    if not virtual_constraints:
        cnst.initialize_constraints()

    fk_chain_data = fk_chain.build(finger_joints, name_or_none=name_or_none,
                                   add_prefix=add_prefix, add_suffix=add_suffix,
                                   replace_in_name=replace_in_name,
                                   scale=scale * 0.1, point_constraint=True, orient_constraint=True,
                                   scale_constraint=False, virtual_constraints=virtual_constraints)

    result = {"Module_Group": fk_chain_data["Module_Group"],
              "Main_Node": fk_chain_data["Controls"][0],
              "Controls": fk_chain_data["Controls"],
              "Control_Groups": fk_chain_data["Control_Groups"],
              "Finger_Pmas": dict()}

    # color first one teal and make it slightly bigger:
    colors.set_color_from_str(result["Controls"][0], "oceanBlue")
    utils.transform_shape(result["Controls"][0], [(0, 0, 0), (0, 0, 0), (1.2, 1.2, 1.2)])

    # add curl and sway:
    cmds.addAttr(result["Controls"][0], longName="curl", minValue=-10.0, maxValue=10.0, keyable=True)
    cmds.addAttr(result["Controls"][0], longName="sway", minValue=-10.0, maxValue=10.0, keyable=True)
    cmds.addAttr(result["Controls"][0], longName="bend", minValue=-10.0, maxValue=10.0, keyable=True)
    cmds.addAttr(result["Controls"][0], longName="wiggle", minValue=-10.0, maxValue=10.0, keyable=True)
    cmds.addAttr(result["Controls"][0], longName="stretch", minValue=-10.0, maxValue=10.0, keyable=True)
    cmds.addAttr(result["Controls"][0], longName="splay", minValue=-10.0, maxValue=10.0, keyable=True)

    if "thumb" not in finger_joints[0].lower():
        cmds.addAttr(result["Controls"][0], longName="contracture", minValue=-10.0, maxValue=10.0, keyable=True)

    curl_curve = driven_keys.build_quick_curve(name="{0}_curl_curve".format(result["Controls"][0]),
                                               points=[(-10, 90), (0, 0), (10, -90)])

    sway_curve = driven_keys.build_quick_curve(name="{0}_sway_curve".format(result["Controls"][0]),
                                               points=[(-10, -45), (0, 0), (10, 45)])

    bend_curve = driven_keys.build_quick_curve(name="{0}_bend_curve".format(result["Controls"][0]),
                                               points=[(-10, 90), (0, 0), (10, -90)])

    wiggle_curve = driven_keys.build_quick_curve(name="{0}_wiggle_curve".format(result["Controls"][0]),
                                                 points=[(-10, -45), (0, 0), (10, 45)])

    if "left" in finger_joints[0].lower() or finger_joints[0].lower().endswith("_l"):
        stretch_curve = driven_keys.build_quick_curve(name="{0}_stretch_curve".format(result["Controls"][0]),
                                                      points=[(-10, -0.25 * stretch_amount), (0, 0), (10, 0.5 * stretch_amount)])
    else:
        stretch_curve = driven_keys.build_quick_curve(name="{0}_stretch_curve".format(result["Controls"][0]),
                                                      points=[(-10, 0.25 / stretch_amount), (0, 0), (10, -0.5 * stretch_amount)])

    splay_curve = driven_keys.build_quick_curve(name="{0}_splay_curve".format(result["Controls"][0]),
                                                points=[(-10, -45), (0, 0), (10, 45)])

    if "thumb" not in finger_joints[0].lower():
        contracture_curve = driven_keys.build_quick_curve(name="{0}_contracture_curve".format(result["Controls"][0]),
                                                          points=[(-10, 90), (0, 0), (10, -90)])

    pma_curl = cmds.createNode("plusMinusAverage", name="{0}_curl_pma".format(result["Controls"][0]))
    clamp_curl = cmds.createNode("clamp", name="{0}_curl_clamp".format(result["Controls"][0]))
    cmds.connectAttr("{0}.curl".format(result["Controls"][0]), pma_curl + ".input1D[0]")
    cmds.connectAttr(pma_curl + ".output1D", clamp_curl + ".inputR")
    cmds.setAttr(clamp_curl + ".maxR", 10)
    cmds.setAttr(clamp_curl + ".minR", -10)
    cmds.connectAttr(clamp_curl + ".outputR", curl_curve + ".input")
    result["Finger_Pmas"]["curl"] = pma_curl

    pma_sway = cmds.createNode("plusMinusAverage", name="{0}_sway_pma".format(result["Controls"][0]))
    clamp_sway = cmds.createNode("clamp", name="{0}_sway_clamp".format(result["Controls"][0]))
    cmds.connectAttr("{0}.sway".format(result["Controls"][0]), pma_sway + ".input1D[0]")
    cmds.connectAttr(pma_sway + ".output1D", clamp_sway + ".inputR")
    cmds.setAttr(clamp_sway + ".maxR", 10)
    cmds.setAttr(clamp_sway + ".minR", -10)
    cmds.connectAttr(clamp_sway + ".outputR", sway_curve + ".input")
    result["Finger_Pmas"]["sway"] = pma_sway

    pma_bend = cmds.createNode("plusMinusAverage", name="{0}_bend_pma".format(result["Controls"][0]))
    clamp_bend = cmds.createNode("clamp", name="{0}_bend_clamp".format(result["Controls"][0]))
    cmds.connectAttr("{0}.bend".format(result["Controls"][0]), pma_bend + ".input1D[0]")
    cmds.connectAttr(pma_bend + ".output1D", clamp_bend + ".inputR")
    cmds.setAttr(clamp_bend + ".maxR", 10)
    cmds.setAttr(clamp_bend + ".minR", -10)
    cmds.connectAttr(clamp_bend + ".outputR", bend_curve + ".input")
    result["Finger_Pmas"]["bend"] = pma_bend

    pma_wiggle = cmds.createNode("plusMinusAverage", name="{0}_wiggle_pma".format(result["Controls"][0]))
    clamp_wiggle = cmds.createNode("clamp", name="{0}_wiggle_clamp".format(result["Controls"][0]))
    cmds.connectAttr("{0}.wiggle".format(result["Controls"][0]), pma_wiggle + ".input1D[0]")
    cmds.connectAttr(pma_wiggle + ".output1D", clamp_wiggle + ".inputR")
    cmds.setAttr(clamp_wiggle + ".maxR", 10)
    cmds.setAttr(clamp_wiggle + ".minR", -10)
    cmds.connectAttr(clamp_wiggle + ".outputR", wiggle_curve + ".input")
    result["Finger_Pmas"]["wiggle"] = pma_wiggle

    pma_stretch = cmds.createNode("plusMinusAverage", name="{0}_stretch_pma".format(result["Controls"][0]))
    clamp_stretch = cmds.createNode("clamp", name="{0}_stretch_clamp".format(result["Controls"][0]))
    cmds.connectAttr("{0}.stretch".format(result["Controls"][0]), pma_stretch + ".input1D[0]")
    cmds.connectAttr(pma_stretch + ".output1D", clamp_stretch + ".inputR")
    cmds.setAttr(clamp_stretch + ".maxR", 10)
    cmds.setAttr(clamp_stretch + ".minR", -10)
    cmds.connectAttr(clamp_stretch + ".outputR", stretch_curve + ".input")
    result["Finger_Pmas"]["stretch"] = pma_stretch

    pma_splay = cmds.createNode("plusMinusAverage", name="{0}_splay_pma".format(result["Controls"][0]))
    clamp_splay = cmds.createNode("clamp", name="{0}_splay_clamp".format(result["Controls"][0]))
    cmds.connectAttr("{0}.splay".format(result["Controls"][0]), pma_splay + ".input1D[0]")
    cmds.connectAttr(pma_splay + ".output1D", clamp_splay + ".inputR")
    cmds.setAttr(clamp_splay + ".maxR", 10)
    cmds.setAttr(clamp_splay + ".minR", -10)
    cmds.connectAttr(clamp_splay + ".outputR", splay_curve + ".input")
    result["Finger_Pmas"]["splay"] = pma_splay

    if "thumb" not in finger_joints[0].lower():
        pma_contracture = cmds.createNode("plusMinusAverage", name="{0}_contracture_pma".format(result["Controls"][0]))
        clamp_contracture = cmds.createNode("clamp", name="{0}_contracture_clamp".format(result["Controls"][0]))
        cmds.connectAttr("{0}.contracture".format(result["Controls"][0]), pma_contracture + ".input1D[0]")
        cmds.connectAttr(pma_contracture + ".output1D", clamp_contracture + ".inputR")
        cmds.setAttr(clamp_contracture + ".maxR", 10)
        cmds.setAttr(clamp_contracture + ".minR", -10)
        cmds.connectAttr(clamp_contracture + ".outputR", contracture_curve + ".input")
        result["Finger_Pmas"]["contracture"] = pma_contracture

    # add sdk subgroups and drive them:
    i = 1
    for control, control_grp in zip(result["Controls"], result["Control_Groups"]):
        # (1) curl sway:
        curl_sway_sdk_grp = cmds.createNode("transform", name=control + "_Curl_Sway_SDK_Grp")
        utils.match_sec_to_first(control, curl_sway_sdk_grp)
        cmds.parent(control, curl_sway_sdk_grp)
        cmds.parent(curl_sway_sdk_grp, control_grp)

        cmds.connectAttr(curl_curve + ".output", curl_sway_sdk_grp + ".rz")

        if "thumb" in control.lower():
            multiplier = cmds.createNode("unitConversion")
            cmds.connectAttr(sway_curve + ".output", multiplier + ".input")
            cmds.setAttr(multiplier + ".conversionFactor", -1 * 0.017)
            cmds.connectAttr(multiplier + ".output", curl_sway_sdk_grp + ".ry")
        else:
            cmds.connectAttr(sway_curve + ".output", curl_sway_sdk_grp + ".ry")

        # (2) bend wiggle:
        if i == 1:
            bend_wiggle_sdk_grp = cmds.createNode("transform", name=control + "_Bend_Wiggle_SDK_Grp")
            utils.match_sec_to_first(control, bend_wiggle_sdk_grp)
            cmds.parent(curl_sway_sdk_grp, bend_wiggle_sdk_grp)
            cmds.parent(bend_wiggle_sdk_grp, control_grp)

            cmds.connectAttr(bend_curve + ".output", bend_wiggle_sdk_grp + ".rz")

            if "thumb" in control.lower():
                multiplier = cmds.createNode("unitConversion")
                cmds.connectAttr(wiggle_curve + ".output", multiplier + ".input")
                cmds.setAttr(multiplier + ".conversionFactor", -1 * 0.017)
                cmds.connectAttr(multiplier + ".output", bend_wiggle_sdk_grp + ".ry")
            else:
                cmds.connectAttr(wiggle_curve + ".output", bend_wiggle_sdk_grp + ".ry")

        # (3) stretch:
        cmds.connectAttr(stretch_curve + ".output", curl_sway_sdk_grp + ".tx")

        # (4) splay:
        if i == 1:
            splay_sdk_grp = cmds.createNode("transform", name=control + "_Splay_SDK_Grp")
            utils.match_sec_to_first(control, splay_sdk_grp)
            cmds.parent(bend_wiggle_sdk_grp, splay_sdk_grp)
            cmds.parent(splay_sdk_grp, control_grp)

            if "middle" not in control.lower():
                multiplier = cmds.createNode("unitConversion")
                cmds.connectAttr(splay_curve + ".output", multiplier + ".input")
                cmds.connectAttr(multiplier + ".output", splay_sdk_grp + ".ry")

                if "ring" in control.lower():
                    cmds.setAttr(multiplier + ".conversionFactor", 0.5 * 0.017)
                elif "pinky" in control.lower():
                    cmds.setAttr(multiplier + ".conversionFactor", 1 * 0.017)
                elif "index" in control.lower():
                    cmds.setAttr(multiplier + ".conversionFactor", -0.5 * 0.017)
                elif "thumb" in control.lower():
                    cmds.setAttr(multiplier + ".conversionFactor", 1.5 * 0.017)

        # (5) contracture:
        if "thumb" not in finger_joints[0].lower():
            contracture_sdk_grp = cmds.createNode("transform", name=control + "_Contracture_SDK_Grp")
            utils.match_sec_to_first(control, contracture_sdk_grp)
            parent_sdk_grp = cmds.listRelatives(control, parent=True)[0]
            cmds.parent(contracture_sdk_grp, parent_sdk_grp)
            cmds.parent(control, contracture_sdk_grp)

            multiplier = cmds.createNode("unitConversion")
            cmds.connectAttr(contracture_curve + ".output", multiplier + ".input")
            cmds.connectAttr(multiplier + ".output", contracture_sdk_grp + ".rz")

            if "pinky" in control.lower():
                cmds.setAttr(multiplier + ".conversionFactor", 1.0 * 0.017)
            elif "ring" in control.lower():
                cmds.setAttr(multiplier + ".conversionFactor", 0.75 * 0.017)
            elif "middle" in control.lower():
                cmds.setAttr(multiplier + ".conversionFactor", 0.5 * 0.017)
            elif "index" in control.lower():
                cmds.setAttr(multiplier + ".conversionFactor", 0.25 * 0.017)

        i += 1

    return result
