import maya.cmds as cmds

from rigging_framework.templates.common.utilities import transform as trans_mod
from rigging_framework.templates.common.utilities import rig_controls as rig_ctrls


def build(name, root_skeletal_joint, scale=1.0, up_vector=(0, 1, 0)):
    """
    Builds a global control and hierarchy
    :param str name: Name of the global control 
    :param str root_skeletal_joint: Name of the root skeletal joint
    :param float scale: Scale of the control
    :param float[3] up_vector: Up vector of the control
	
    :return str: Global control name
    """
    if up_vector == (0, 1, 0):
        control_type = "circle_Y"
    else:
        control_type = "circle_Z"

    global_ctrl = rig_ctrls.make_control(control_type=control_type, size=scale/2.0,
                                         name=name, color="purple")

    rig_grp = cmds.createNode("transform", name="Rig_Grp")

    skeleton_grp = cmds.createNode("transform", name="Skeleton_Grp")
    modules_grp = cmds.createNode("transform", name="Modules_Grp")
    do_not_touch_grp = cmds.createNode("transform", name="Do_Not_Touch")

    cmds.parent(modules_grp, skeleton_grp, do_not_touch_grp, rig_grp)

    trans_mod.make_space_for(global_ctrl, "Grp", parent=modules_grp)

    cmds.addAttr(global_ctrl, longName="global_scale", defaultValue=1, minValue=0.01, maxValue=1000, keyable=True)
    cmds.connectAttr(global_ctrl+".global_scale", global_ctrl+".sx")
    cmds.connectAttr(global_ctrl+".global_scale", global_ctrl+".sy")
    cmds.connectAttr(global_ctrl+".global_scale", global_ctrl+".sz")

    cmds.addAttr(global_ctrl, longName="extra_controls", attributeType="enum", enumName="Off:On", keyable=True)

    # scale constrain rig joints to global control:
    cmds.parent(root_skeletal_joint, skeleton_grp)

    cmds.parentConstraint(global_ctrl, skeleton_grp)
    cmds.scaleConstraint(global_ctrl, skeleton_grp)
    
    return global_ctrl


def post_process(name):
    """
    Post process scene parenting all modules under global control
    :param name:
    :return:
    """
    modules = cmds.ls("*_Module_Group", assemblies=True) + cmds.ls("*_Foot_System_Grp")
    cmds.parent(modules, name)

    settings_grp = cmds.ls("*_Settings_Grp*") + cmds.ls("*_Angle_Between_Grp")
    cmds.parent(settings_grp, "Rig_Grp|Do_Not_Touch")

    cmds.setAttr("Rig_Grp|Do_Not_Touch.v", False)


def gather_data_from_placer(placer_transform):
    """
    Gathers data from placer top node
    :param placer_transform:
    :return:
    """

    # get name:
    node = [x for x in cmds.listRelatives(placer_transform, children=True, fullPath=True)
            if x.split("|")[-1].startswith("Name__")][0]
    module_name = node.split("__")[-1]

    placers = {}
    for placer_str in ["Skeleton_Root"]:
        node = [x for x in cmds.listRelatives(placer_transform, children=True, fullPath=True)
                if x.split("|")[-1].startswith(placer_str)][0]
        cns = node + "|" + cmds.listConnections(node + ".tx", type="pointConstraint")[0]
        placers[placer_str] = cmds.pointConstraint(cns, q=True, targetList=True)[0]

    # get scale:
    scale = 1.0
    for x in cmds.listRelatives(placer_transform, children=True, fullPath=True):
        if x.endswith("Adjust"):
            scale = cmds.getAttr(x + ".Scale")

    return dict(module_name=module_name,
                skeleton_root=placers["Skeleton_Root"],
                scale=scale)


def set_data_to_placer(placer_transform, placer_data):
    """
    Sets data to placer top node
    :param placer_transform:
    :param placer_data:
    :return:
    """
    node = [x for x in cmds.listRelatives(placer_transform, children=True, fullPath=True)
            if x.split("|")[-1].startswith("Name__")][0]
    cmds.rename(node, "Name__{}".format(placer_data["module_name"]))

    if "skeleton_root" in placer_data.keys():
        cmds.pointConstraint(placer_data["skeleton_root"], placer_transform + "|Skeleton_Root_Placer_Joint")

    scale = placer_data["scale"] / 10.0

    cmds.setAttr(placer_transform + "|Global_Control_Adjust.s", scale, scale, scale)
