import maya.cmds as cmds
from rigging_framework.templates.common.utilities import rig_controls
from rigging_framework.templates.common.utilities import connections as conns_mod
from rigging_framework.templates.humanoid_rig.modules import finger_rig
from rigging_framework.templates.common.utilities import transform as trans_mod


def build(name, cog_ctrl, ikfk_switch_ctrl, ik_main_node, hand_joint, thumb_joints, index_joints, middle_joints,
          ring_joints, pinky_joints, scale):
    """
    Builds an hand rig
    
    :param str name: Name of the module
    :param str cog_ctrl: Cog control name in the rig (needed to parent hand under) 
    :param str ikfk_switch_ctrl: Name of the ikfk switch control
    :param str ik_main_node: Name of the main ik node to attach the hand to
    :param str hand_joint: Name of the hand joint
    :param list[str] thumb_joints: List of thumb joints
    :param list[str] index_joints: List of index joints
    :param list[str] middle_joints: List of middle joints
    :param list[str] ring_joints: List of ring joints
    :param list[str] pinky_joints: List of pinky joint
    :param float scale: Scale of the controls
    
    :return str: Ik hand control name
    """

    if hand_joint.lower().startswith("left") or hand_joint.lower().endswith("_l"):
        side = "Left"
    else:
        side = "Right"

    # making sure arm controls follow shoulder joint:

    # create hand ik controls:
    hand_position = cmds.xform(hand_joint, query=True, translation=True, worldSpace=True)

    # only align the hand with the joint if worldspace parameter is not set to true:
    # hand_rotation = rig_orientation
    # if worldspace != "True":
    hand_rotation = cmds.xform(hand_joint, query=True, rotation=True, worldSpace=True)

    ik_hand_control = rig_controls.make_control("box", size=scale * 0.5,
                                                name="{0}_Ik_Ctrl".format(name),
                                                color="red", position=hand_position, rotation=hand_rotation)

    ik_hand_control_group = trans_mod.make_space_for(ik_hand_control)

    # add attribute for soft ik on ik hand control:
    cmds.addAttr(ik_hand_control, longName="OTHER", attributeType="enum",
                 enumName=" ", keyable=True)
    cmds.setAttr(ik_hand_control + ".OTHER", lock=True)
    cmds.addAttr(ik_hand_control, longName="soft_ik", attributeType="float",
                 minValue=0, maxValue=50, keyable=True)
    cmds.connectAttr("{0}.soft_ik".format(ik_hand_control), "{0}.soft_ik".format(ik_main_node[0]))

    # parent the hand control under the controls group:
    cmds.parent(ik_hand_control_group, cog_ctrl)

    # constraint the main node to the hand control and the ik joint as well:
    cmds.pointConstraint(ik_hand_control, ik_main_node)
    cmds.orientConstraint(ik_hand_control, "IK_{0}".format(hand_joint), maintainOffset=True)

    # add visibility switch for ik foot controls:
    vis_curves = [x for x in cmds.listConnections("{0}.ikfk_blend".format(ikfk_switch_ctrl),
                                                  source=False,
                                                  destination=True, plugs=False) if
                  cmds.nodeType(x) == "animCurveUL"]

    ik_vis_crv = [x for x in vis_curves if "IK" in x][0]
    cmds.connectAttr(ik_vis_crv + ".output", ik_hand_control_group + ".visibility")

    fingers = {"Thumb": thumb_joints, "Index": index_joints,
               "Middle": middle_joints, "Ring": ring_joints,
               "Pinky": pinky_joints}

    # build finger control:
    hand_joint_position = cmds.xform(hand_joint, query=True, translation=True, worldSpace=True)
    hand_joint_rotation = cmds.xform(hand_joint, query=True, rotation=True, worldSpace=True)

    finger_control = rig_controls.make_control("circle_X", name="{0}_Fingers_Ctrl".format(side),
                                               color="oceanBlue", position=hand_joint_position,
                                               rotation=hand_joint_rotation,
                                               scale_comps=(0.3 * scale, 0.3 * scale, 0.3 * scale),
                                               scale=(1, 1, 1))

    # !!! the fingers control should be controlling the hand joint, so we swap it all here:
    elbow_grp = cmds.createNode("transform", name=finger_control + "_Grp")
    elbow_joint = cmds.listRelatives(hand_joint, parent=True)[0]
    cmds.parentConstraint(elbow_joint, elbow_grp)

    offset_grp = cmds.createNode("transform", name=finger_control + "_Offset_Grp")
    cmds.parent(offset_grp, elbow_grp)
    cmds.matchTransform(offset_grp, finger_control)
    cmds.parent(finger_control, offset_grp)

    conns_mod.move_connection(hand_joint + ".r", offset_grp + ".r")
    cmds.connectAttr(hand_joint + ".t", offset_grp + ".t")  # this to move fingers

    cmds.orientConstraint(finger_control, hand_joint)

    cmds.parent(elbow_grp, cog_ctrl)

    # finger_control_group = utils.make_space_for(finger_control)
    # cmds.parent(finger_control_group, cog_ctrl)
    # cmds.parentConstraint(hand_joint, finger_control_group)

    cmds.addAttr(finger_control, longName="fingers_curl", minValue=-10, maxValue=10, keyable=True)
    cmds.addAttr(finger_control, longName="fingers_sway", minValue=-10, maxValue=10, keyable=True)
    cmds.addAttr(finger_control, longName="fingers_bend", minValue=-10, maxValue=10, keyable=True)
    cmds.addAttr(finger_control, longName="fingers_wiggle", minValue=-10, maxValue=10, keyable=True)
    cmds.addAttr(finger_control, longName="fingers_stretch", minValue=-10, maxValue=10, keyable=True)
    cmds.addAttr(finger_control, longName="fingers_splay", minValue=-10, maxValue=10, keyable=True)
    cmds.addAttr(finger_control, longName="fingers_contracture", minValue=-10, maxValue=10, keyable=True)

    cmds.addAttr(finger_control, longName="thumb_curl", minValue=-10, maxValue=10, keyable=True)
    cmds.addAttr(finger_control, longName="thumb_sway", minValue=-10, maxValue=10, keyable=True)
    cmds.addAttr(finger_control, longName="thumb_bend", minValue=-10, maxValue=10, keyable=True)
    cmds.addAttr(finger_control, longName="thumb_wiggle", minValue=-10, maxValue=10, keyable=True)
    cmds.addAttr(finger_control, longName="thumb_stretch", minValue=-10, maxValue=10, keyable=True)
    cmds.addAttr(finger_control, longName="thumb_splay", minValue=-10, maxValue=10, keyable=True)

    for finger, finger_joints in fingers.items():
        current_finger_data = finger_rig.build(finger_joints, name_or_none=finger, add_prefix=side + "_",
                                               scale=scale * 2.0)

        for attr in "curl", "sway", "bend", "wiggle", "stretch", "splay", "contracture":
            if "thumb" in finger.lower():
                if cmds.objExists("{0}.{1}".format(current_finger_data["Main_Node"], attr)):
                    cmds.connectAttr("{0}.thumb_{1}".format(finger_control, attr),
                                     current_finger_data["Finger_Pmas"][attr] + ".input1D[1]")
            else:
                if cmds.objExists("{0}.{1}".format(current_finger_data["Main_Node"], attr)):
                    cmds.connectAttr("{0}.fingers_{1}".format(finger_control, attr),
                                     current_finger_data["Finger_Pmas"][attr] + ".input1D[1]")

        current_finger_control_group = current_finger_data["Module_Group"]
        cmds.parent(current_finger_control_group, finger_control)
        
    return ik_hand_control


def gather_data_from_placer(placer_transform):
    """
    Gathers data from placer top node
    :param placer_transform:
    :return:
    """
    node = [x for x in cmds.listRelatives(placer_transform, children=True, fullPath=True)
            if x.split("|")[-1].startswith("Name__")][0]
    module_name = node.split("__")[-1]

    # getting the placers information
    placers = {}
    for placer_str in "Hand", "Thumb", "Index", "Middle", "Ring", "Pinky":
        node = [x for x in cmds.listRelatives(placer_transform, children=True, fullPath=True)
                if x.split("|")[-1].startswith(placer_str)][0]
        cns = node + "|" + cmds.listConnections(node + ".tx", type="pointConstraint")[0]
        placers[placer_str] = cmds.pointConstraint(cns, q=True, targetList=True)[0]

    # getting connection info:
    plugged_dict = {
        "IKFK_Spine": None,
        "IKFK_Arm": None,
    }

    for x in cmds.listRelatives(placer_transform, children=True, fullPath=True):
        if x.endswith("Plug") and "Input" in x:
            cns = cmds.listConnections(x + ".tx", type="pointConstraint")[0]
            output = cmds.pointConstraint(cns, q=True, targetList=True)[0]

            plugged_placer = cmds.listRelatives(output, parent=True)[0]
            if plugged_placer != placer_transform:
                node = [z for z in cmds.listRelatives(plugged_placer, children=True, fullPath=True)
                        if z.split("|")[-1].startswith("Name__")][0]
                plugged_name = node.split("__")[-1]
                for key in plugged_dict.keys():
                    if key in x:
                        plugged_dict[key] = [plugged_name, output.split("|")[-1]]

    return dict(module_name=module_name,
                hand_joint=placers["Hand"],
                thumb_joint=placers["Thumb"],
                index_joint=placers["Index"],
                middle_joint=placers["Middle"],
                ring_joint=placers["Ring"],
                pinky_joint=placers["Pinky"],
                input_info_spine=dict(module=plugged_dict["IKFK_Spine"][0], plug=plugged_dict["IKFK_Spine"][1]),
                input_info_arm=dict(module=plugged_dict["IKFK_Arm"][0], plug=plugged_dict["IKFK_Arm"][1])
                )


def set_data_to_placer(placer_transform, placer_data):
    """
    Sets data to placers
    :param placer_transform:
    :param placer_data:
    :return:
    """
    node = [x for x in cmds.listRelatives(placer_transform, children=True, fullPath=True)
            if x.split("|")[-1].startswith("Name__")][0]
    cmds.rename(node, "Name__{}".format(placer_data["module_name"]))

    cmds.pointConstraint(placer_data["hand_joint"], placer_transform + "|Hand_Placer_Joint")
    cmds.pointConstraint(placer_data["thumb_joint"], placer_transform + "|Thumb_Start_Placer_Joint")
    cmds.pointConstraint(placer_data["index_joint"], placer_transform + "|Index_Start_Placer_Joint")
    cmds.pointConstraint(placer_data["middle_joint"], placer_transform + "|Middle_Start_Placer_Joint")
    cmds.pointConstraint(placer_data["ring_joint"], placer_transform + "|Ring_Start_Placer_Joint")
    cmds.pointConstraint(placer_data["pinky_joint"], placer_transform + "|Pinky_Start_Placer_Joint")


def load_placer_connection(placer_transform, placer_data):
    """
    Load placer connections
    :param placer_transform:
    :param placer_data:
    :return:
    """
    # set connection info:
    for input_info in ["input_info_spine", "input_info_arm"]:
        if placer_data[input_info]:
            connected_module = [x for x in cmds.ls(assemblies=True) if "Name__{}".format(placer_data[input_info]["module"])
                                in cmds.listRelatives(x, children=True)][0]

            for x in cmds.listRelatives(placer_transform, children=True, fullPath=True):
                if x.endswith("Plug") and "Input" in x and input_info.replace("input_info_", "").replace("_","") in x.lower():

                    constraint = cmds.listConnections(x, type="constraint", source=True, destination=False)
                    if constraint:
                        cmds.delete(constraint)

                    cmds.pointConstraint(connected_module + "|" + placer_data[input_info]["plug"], x)