import maya.cmds as cmds
from rigging_framework.templates.common.utilities import transform as trans_mod
from rigging_framework.templates.common.utilities import rig_controls
from rigging_framework.templates.common.utilities import connections as conns_mod


def build(name, pelvis_jnt, cog_ctrl,
          left_hip_jnt, left_ik_hip_joint, left_hip_control_fk_grp, left_leg_ikfk_switch, left_ankle_jnt,
          right_hip_jnt, right_ik_hip_joint, right_hip_control_fk_grp, right_leg_ikfk_switch, right_ankle_jnt,
          scale):
    """
    :param str left_hip_jnt: left hip joint
    :param str left_ik_hip_joint: left ik hip joint
    :param str left_hip_control_fk_grp: left hip fk control offset group
    :param str left_leg_ikfk_switch: left leg ikfk switch
    :param str left_ankle_jnt: left ankle joint
    :param str right_hip_jnt: right hip joint
    :param str right_ik_hip_joint: right ik hip joint (assumes we have fk and ik joints)
    :param str right_hip_control_fk_grp: right hip fk control offset group
    :param str right_leg_ikfk_switch: right leg ikfk switch
    :param str right_ankle_jnt: right ankle joint
    :param str pelvis_jnt: pelvis joint
    :param str cog_ctrl: COG control
    :param str name: name of the module
    :param float scale: controls scale
    :return str: pelvis control name
    """

    pelvis_position = cmds.xform(pelvis_jnt, query=True, translation=True, worldSpace=True)
    pelvis_rotation = cmds.xform(pelvis_jnt, query=True, rotation=True, worldSpace=True)

    pelvis_control = rig_controls.make_control("circle_X", name=name + "_Pelvis_Ctrl",
                                               color="blue", position=pelvis_position, rotation=pelvis_rotation,
                                               scale_comps=(scale, scale, scale))

    pelvis_control_group = trans_mod.make_space_for(pelvis_control)
    cmds.parent(pelvis_control_group, cog_ctrl)

    for hip_joint, ik_hip_joint, hip_control_fk_grp in zip([left_hip_jnt, right_hip_jnt],
                                                           [left_ik_hip_joint, right_ik_hip_joint],
                                                           [left_hip_control_fk_grp, right_hip_control_fk_grp]):

        hip_position = cmds.xform(hip_joint, query=True, translation=True, worldSpace=True)

        side = "Right"
        control_rotation = (180, 0, 0)
        if hip_joint == left_hip_jnt:
            control_rotation = (0, 0, 0)
            side = "Left"

        hip_control = rig_controls.make_control("clavicle", name="{}_Hip_Ctrl".format(side),
                                                color="blue", position=hip_position, rotation=pelvis_rotation,
                                                rotate_comps=control_rotation,
                                                scale_comps=(scale, scale, scale))

        cmds.rotate(control_rotation[0], control_rotation[1], control_rotation[2], hip_control, relative=True)

        hip_control_group = trans_mod.make_space_for(hip_control)
        cmds.parent(hip_control_group, cog_ctrl)
        cmds.parentConstraint(pelvis_control, hip_control_group, maintainOffset=True, skipRotate=["x", "y", "z"])

        # now we need to hook up joints to the hip controls and IK joint too:
        cmds.pointConstraint(hip_control, hip_joint)
        cmds.pointConstraint(hip_control, ik_hip_joint)

        # now we parent hip control fk group to the hip control
        cmds.parent(hip_control_fk_grp, hip_control)

    # we hook up pelvis joint to the hip control
    cmds.parentConstraint(pelvis_control, pelvis_jnt)

    # we need to lock feet positions when IK is set to 1, otherwise there is some foot drift
    # for that we are going to use condition node which will switch between blend node and parentConstraint
    # which will only be triggered on IK=1

    for ankle_rig_joint, ikfk_switch_control in zip([left_ankle_jnt, right_ankle_jnt],
                                                    [left_leg_ikfk_switch, right_leg_ikfk_switch]):
        condition_t = cmds.createNode("condition")
        condition_r = cmds.createNode("condition")

        cmds.setAttr(condition_t + ".secondTerm", 1)
        cmds.setAttr(condition_r + ".secondTerm", 1)

        cmds.connectAttr(ikfk_switch_control + ".ikfk_blend", condition_t + ".firstTerm")
        cmds.connectAttr(ikfk_switch_control + ".ikfk_blend", condition_r + ".firstTerm")

        pair_blend = [x for x in cmds.listConnections(ankle_rig_joint, source=True, destination=False)
                      if cmds.nodeType(x) == "pairBlend"][0]

        cmds.connectAttr(pair_blend + ".outTranslate", condition_t + ".colorIfTrue")
        cmds.connectAttr(pair_blend + ".outRotate", condition_r + ".colorIfTrue")

        conns_mod.disconnect_attr(ankle_rig_joint + ".t")
        conns_mod.disconnect_attr(ankle_rig_joint + ".r")

        p_cns = cmds.parentConstraint("IK_{0}".format(ankle_rig_joint), "FK_{0}".format(ankle_rig_joint),
                                      ankle_rig_joint, maintainOffset=True)[0]

        cmds.connectAttr(ikfk_switch_control + ".ikfk_blend", p_cns + ".FK_{0}W1".format(ankle_rig_joint))
        rev_ik = cmds.createNode("reverse")
        cmds.connectAttr(ikfk_switch_control + ".ikfk_blend", "{0}.inputX".format(rev_ik))
        cmds.connectAttr("{0}.outputX".format(rev_ik), p_cns + ".IK_{0}W0".format(ankle_rig_joint))

        for axis, color in zip("XYZ", "RGB"):
            cmds.connectAttr(p_cns + ".constraintTranslate{0}".format(axis), condition_t + f".colorIfFalse{color}")
            cmds.connectAttr(p_cns + ".constraintRotate{0}".format(axis), condition_r + f".colorIfFalse{color}")
            conns_mod.disconnect_attr(ankle_rig_joint + ".translate{0}".format(axis))
            conns_mod.disconnect_attr(ankle_rig_joint + ".rotate{0}".format(axis))

        cmds.connectAttr("{0}.outColor".format(condition_t), ankle_rig_joint + ".t")
        cmds.connectAttr("{0}.outColor".format(condition_r), ankle_rig_joint + ".r")

        return pelvis_control


def gather_data_from_placer(placer_transform):
    """
    Gathers data from placers
    :param placer_transform:
    :return:
    """
    node = [x for x in cmds.listRelatives(placer_transform, children=True, fullPath=True)
            if x.split("|")[-1].startswith("Name__")][0]
    module_name = node.split("__")[-1]

    placers = {}
    for placer_str in ["Pelvis"]:
        node = [x for x in cmds.listRelatives(placer_transform, children=True, fullPath=True)
                if x.split("|")[-1].startswith(placer_str)][0]
        cns = node + "|" + cmds.listConnections(node + ".tx", type="pointConstraint")[0]
        placers[placer_str] = cmds.pointConstraint(cns, q=True, targetList=True)[0]

    # get connection info:
    plugged_dict = {
        "IKFK_LeftLeg": None,
        "IKFK_RightLeg": None,
        "IKFK_Spine": None,
    }

    for x in cmds.listRelatives(placer_transform, children=True, fullPath=True):
        if x.endswith("Plug") and "Input" in x:
            cns = cmds.listConnections(x + ".tx", type="pointConstraint")[0]
            output = cmds.pointConstraint(cns, q=True, targetList=True)[0]

            plugged_placer = cmds.listRelatives(output, parent=True)[0]
            if plugged_placer != placer_transform:
                node = [z for z in cmds.listRelatives(plugged_placer, children=True, fullPath=True)
                        if z.split("|")[-1].startswith("Name__")][0]
                plugged_name = node.split("__")[-1]
                for key in plugged_dict.keys():
                    if key in x:
                        plugged_dict[key] = [plugged_name, output.split("|")[-1]]

    return dict(module_name=module_name,
                pelvis_joint=placers["Pelvis"],
                input_info_left_leg=dict(module=plugged_dict["IKFK_LeftLeg"][0], plug=plugged_dict["IKFK_LeftLeg"][1]),
                input_info_right_leg=dict(module=plugged_dict["IKFK_RightLeg"][0], plug=plugged_dict["IKFK_RightLeg"][1]),
                input_info_spine=dict(module=plugged_dict["IKFK_Spine"][0], plug=plugged_dict["IKFK_Spine"][1]))


def set_data_to_placer(placer_transform, placer_data):
    """
    Sets data to placers
    :param placer_transform:
    :param placer_data:
    :return:
    """
    node = [x for x in cmds.listRelatives(placer_transform, children=True, fullPath=True)
            if x.split("|")[-1].startswith("Name__")][0]
    cmds.rename(node, "Name__{}".format(placer_data["module_name"]))

    cmds.pointConstraint(placer_data["pelvis_joint"], placer_transform + "|Pelvis_Placer_Joint")


def load_placer_connection(placer_transform, placer_data):
    """
    Load placer connections
    :param placer_transform:
    :param placer_data:
    :return:
    """
    # set connection info:
    for input_info in ["input_info_left_leg", "input_info_right_leg", "input_info_spine"]:
        if placer_data[input_info]:
            connected_module = [x for x in cmds.ls(assemblies=True) if "Name__{}".format(placer_data[input_info]["module"])
                                in cmds.listRelatives(x, children=True)][0]

            for x in cmds.listRelatives(placer_transform, children=True, fullPath=True):
                if x.endswith("Plug") and "Input" in x and input_info.replace("input_info_", "").replace("_","") in x.lower():

                    constraint = cmds.listConnections(x, type="constraint", source=True, destination=False)
                    if constraint:
                        cmds.delete(constraint)

                    cmds.pointConstraint(connected_module + "|" + placer_data[input_info]["plug"], x)
