import maya.cmds as cmds

import maya.api.OpenMaya as apiOM


from rigging_framework.templates.common.utilities import utils as utils
from rigging_framework.templates.common.utilities import rigging_math
from rigging_framework.templates.common.utilities import rig_controls
from rigging_framework.templates.common.utilities import driven_keys
from rigging_framework.templates.common.utilities import constraints as cnst
from rigging_framework.templates.common.utilities import transform as trans_mod


def get_furthest(axis, *objects):
    """
    gets furthest axis value for given objects (axis value can be +x, -x, +y, -y... etc...)
    :param axis:
    :param objects:
    :return:
    """
    direction = int(axis[0].replace("+", "") + "1")  # converting direction to +1 and -1
    axis = axis[1]

    result = -10000000 * direction
    axis_dict = {"x": 0, "y": 1, "z": 2}

    for obj in objects:
        position = cmds.xform(obj, query=True, translation=True, worldSpace=True)
        axis_position = position[axis_dict[axis]]

        if direction == 1:
            if axis_position > result:
                result = axis_position
        if direction == -1:
            if axis_position < result:
                result = axis_position

    return result


def build(name, foot_ik, ik_ankle_bone, ik_toe_base_bone, ik_toes_end_bone, toe_pivot, rear_pivot, out_pivot, in_pivot,
          ball_pivot, spin_pivot, mirror_foot=False, aim_vector=(1, 0, 0), up_vector=(0, 1, 0),
          virtual_constraints=False):
    """
    :param str foot_ik: this could be foot ik or locator driving the foot
    :param str ik_ankle_bone: IK ankle bone
    :param str ik_toe_base_bone: IK toe base bone 
    :param str ik_toes_end_bone: IK toes end bone
    :param str toe_pivot: Transform for toe pivot
    :param str rear_pivot: Transform for rear pivot
    :param str out_pivot: Transform for outside foot pivot
    :param str in_pivot: Transform for inner foot pivot
    :param str ball_pivot: Transform for ball pivot
    :param str spin_pivot: Transform for foot spinning pivot
    :param str name: Module name
    :param bool mirror_foot: Do we want to mirror/flip the foot behaviour
    :param float[3] aim_vector: Where the rig should be aiming (vector)
    :param float[3] up_vector: Where is the rig up vector (vector)
    :param bool virtual_constraints: Do we want to constraint bones?
    
    :return dict:     
        result = {"Module_Group": Name of the module group
                  "Control_Groups": Foot control groups
                  "Controls": Foot controls}
    """

    if not virtual_constraints:
        cnst.initialize_constraints()

    # calculate orientation based on up and aim:
    module_matrix = rigging_math.look_at(apiOM.MVector(aim_vector), apiOM.MVector(up_vector))
    module_transform = apiOM.MTransformationMatrix(module_matrix)

    transform_node = cmds.createNode("transform")
    cmds.xform(transform_node, matrix=module_transform.asMatrix(), worldSpace=True)
    orientation = cmds.getAttr(transform_node + ".rotate")[0]
    cmds.delete(transform_node)

    mirror = 1
    if mirror_foot:
        mirror = -1

    # create a foot control based on pivots:
    foot_control = rig_controls.make_control("box", name="{0}_IK_Foot_Ctrl".format(name), size=0.001, color="red",
                                             rotation=orientation)

    # adjust foot control shape so it fits around pivots:
    # get furthest z, get closest z , get furthest x, get clsoest x, get furthest y, get closest y
    bounds = dict()
    bounds["max_x"] = get_furthest("+x", toe_pivot, rear_pivot, out_pivot, in_pivot, ball_pivot, spin_pivot, foot_ik)
    bounds["min_x"] = get_furthest("-x", toe_pivot, rear_pivot, out_pivot, in_pivot, ball_pivot, spin_pivot, foot_ik)
    bounds["max_y"] = get_furthest("+y", toe_pivot, rear_pivot, out_pivot, in_pivot, ball_pivot, spin_pivot, foot_ik)
    bounds["min_y"] = get_furthest("-y", toe_pivot, rear_pivot, out_pivot, in_pivot, ball_pivot, spin_pivot, foot_ik)
    bounds["max_z"] = get_furthest("+z", toe_pivot, rear_pivot, out_pivot, in_pivot, ball_pivot, spin_pivot, foot_ik)
    bounds["min_z"] = get_furthest("-z", toe_pivot, rear_pivot, out_pivot, in_pivot, ball_pivot, spin_pivot, foot_ik)

    comps = dict()
    comps["max_z"] = ["{0}.cv[0]", "{0}.cv[3:7]", "{0}.cv[12:13]", "{0}.cv[16]"]
    comps["min_z"] = ["{0}.cv[1:2]", "{0}.cv[8:11]", "{0}.cv[14:15]"]
    comps["max_y"] = ["{0}.cv[0:4]", "{0}.cv[7:8]", "{0}.cv[11:12]"]
    comps["min_y"] = ["{0}.cv[5:6]", "{0}.cv[9:10]", "{0}.cv[13:16]"]
    comps["max_x"] = ["{0}.cv[0:1]", "{0}.cv[4:5]", "{0}.cv[10:14]"]
    comps["min_x"] = ["{0}.cv[2:3]", "{0}.cv[6:9]", "{0}.cv[15:16]"]

    clusters = []

    # get sides for each and create clusters, then move those clusters accordingly
    for key, components in comps.items():
        components = [c.format(foot_control) for c in components]
        clu_transform = cmds.cluster(components)[1]
        cmds.setAttr(clu_transform + ".t{0}".format(key[4]), bounds[key])
        clusters.append(clu_transform)

    rig_controls.store_shape(foot_control)
    utils.match_sec_to_first(foot_ik, foot_control, rotation=False)
    cmds.delete(clusters)
    foot_control_grp = trans_mod.make_space_for(foot_control)
    rig_controls.restore_shape()

    # build footrig system:

    # create SC handles:
    ankle_to_toebase_ik = cmds.ikHandle(startJoint=ik_ankle_bone, endEffector=ik_toe_base_bone, sol="ikSCsolver",
                                        name=name + "_Ankle_To_ToeBase_SC")[0]

    toebase_to_toesend_ik = cmds.ikHandle(startJoint=ik_toe_base_bone, endEffector=ik_toes_end_bone, sol="ikSCsolver",
                                          name=name + "_ToeBase_To_ToesEnd_SC")[0]

    # create groups for iks:
    ball_pivot_grp = cmds.createNode("transform", name="{0}_Ball_Pivot_Grp".format(name))
    utils.match_sec_to_first(ball_pivot, ball_pivot_grp, rotation=False)

    toes_pivot_grp = cmds.createNode("transform", name="{0}_Toes_Pivot_Grp".format(name))
    utils.match_sec_to_first(toe_pivot, toes_pivot_grp, rotation=False)

    rear_pivot_grp = cmds.createNode("transform", name="{0}_Rear_Pivot_Grp".format(name))
    utils.match_sec_to_first(rear_pivot, rear_pivot_grp, rotation=False)

    spin_pivot_grp = cmds.createNode("transform", name="{0}_Spin_Pivot_Grp".format(name))
    utils.match_sec_to_first(spin_pivot, spin_pivot_grp, rotation=False)

    in_pivot_grp = cmds.createNode("transform", name="{0}_In_Pivot_Grp".format(name))
    utils.match_sec_to_first(in_pivot, in_pivot_grp, rotation=False)

    out_pivot_grp = cmds.createNode("transform", name="{0}_Out_Pivot_Grp".format(name))
    utils.match_sec_to_first(out_pivot, out_pivot_grp, rotation=False)

    footrig_top_grp = cmds.createNode("transform", name="{0}_Footrig_Top_Grp".format(name))
    utils.match_sec_to_first(ik_ankle_bone, footrig_top_grp, rotation=False)

    # now orient all the groups like the rig orientation:
    cmds.xform(ball_pivot_grp, rotation=orientation, worldSpace=True)
    cmds.xform(toes_pivot_grp, rotation=orientation, worldSpace=True)
    cmds.xform(rear_pivot_grp, rotation=orientation, worldSpace=True)
    cmds.xform(spin_pivot_grp, rotation=orientation, worldSpace=True)
    cmds.xform(in_pivot_grp, rotation=orientation, worldSpace=True)
    cmds.xform(out_pivot_grp, rotation=orientation, worldSpace=True)
    cmds.xform(footrig_top_grp, rotation=orientation, worldSpace=True)

    # build group hierarchy
    cmds.parent(ball_pivot_grp, toes_pivot_grp)
    cmds.parent(toes_pivot_grp, rear_pivot_grp)
    cmds.parent(rear_pivot_grp, spin_pivot_grp)
    cmds.parent(spin_pivot_grp, in_pivot_grp)
    cmds.parent(in_pivot_grp, out_pivot_grp)
    cmds.parent(out_pivot_grp, footrig_top_grp)

    # parent iks to relevant groups
    cmds.parent(ankle_to_toebase_ik, ball_pivot_grp)
    cmds.parent(toebase_to_toesend_ik, toes_pivot_grp)
    cmds.parent(foot_ik, ball_pivot_grp)

    # add attributes as drivers for those groups
    cmds.addAttr(foot_control, longName="ROLLS", attributeType="enum", enumName=" ", keyable=True)
    cmds.setAttr(foot_control + ".ROLLS", lock=True)
    cmds.addAttr(foot_control, longName="ball_roll", minValue=0, maxValue=10, keyable=True)
    cmds.addAttr(foot_control, longName="roll_toes", minValue=-10, maxValue=10, keyable=True)
    cmds.addAttr(foot_control, longName="PIVOTS", attributeType="enum", enumName=" ", keyable=True)
    cmds.setAttr(foot_control + ".PIVOTS", lock=True)
    cmds.addAttr(foot_control, longName="toe_heel_pivot", minValue=-10, maxValue=10, keyable=True)
    cmds.addAttr(foot_control, longName="bank_in_out_pivot", minValue=-10, maxValue=10, keyable=True)
    cmds.addAttr(foot_control, longName="SPINS", attributeType="enum", enumName=" ", keyable=True)
    cmds.setAttr(foot_control + ".SPINS", lock=True)
    cmds.addAttr(foot_control, longName="front_spin", minValue=-10, maxValue=10, keyable=True)
    cmds.addAttr(foot_control, longName="heel_spin", minValue=-10, maxValue=10, keyable=True)
    cmds.addAttr(foot_control, longName="ball_spin", minValue=-10, maxValue=10, keyable=True)
    cmds.addAttr(foot_control, longName="OTHER", attributeType="enum", enumName=" ", keyable=True)

    # adding soft ik, only if soft ik attribute exists on the IK:
    if cmds.objExists(foot_ik[0] + ".soft_ik"):
        cmds.setAttr(foot_control + ".OTHER", lock=True)
        cmds.addAttr(foot_control, longName="soft_ik", minValue=0, maxValue=50, keyable=True, defaultValue=0)
        cmds.connectAttr(foot_control + ".soft_ik", foot_ik[0] + ".soft_ik")

    # determine which axis is up, which one roll, and which one aim
    # potentially determine direction and multiply driven keys by it...
    aim_axis = "X"
    up_axis = "Y"
    roll_axis = "Z"

    # create driven keys for the attributes:
    ball_roll_curve = driven_keys.build_quick_curve("{0}_Ball_Roll_Sdk".format(name), [(0, 0), (10, -100)])
    cmds.connectAttr(foot_control + ".ball_roll", ball_roll_curve + ".input")
    cmds.connectAttr(ball_roll_curve + ".output", ball_pivot_grp + ".rotate{}".format(roll_axis))

    toe_pivot_curve = driven_keys.build_quick_curve("{0}_Toe_Pivot_Sdk".format(name), [(0, 0), (10, -100)])
    cmds.connectAttr(foot_control + ".toe_heel_pivot", toe_pivot_curve + ".input")
    cmds.connectAttr(toe_pivot_curve + ".output", toes_pivot_grp + ".rotate{}".format(roll_axis))

    heel_pivot_curve = driven_keys.build_quick_curve("{0}_Heel_Pivot_Sdk".format(name), [(0, 0), (-10, 100)])
    cmds.connectAttr(foot_control + ".toe_heel_pivot", heel_pivot_curve + ".input")
    cmds.connectAttr(heel_pivot_curve + ".output", rear_pivot_grp + ".rotate{}".format(roll_axis))

    bank_in_pivot_curve = driven_keys.build_quick_curve("{0}_Bank_In_Pivot_Sdk".format(name),
                                                        [(0, 0), (10 * mirror, 100 * mirror)])
    cmds.connectAttr(foot_control + ".bank_in_out_pivot", bank_in_pivot_curve + ".input")
    cmds.connectAttr(bank_in_pivot_curve + ".output", in_pivot_grp + ".rotate{}".format(aim_axis))

    bank_out_pivot_curve = driven_keys.build_quick_curve("{0}_Bank_Out_Pivot_Sdk".format(name),
                                                         [(0, 0), (10 * -mirror, 100 * -mirror)])
    cmds.connectAttr(foot_control + ".bank_in_out_pivot", bank_out_pivot_curve + ".input")
    cmds.connectAttr(bank_out_pivot_curve + ".output", out_pivot_grp + ".rotate{}".format(aim_axis))

    front_spin_curve = driven_keys.build_quick_curve("{0}_Front_Spin_Sdk".format(name),
                                                     [(-10, -100), (0, 0), (10, 100)])
    cmds.connectAttr(foot_control + ".front_spin", front_spin_curve + ".input")
    cmds.connectAttr(front_spin_curve + ".output", toes_pivot_grp + ".rotate{}".format(up_axis))

    heel_spin_curve = driven_keys.build_quick_curve("{0}_Heel_Spin_Sdk".format(name),
                                                    [(-10, -100), (0, 0), (10, 100)])
    cmds.connectAttr(foot_control + ".heel_spin", heel_spin_curve + ".input")
    cmds.connectAttr(heel_spin_curve + ".output", rear_pivot_grp + ".rotate{}".format(up_axis))

    ball_spin_curve = driven_keys.build_quick_curve("{0}_Ball_Spin_Sdk".format(name),
                                                    [(-10, -100), (0, 0), (10, 100)])
    cmds.connectAttr(foot_control + ".ball_spin", ball_spin_curve + ".input")
    cmds.connectAttr(ball_spin_curve + ".output", spin_pivot_grp + ".rotate{}".format(up_axis))

    # now toe roll:
    # create empty group in the foot rig toes pivot group - at the position of toe_base_joint
    toes_roller_grp = cmds.createNode("transform", name="{0}_Toes_Roller".format(name))
    utils.match_sec_to_first(ik_toe_base_bone, toes_roller_grp, rotation=False)
    cmds.parent(toes_roller_grp, toes_pivot_grp)
    cmds.parent(toebase_to_toesend_ik, toes_roller_grp)

    roll_toes_curve = driven_keys.build_quick_curve("{0}_Roll_Toes_Sdk".format(name),
                                                    [(-10, 100), (0, 0), (10, -100)])
    cmds.connectAttr(foot_control + ".roll_toes", roll_toes_curve + ".input")
    cmds.connectAttr(roll_toes_curve + ".output", toes_roller_grp + ".rotate{0}".format(roll_axis))

    cmds.parent(footrig_top_grp, foot_control)

    # scale constraining the system:
    # cnst.scale_constraint(foot_control, ik_ankle_bone)
    # cnst.scale_constraint(foot_control, ik_toe_base_bone)
    # cnst.scale_constraint(foot_control, ik_toes_end_bone)

    if not virtual_constraints:
        cnst.build_constraints()

    result = {"Module_Group": foot_control_grp,
              "Control_Groups": [foot_control_grp],
              "Controls": [foot_control]}

    return result
