import maya.cmds as cmds

import maya.api.OpenMaya as apiOM

import rigging_framework.templates.common.utilities.transform as trans_mod
from rigging_framework.templates.humanoid_rig.modules import fk_foot_rig
from rigging_framework.templates.humanoid_rig.modules import ik_foot_rig
from rigging_framework.templates.common.utilities import utils as utils
from rigging_framework.templates.common.utilities import constraints as cnst
from rigging_framework.templates.common.utilities import driven_keys
from rigging_framework.templates.common.utilities import rigging_math
from rigging_framework.templates.common.utilities import minimization


def build(name, ikfk_switch_control, ikfk_switch_attr, leg_ik, ik_ankle_bone, fk_ankle_bone, toe_base_bone,
          toes_end_bone, toe_pivot, rear_pivot, out_pivot, in_pivot, ball_pivot, spin_pivot,
          scale=1.0, aim_vector=(1, 0, 0), mirror_foot=False, up_vector=(0, 1, 0), virtual_constraints=False):
    """
    Builds ikfk footrig system
    
    :param str name: Name of the module 
    :param str ikfk_switch_control: IKFK switch control
    :param str ikfk_switch_attr: IKFK attribute name
    :param str leg_ik: IK node for the leg (or locator driving IK)
    :param str ik_ankle_bone: Ik ankle joint
    :param str fk_ankle_bone: FK ankle joint
    :param str toe_base_bone: Toe base joint
    :param str toes_end_bone: Toes end joint
    :param str toe_pivot: Transform for toe pivot
    :param str rear_pivot: Transform for rear/heel pivot
    :param str out_pivot: Transform for outside pivot
    :param str in_pivot: Transform for inner/inside pivot
    :param str ball_pivot: Transform for ball/toebase pivot
    :param str spin_pivot: Transform for footspin pivot
    :param float scale: Scale for controls
    :param bool mirror_foot: Do we want to mirror the foot rig?
    :param float[3] aim_vector: Where there rig should be aiming? (Vector)
    :param float[3] up_vector: Where is the up direction for the footrig (Vector)
    :param bool virtual_constraints: Do we want to build constraints, or leave that for later.
    
    :return:
        result = {"Module_Group": main module group
                  "IK_Foot_Control": ik foot control,
                  "Controls": ik foot controls and fk foot controls,
                  "Control_Groups": if foot control groups and fk foot control groups,
                  "IK_Curve": visibility curve (animCurve) for IK,
                  "FK_Curve": visibility curve (animCurve) for FK}
    """
    
    if not virtual_constraints:
        cnst.initialize_constraints()

    # calculate orientation based on up and aim:
    module_matrix = rigging_math.look_at(apiOM.MVector(aim_vector), apiOM.MVector(up_vector))
    module_transform = apiOM.MTransformationMatrix(module_matrix)

    # create main groups:
    footrig_system_grp = cmds.createNode("transform", name="{0}_Module_Group".format(name))
    cmds.xform(footrig_system_grp, matrix=module_transform.asMatrix(), worldSpace=True)
    orientation = cmds.getAttr(footrig_system_grp + ".rotate")[0]

    # add attribute on the module group so is easy to find/identify:
    cmds.addAttr(footrig_system_grp, longName="moduleType", dataType="string")
    cmds.setAttr(footrig_system_grp + ".moduleType", "IKFK_Foot", type="string")

    # now we need to duplicate regular chain into IK chain for feet:
    utils.duplicate_unparent(toe_base_bone, 2, "IK")

    ik_toe_base_bone = "IK_{0}".format(toe_base_bone)
    ik_toes_end_bone = "IK_{0}".format(toes_end_bone)
    cmds.parent(ik_toe_base_bone, ik_ankle_bone)

    # build ik foot rig (Soft ik feature handled inside ik foot rig)
    ik_foot_rig_data = ik_foot_rig.build("{0}_Ik_FR".format(name), leg_ik, ik_ankle_bone, ik_toe_base_bone,
                                         ik_toes_end_bone, toe_pivot, rear_pivot, out_pivot, in_pivot, ball_pivot,
                                         spin_pivot, mirror_foot, aim_vector, up_vector, virtual_constraints)

    cmds.parent(ik_foot_rig_data["Module_Group"], footrig_system_grp)

    # now we need to duplicate regular chain into FK chain for feet:
    utils.duplicate_unparent(toe_base_bone, 2, "FK")
    fk_toe_base_bone = "FK_{0}".format(toe_base_bone)
    fk_toes_end_bone = "FK_{0}".format(toes_end_bone)
    cmds.parent(fk_toe_base_bone, fk_ankle_bone)

    # now build fk foot rig
    fk_foot_rig_data = fk_foot_rig.build(fk_toe_base_bone, fk_toes_end_bone, add_suffix="_FK_Ctrl",
                                         name_or_none=name + "_FK",
                                         scale=scale,
                                         virtual_constraints=virtual_constraints)

    cmds.parent(fk_foot_rig_data["Module_Group"], footrig_system_grp)

    # now parent and scale constraint fk stuff to fk ankle bone
    cnst.parent_constraint(fk_ankle_bone, fk_foot_rig_data["Module_Group"], maintain_offset=True)
    cnst.scale_constraint(fk_ankle_bone, fk_foot_rig_data["Module_Group"], maintain_offset=True)

    vis_curves = [x for x in cmds.listConnections("{0}.{1}".format(ikfk_switch_control, ikfk_switch_attr), source=False,
                                                  destination=True, plugs=False) if cmds.nodeType(x) == "animCurveUL"]

    # if script cannot find any visibility curves we gotta make them:
    if not vis_curves:
        fk_vis_crv = driven_keys.build_quick_curve("{0}_FK_Visibility_Curve".format(name), [(0, 0), (1, 1)])
        cmds.connectAttr("{0}.{1}".format(ikfk_switch_control, ikfk_switch_attr), fk_vis_crv + ".input")

        ik_vis_crv = driven_keys.build_quick_curve("{0}_IK_Visibility_Curve".format(name), [(0, 1), (1, 0)])
        cmds.connectAttr("{0}.{1}".format(ikfk_switch_control, ikfk_switch_attr), ik_vis_crv + ".input")
    else:
        ik_vis_crv = [x for x in vis_curves if "IK" in x][0]
        fk_vis_crv = [x for x in vis_curves if "FK" in x][0]

    # add visibility switch for ik foot controls:
    cmds.connectAttr("{0}.output".format(ik_vis_crv), ik_foot_rig_data["Module_Group"] + ".visibility")

    # add reverse visibility for FKs
    cmds.connectAttr("{0}.output".format(fk_vis_crv), fk_foot_rig_data["Module_Group"] + ".visibility")

    # build the blending:
    for i, joint in enumerate([toe_base_bone, toes_end_bone]):
        fk_joint = "FK_" + joint
        ik_joint = "IK_" + joint
        pair_blend = cmds.createNode("pairBlend")
        blend_colors = cmds.createNode("blendColors")
        cmds.connectAttr(ik_joint + ".translate", pair_blend + ".inTranslate1")
        cmds.connectAttr(fk_joint + ".translate", pair_blend + ".inTranslate2")
        cmds.connectAttr(ik_joint + ".rotate", pair_blend + ".inRotate1")
        cmds.connectAttr(fk_joint + ".rotate", pair_blend + ".inRotate2")
        cmds.connectAttr(ik_joint + ".scale", blend_colors + ".color2")
        cmds.connectAttr(fk_joint + ".scale", blend_colors + ".color1")
        cmds.connectAttr("{0}.{1}".format(ikfk_switch_control, ikfk_switch_attr), pair_blend + ".weight")
        cmds.connectAttr("{0}.{1}".format(ikfk_switch_control, ikfk_switch_attr), blend_colors + ".blender")

        # we don't want to plug anything to the first joint translation as its in a different space
        cmds.connectAttr(pair_blend + ".outTranslate", joint + ".translate")
        cmds.connectAttr(pair_blend + ".outRotate", joint + ".rotate")
        # cmds.connectAttr(blend_colors + ".output", joint + ".scale")

    # scale constraint IK chain and FK chains to the foot controls so it scales with the rig:
    cnst.scale_constraint(footrig_system_grp, ik_ankle_bone)
    cnst.scale_constraint(footrig_system_grp, fk_ankle_bone)

    # disconnect scale from ankle bone if there is any:
    try:
        utils.disconnect_attr(
            trans_mod.get_parent(toe_base_bone) + ".scale")
    except TypeError:
        cmds.warning("Cannot disconnect scaling from: {0}".format(
            trans_mod.get_parent(toe_base_bone)))

    if not virtual_constraints:
        cnst.build_constraints()

    result = {"Module_Group": footrig_system_grp,
              "IK_Foot_Control": ik_foot_rig_data["Controls"][0],
              "Controls": ik_foot_rig_data["Controls"] + fk_foot_rig_data["Controls"],
              "Control_Groups": ik_foot_rig_data["Control_Groups"] + fk_foot_rig_data["Control_Groups"],
              "IK_Curve": ik_vis_crv,
              "FK_Curve": fk_vis_crv}

    return result


def gather_data_from_placer(placer_transform):
    """
    Gathers data from placer top node
    :param placer_transform:
    :return:
    """
    node = [x for x in cmds.listRelatives(placer_transform, children=True, fullPath=True)
            if x.split("|")[-1].startswith("Name__")][0]
    module_name = node.split("__")[-1]

    placers = {}
    for placer_str in "Toe_Base", "Toes_End":
        node = [x for x in cmds.listRelatives(placer_transform, children=True, fullPath=True)
                if x.split("|")[-1].startswith(placer_str)][0]
        cns = node + "|" + cmds.listConnections(node + ".tx", type="pointConstraint")[0]
        placers[placer_str] = cmds.pointConstraint(cns, q=True, targetList=True)[0]

    # gather pivot positions:
    pivots = {}
    for x in cmds.listRelatives(placer_transform, children=True, fullPath=True):
        if x.endswith("Adjust"):
            position = cmds.xform(x, q=True, translation=True, worldSpace=True)
            pivots[x.split("|")[-1]] = position

    # get connection info:
    plugged_name = None
    output = ""
    for x in cmds.listRelatives(placer_transform, children=True, fullPath=True):
        if x.endswith("Plug") and "Input" in x:
            cns = cmds.listConnections(x + ".tx", type="pointConstraint")[0]
            output = cmds.pointConstraint(cns, q=True, targetList=True)[0]

            plugged_placer = cmds.listRelatives(output, parent=True)[0]
            if plugged_placer != placer_transform:
                node = [z for z in cmds.listRelatives(plugged_placer, children=True, fullPath=True)
                        if z.split("|")[-1].startswith("Name__")][0]
                plugged_name = node.split("__")[-1]

    return dict(module_name=module_name,
                toe_base_joint=placers["Toe_Base"],
                toes_end_joint=placers["Toes_End"],
                pivot_positions=pivots,
                input_module=plugged_name,
                input_plug=output.split("|")[-1] if plugged_name else None)


def set_data_to_placer(placer_transform, placer_data):
    """
    Sets data from placer top node
    :param placer_transform:
    :param placer_data:
    :return:
    """
    node = [x for x in cmds.listRelatives(placer_transform, children=True, fullPath=True)
            if x.split("|")[-1].startswith("Name__")][0]
    cmds.rename(node, "Name__{}".format(placer_data["module_name"]))

    cmds.pointConstraint(placer_data["toe_base_joint"], placer_transform + "|Toe_Base_Placer_Joint")
    cmds.pointConstraint(placer_data["toes_end_joint"], placer_transform + "|Toes_End_Placer_Joint")

    # set pivots positions
    for x in cmds.listRelatives(placer_transform, children=True, fullPath=True):
        if x.endswith("Adjust"):
            cmds.xform(x, translation=placer_data["pivot_positions"][x.split('|')[-1]], worldSpace=True)


def load_placer_connection(placer_transform, placer_data):
    """
    This loads connection based on the placer data
    :param placer_transform:
    :param placer_data:
    :return:
    """
    # set connection info:
    if placer_data["input_module"]:
        connected_module = [x for x in cmds.ls(assemblies=True) if "Name__{}".format(placer_data["input_module"])
                            in cmds.listRelatives(x, children=True)][0]

        for x in cmds.listRelatives(placer_transform, children=True, fullPath=True):
            if x.endswith("Plug") and "Input" in x:

                constraint = cmds.listConnections(x, type="constraint", source=True, destination=False)
                if constraint:
                    cmds.delete(constraint)

                cmds.pointConstraint(connected_module + "|" + placer_data["input_plug"], x)


def match_with_offset(obj1, obj2, offset_obj):
    """
    Match control with offset
    :param obj1:
    :param obj2:
    :param offset_obj:
    :return:
    """
    trans_mod.zero_out(obj1)

    obj_trs = cmds.createNode("transform")
    obj_trs_child = cmds.createNode("transform")
    cmds.parent(obj_trs_child, obj_trs)
    cmds.matchTransform(obj_trs, offset_obj)
    cmds.matchTransform(obj_trs_child, obj1)

    cmds.matchTransform(obj_trs, obj2)
    cmds.matchTransform(obj1, obj_trs_child)

    cmds.delete(obj_trs)


def ikfk_match(module_grp, ik_to_fk=True):
    """
    Match ik foot to fk foot and vice versa
    :param module_grp:
    :param ik_to_fk:
    :return:
    """
    module_name = module_grp.replace("_Module_Group", "")
    fk_controls = [x for x in cmds.listRelatives(module_grp, allDescendents=True) if x.endswith("FK_Ctrl")]

    fk_joints = []
    for fk_ctrl in fk_controls:
        cns = cmds.listConnections("{}.r".format(fk_ctrl), destination=True, source=False)[0]
        bone = cmds.listConnections("{}.constraintTranslateX".format(cns), destination=True, source=False)[0]
        fk_joints.append(bone)

    fk_joints += cmds.listRelatives(fk_joints[-1], parent=True)

    cns = cmds.listConnections("{}.tx".format(fk_joints[-1]), source=True, destination=False)[0]
    ctrl = cmds.parentConstraint(cns, q=True, targetList=True)[0]
    fk_controls += [ctrl]

    ik_joints = [x.replace("FK", "IK") for x in fk_joints]

    ik_foot_ctrl = [x for x in cmds.listRelatives(module_grp, allDescendents=True) if x.endswith("IK_Foot_Ctrl")][0]

    if ik_to_fk:
        cmds.setAttr(ik_foot_ctrl + ".ball_roll", 0)
        cmds.setAttr(ik_foot_ctrl + ".roll_toes", 0)
        cmds.setAttr(ik_foot_ctrl + ".toe_heel_pivot", 0)
        cmds.setAttr(ik_foot_ctrl + ".bank_in_out_pivot", 0)
        cmds.setAttr(ik_foot_ctrl + ".front_spin", 0)
        cmds.setAttr(ik_foot_ctrl + ".heel_spin", 0)
        cmds.setAttr(ik_foot_ctrl + ".ball_spin", 0)

        # this snaps foot ik control to fk joint but with offset - from ik_joint retained
        match_with_offset(ik_foot_ctrl, fk_joints[-1], ik_joints[-1])

        minimization.change_attribute_to_minimize_distance(ik_foot_ctrl, "roll_toes", ik_joints[0], fk_joints[0],
                                                           negative=True)
    else:  # fk to ik
        for fk_ctrl, ik_joint in zip(reversed(fk_controls), reversed(ik_joints)):
            cmds.matchTransform(fk_ctrl, ik_joint, position=False)
