import os

import maya.cmds as cmds

from rigging_framework.templates.common.utilities import utils
from rigging_framework.templates.common.utilities import colors
from rigging_framework.templates.common.utilities import joints
from rigging_framework.templates.common.utilities import rig_controls
from rigging_framework.templates.common.utilities import driven_keys
from rigging_framework.templates.common.utilities import constraints as cnst
from rigging_framework.templates.common.modules import fk_chain
from rigging_framework.templates.common.utilities import transform as trans_mod


def remove_secondary_joints(joint_chain, core_joint_names):
    """
    removes all the joints from the chain that don't have the core joint names in them and that are duplicates,
    physically deletes them from the scene and removes them from the joint_chain
    :param joint_chain:
    :param core_joint_names:
    :return:
    """
    filtered = []
    for joint in cmds.listRelatives(joint_chain[0], allDescendents=True):
        for core_joint in core_joint_names:
            if core_joint in joint and len(cmds.ls(joint)) == 1:
                filtered.append(joint)

    # deleting chain elements that are not core joints:
    for joint in cmds.listRelatives(joint_chain[0], allDescendents=True):
        if len(cmds.ls(joint)) > 1:
            for delete_me in cmds.ls(joint, long=True):
                if joint_chain[0] in delete_me:
                    cmds.delete(delete_me)

    return filtered


def build(name, start_joint, mid_joint, end_joint, scale=1.0, pole_vector_distance=1.0, virtual_constraints=False,
          name_or_none=None,
          replace_in_name=(("_Jnt", ""), ("FK_", "")),
          add_prefix="",
          add_suffix="_FK_Ctrl",
          ):
    """
    Builds the module based on given parameters
    This module has to be build on almost straight limb - it will not straighten the bent limb...
    
    :param str name: Name of the module
    :param str start_joint: Upper arm or Upper leg
    :param str mid_joint: Knee or elbow 
    :param str end_joint: Ankle or hand
    :param scale float scale: Scale of controls
    :param float pole_vector_distance: multiplier how far pole vector should land
    :param bool virtual_constraints: do we want to constraint right away or leave it for later
    :param str name_or_none: Name of the FK module - otherwise it will name after joints
    :param list(tuple) replace_in_name: Substrings in fk joints to be replaced default FK and JNT
    :param str add_prefix: add prefix to FK controls?
    :param str add_suffix:
    
    :return: dictionary 
        result = {"Main_Node": End locator of the limb system (Use this instead of IK)
                  "Controls": Pole vector control + FK controls
                  "Module_Group": Main module group
                  "Gizmos_Grp": Gizmos (Display stuff)
                  "Control_Groups": Offsets for all the controls above
                  "IKFK_Switch_Ctrl": IKFK switch control
                  "IK_Handle": IK handle node}
    """

    if not virtual_constraints:
        cnst.initialize_constraints()

    # build module group and snap it to the start_joint
    module_group = cmds.createNode("transform", name="{0}_Module_Group".format(name))
    cmds.delete(cmds.pointConstraint(start_joint, module_group))

    # add attribute on the module group so is easy to find/identify:
    cmds.addAttr(module_group, longName="moduleType", dataType="string")
    cmds.setAttr(module_group + ".moduleType", "IKFK_Limb", type="string")

    # create duplicate chain for FK
    fk_joint_chain = utils.duplicate_unparent(start_joint, depth=3, prefix='FK')
    fk_joint_chain = remove_secondary_joints(fk_joint_chain, [start_joint, mid_joint, end_joint])

    colors.set_color_from_str(fk_joint_chain[0], "blue")

    fk_start_joint = "FK_" + start_joint
    fk_mid_joint = "FK_" + mid_joint
    fk_end_joint = "FK_" + end_joint

    fk_joints_grp = utils.make_empty_group("{0}_FK_Joints_Grp".format(name))
    cmds.delete(cmds.pointConstraint(start_joint, fk_joints_grp))
    cmds.parent(fk_start_joint, fk_joints_grp)

    # build FK chain
    fk_data = fk_chain.build([fk_start_joint, fk_mid_joint, fk_end_joint],
                             name_or_none=name_or_none,
                             replace_in_name=replace_in_name,
                             add_prefix=add_prefix,
                             add_suffix=add_suffix,
                             scale=scale, scale_constraint=False, virtual_constraints=virtual_constraints)

    cmds.parent(fk_data["Module_Group"], module_group)
    fk_controls = fk_data["Controls"]
    fk_control_grp = fk_data["Control_Groups"][0]

    # create duplicate chain for IK
    ik_joint_chain = utils.duplicate_unparent(start_joint, depth=3, prefix='IK')
    ik_joint_chain = remove_secondary_joints(ik_joint_chain, [start_joint, mid_joint, end_joint])
    colors.set_color_from_str(ik_joint_chain[0], "red")

    ik_start_joint = "IK_" + start_joint
    ik_mid_joint = "IK_" + mid_joint
    ik_end_joint = "IK_" + end_joint

    ik_joints_grp = utils.make_empty_group("{0}_IK_Joints_Grp".format(name))
    cmds.delete(cmds.pointConstraint(start_joint, ik_joints_grp))
    cmds.parent(ik_start_joint, ik_joints_grp)

    # hook up reversal of the scale to ik and fk joints groups:
    trans_rev = cmds.createNode("transform", name="{0}_Scale_Reversal_Trans".format(name))
    cnst.scale_constraint(module_group, trans_rev)
    md_rev = cmds.createNode("multiplyDivide", name="{0}_Scale_Reversal_MD".format(name))
    cmds.setAttr(md_rev+".input1", 1, 1, 1)
    cmds.connectAttr(trans_rev + ".scale", md_rev + ".input2")
    cmds.setAttr(md_rev + ".operation", 2)
    cmds.connectAttr(md_rev + ".output", ik_joints_grp + ".scale")
    cmds.connectAttr(md_rev + ".output", fk_joints_grp + ".scale")
    cmds.parent(trans_rev, module_group)
    cmds.setAttr(trans_rev + ".inheritsTransform", False)

    # build upper and lower ik handles:
    ik_handle = cmds.ikHandle(startJoint=ik_start_joint, endEffector=ik_end_joint, sol="ikRPsolver",
                              name=name+"_Upper_IKHandle")[0]

    # build ik foot/hand locator
    end_locator = cmds.spaceLocator(name=name+"_IK_End_Loc")[0]
    utils.match_sec_to_first(ik_end_joint, end_locator, rotation=False)

    # now let's try to add stretchiness to the ik

    # add enums for the foot locator
    cmds.addAttr(end_locator, longName="stretch", attributeType="enum", enumName="Off:On:SquashMode", keyable=True)

    # add soft ik the foot locator...
    cmds.addAttr(end_locator, longName="soft_ik", attributeType="float", minValue=0,
                 defaultValue=0, maxValue=50, keyable=True)

    # build measuring curves, default, current, and for 3 bones - up, mid, low:
    mid_crv = utils.build_linear_curve(name="{}_len_crv".format(ik_mid_joint))
    low_crv = utils.build_linear_curve(name="{}_len_crv".format(ik_end_joint))

    cvs = cmds.ls('%s.cp[*]' % mid_crv, fl=True)
    clu1_handle_mid = cmds.cluster(cvs[0], name="{}_start_clu".format(mid_crv))[1]
    clu2_handle_mid = cmds.cluster(cvs[1], name="{}_end_clu".format(mid_crv))[1]
    cmds.delete(cmds.pointConstraint(ik_start_joint, clu1_handle_mid))
    cmds.delete(cmds.pointConstraint(ik_mid_joint, clu2_handle_mid))

    cvs = cmds.ls('%s.cp[*]' % low_crv, fl=True)
    clu1_handle_low = cmds.cluster(cvs[0], name="{}_start_clu".format(low_crv))[1]
    clu2_handle_low = cmds.cluster(cvs[1], name="{}_end_clu".format(low_crv))[1]
    cmds.delete(cmds.pointConstraint(ik_mid_joint, clu1_handle_low))
    cmds.delete(cmds.pointConstraint(ik_end_joint, clu2_handle_low))

    # default curve is the merger of two short ones:
    low_curve_dup = cmds.duplicate(low_crv)
    mid_curve_dup = cmds.duplicate(mid_crv)
    def_crv = cmds.attachCurve(mid_curve_dup, low_curve_dup, ch=0, rpo=1, kmk=1, m=0, bb=0.5, bki=0, p=0.1)[0]
    def_crv = cmds.rename(def_crv, "{}_def_len_crv".format(ik_start_joint))
    cmds.delete(low_curve_dup)

    cur_crv = cmds.duplicate(def_crv, name="{}_cur_len_crv".format(ik_start_joint))[0]
    cvs = cmds.ls('%s.cp[*]' % cur_crv, fl=True)
    clu1_handle_def = cmds.cluster(cvs[0], name="{}_start_clu".format(cur_crv))[1]
    clu_mid_def = cmds.cluster(cvs[1], name="{}_mid_clu".format(cur_crv))[1]
    clu2_handle_def = cmds.cluster(cvs[-1], name="{}_end_clu".format(cur_crv))[1]
    cnst.point_constraint(ik_start_joint, clu1_handle_def)
    cnst.point_constraint(end_locator, clu2_handle_def)
    cnst.point_constraint([clu1_handle_def, clu2_handle_def], clu_mid_def)

    # import ik system
    common_data_folder = os.path.dirname(os.path.realpath(__file__)).replace("modules", "data")
    path = os.path.join(common_data_folder, 'ik_system.ma')
    utils.import_and_add_prefix(path, name)
    soft_ik_sys = name + "_soft_ik_system"

    # hookup measuring curves to curve infos
    cmds.connectAttr(cur_crv+".worldSpace[0]", name + "_limb_cur_len_crvInfo.inputCurve")
    cmds.connectAttr(def_crv+".worldSpace[0]", name + "_limb_def_len_crvInfo.inputCurve")
    cmds.connectAttr(mid_crv+".worldSpace[0]", name + "_limb_up_len_crvInfo.inputCurve")
    cmds.connectAttr(low_crv+".worldSpace[0]", name + "_limb_low_len_crvInfo.inputCurve")

    # constraint ik between start joint and end locator (this cannot be defered)
    cns = cmds.pointConstraint(end_locator, ik_handle)[0]
    cmds.pointConstraint(start_joint, ik_handle)

    # connect the constraint to soft ik system
    cmds.connectAttr(soft_ik_sys + ".ik_adjust_percent_rev", cns + "." + end_locator + "W0")
    cmds.connectAttr(soft_ik_sys + ".ik_adjust_percent", cns + "." + start_joint + "W1")

    def connect_joint_length(incoming_connection, joint_input):
        """
        connects joint length but if its negative then reverses it
        """
        if cmds.getAttr(joint_input) > 0:
            cmds.connectAttr(incoming_connection, joint_input)
        else:
            rev_md = cmds.createNode("multiplyDivide", name="md_reverse_{}_length".format(joint_input.split('.')[0]))
            cmds.setAttr(rev_md+".input2", -1, -1, -1)
            cmds.connectAttr(incoming_connection, rev_md+".input1X")
            cmds.connectAttr(rev_md+".outputX", joint_input)

    # connect soft ik outputs for 2 chains to their respective joints lengths
    connect_joint_length(name + "_limb_up_ik_length_md.outputX", ik_mid_joint + joints.get_jnt_trans(ik_mid_joint))
    connect_joint_length(name + "_limb_low_ik_length_md.outputX", ik_end_joint + joints.get_jnt_trans(ik_end_joint))

    # hook up some more attributes to soft ik system
    cmds.connectAttr(end_locator + ".stretch", name + "_soft_ik_core_squashing_mode_crv.input")
    cmds.connectAttr(end_locator + ".stretch", name + "_soft_ik_core_stretchy_ik_crv.input")
    cmds.connectAttr(end_locator + ".soft_ik", soft_ik_sys + ".soft_ik")
    cmds.connectAttr(trans_rev + ".scaleX", soft_ik_sys + ".soft_ik_scale")

    # organize common elements into scaling and non_scaling
    ik_system_grp = utils.make_empty_group("{}_IK_System_Grp".format(name))
    cmds.parent(def_crv, ik_system_grp)
    cmds.parent(mid_crv, low_crv, ik_system_grp)
    cmds.parent(soft_ik_sys, ik_system_grp)
    cmds.parent(cur_crv, ik_system_grp)
    cmds.parent(clu1_handle_def, clu1_handle_low, clu1_handle_mid, ik_system_grp)
    cmds.parent(clu2_handle_def, clu2_handle_low, clu2_handle_mid, ik_system_grp)
    cmds.parent(clu_mid_def, ik_system_grp)
    cmds.parent(ik_handle, ik_system_grp)
    cmds.parent(end_locator, ik_system_grp)
    cmds.parent(ik_system_grp, module_group)

    # hook up global scale
    groups = [ik_joints_grp, fk_joints_grp]
    for group in groups:
        cmds.parent(group, ik_system_grp)

    # pole vector for knee or elbow
    pole_vector_control = rig_controls.make_control("box", name="{0}_Vector_IK_Ctrl".format(name),
                                                    size=scale * 0.5, color="red")

    # constraint the group in the middle of the two ends:
    pole_vector_control_group = trans_mod.make_space_for(pole_vector_control)

    # calculate the length of the limb
    limb_length = cmds.getAttr(name + "_limb_up_ik_length_md.outputX") + \
                  cmds.getAttr(name + "_limb_low_ik_length_md.outputX")

    # create a transform in the knee and snap it:
    parent_transform = cmds.createNode("transform")
    child_transform = cmds.createNode("transform")

    # calculate ratio of upper to lower (placing the pole vector in ideal spot):
    point_constraint = cmds.pointConstraint(ik_start_joint, ik_end_joint, parent_transform)[0]
    upper_percent = cmds.getAttr(name + "_limb_up_ik_length_md.outputX")
    lower_percent = cmds.getAttr(name + "_limb_low_ik_length_md.outputX")
    cmds.setAttr(point_constraint + ".{0}W0".format(ik_start_joint), lower_percent)
    cmds.setAttr(point_constraint + ".{0}W1".format(ik_end_joint), upper_percent)
    cmds.delete(point_constraint)

    cmds.aimConstraint(mid_joint, parent_transform)
    cmds.parent(child_transform, parent_transform)
    cmds.setAttr(child_transform + ".t", 0, 0, 0)
    cmds.setAttr(child_transform + ".r", 0, 0, 0)
    cmds.setAttr(child_transform + ".tx", limb_length * 0.25 * pole_vector_distance)

    cmds.delete(cmds.pointConstraint(child_transform, pole_vector_control_group))

    cmds.delete(parent_transform)

    # make the pole vector constraint
    cmds.poleVectorConstraint(pole_vector_control, ik_handle)

    lower_ik_guide = utils.build_linear_curve(name="{}_Knee_Guide_Crv".format(name))
    cmds.setAttr(lower_ik_guide + ".overrideEnabled", True)
    cmds.setAttr(lower_ik_guide + ".overrideDisplayType", 1)

    # snap those curve cvs to right joint using clusters:
    cvs = cmds.ls('%s.cp[*]' % lower_ik_guide, fl=True)
    cmds.xform(cvs[0], translation=cmds.xform(ik_mid_joint, q=True, rp=True, ws=True), ws=True)
    cmds.xform(cvs[1], translation=cmds.xform(pole_vector_control, q=True, rp=True, ws=True), ws=True)

    clu1_lower_handle_def = cmds.cluster(cvs[0], name="{}_start_clu".format(lower_ik_guide))[1]
    clu2_lower_handle_def = cmds.cluster(cvs[1], name="{}_end_clu".format(lower_ik_guide))[1]
    cnst.point_constraint(ik_mid_joint, clu1_lower_handle_def)
    cnst.point_constraint(pole_vector_control, clu2_lower_handle_def)

    # clean up those guide curves:
    cmds.parent(clu1_lower_handle_def, clu2_lower_handle_def,
                ik_system_grp)

    gizmo_group = utils.make_empty_group("{}_Gizmo_Grp".format(name))
    cmds.parent(lower_ik_guide, gizmo_group)
    cmds.parent(gizmo_group, ik_system_grp)

    cmds.setAttr(lower_ik_guide + ".inheritsTransform", False)

    # now build ikfk switching and visibility switching
    ikfk_switch_ctrl = rig_controls.make_control("diamond", name="{0}_IKFK_Switch_Ctrl".format(name),
                                                 size=scale * 0.2, color="red", rotate_comps=(0, 0, 90))

    ikfk_switch_ctrl_grp = trans_mod.make_space_for(ikfk_switch_ctrl)
    utils.match_sec_to_first(end_joint, ikfk_switch_ctrl_grp, rotation=False)
    cnst.point_constraint(end_joint, ikfk_switch_ctrl_grp)
    cmds.addAttr(ikfk_switch_ctrl, longName="ikfk_blend", minValue=0, maxValue=1, keyable=True)

    fk_vis_crv = driven_keys.build_quick_curve("{0}_FK_Visibility_Curve".format(name), [(0, 0), (1, 1)])
    cmds.connectAttr(ikfk_switch_ctrl + ".ikfk_blend", fk_vis_crv + ".input")
    cmds.connectAttr(fk_vis_crv + ".output", fk_control_grp + ".visibility")

    ik_vis_crv = driven_keys.build_quick_curve("{0}_IK_Visibility_Curve".format(name), [(0, 1), (1, 0)])
    cmds.connectAttr(ikfk_switch_ctrl + ".ikfk_blend", ik_vis_crv + ".input")
    cmds.connectAttr(ik_vis_crv + ".output", pole_vector_control_group + ".visibility")
    cmds.connectAttr(ik_vis_crv + ".output", gizmo_group + ".visibility")
    cmds.connectAttr(ik_vis_crv + ".output", end_locator + ".visibility")

    # build blendColors or pairBlends to blend ik and fk...
    for i, joint in enumerate([start_joint, mid_joint, end_joint]):
        fk_joint = "FK_" + joint
        ik_joint = "IK_" + joint
        pair_blend = cmds.createNode("pairBlend")
        blend_colors = cmds.createNode("blendColors")
        cmds.connectAttr(ik_joint + ".translate", pair_blend + ".inTranslate1")
        cmds.connectAttr(fk_joint + ".translate", pair_blend + ".inTranslate2")
        cmds.connectAttr(ik_joint + ".rotate", pair_blend + ".inRotate1")
        cmds.connectAttr(fk_joint + ".rotate", pair_blend + ".inRotate2")
        cmds.connectAttr(ik_joint + ".scale", blend_colors + ".color2")
        cmds.connectAttr(fk_joint + ".scale", blend_colors + ".color1")
        cmds.connectAttr(ikfk_switch_ctrl + ".ikfk_blend", pair_blend + ".weight")
        cmds.connectAttr(ikfk_switch_ctrl + ".ikfk_blend", blend_colors + ".blender")

        # we don't want to plug anything to the first joint translation as its in a different space
        if i != 0:
            # cmds.connectAttr(pair_blend + ".outTranslate", joint + ".translate")
            md_scale_mult = cmds.createNode("multiplyDivide")
            cmds.connectAttr(pair_blend + ".outTranslate", md_scale_mult + ".input1")
            cmds.connectAttr(md_rev + ".output", md_scale_mult + ".input2")
            cmds.connectAttr(md_scale_mult + ".output", joint + ".translate")

            cmds.connectAttr(pair_blend + ".outRotate", joint + ".rotate")
        else:
            # for top joint we use orientConstraint otherwise we will have issues with skeleton not locked in rotate
            orient_cns = cmds.orientConstraint(fk_joint, ik_joint, joint, maintainOffset=True)[0]
            cmds.connectAttr(fk_vis_crv + ".output", "{0}.{1}W0".format(orient_cns, fk_joint))
            cmds.connectAttr(ik_vis_crv + ".output", "{0}.{1}W1".format(orient_cns, ik_joint))

            # let's add point constraint for the first joint (EDIT):
            point_cns = cmds.pointConstraint(fk_joint, ik_joint, joint, maintainOffset=True)[0]
            cmds.connectAttr(fk_vis_crv + ".output", "{0}.{1}W0".format(point_cns, fk_joint))
            cmds.connectAttr(ik_vis_crv + ".output", "{0}.{1}W1".format(point_cns, ik_joint))

        cmds.connectAttr(blend_colors + ".output", joint + ".scale")

    # add ik mode attribute:
    cmds.addAttr(ikfk_switch_ctrl, attributeType="enum", longName="ik_mode",
                 enumName="Regular:Stretchy:Squash", keyable=True)
    cmds.connectAttr(ikfk_switch_ctrl + ".ik_mode", end_locator + ".stretch")

    # clean up - create system control group
    control_group = utils.make_empty_group("{}_Ctrl_Grp".format(name))
    cmds.parent(pole_vector_control_group, control_group)
    cmds.parent(ikfk_switch_ctrl_grp, control_group)
    cmds.parent(control_group, module_group)

    cmds.delete(mid_crv, low_crv, constructionHistory=True)
    cmds.setAttr(cur_crv + ".inheritsTransform", False)

    cmds.setAttr(ik_system_grp + ".visibility", False)

    if not virtual_constraints:
        cnst.build_constraints()

    result = {"Main_Node": end_locator,
              "Controls": [pole_vector_control] + fk_controls,
              "Module_Group": module_group,
              "Gizmos_Grp": gizmo_group,
              "Control_Groups": [trans_mod.get_parent(x) for x in [pole_vector_control] + fk_controls],
              "IKFK_Switch_Ctrl": ikfk_switch_ctrl,
              "IK_Handle": ik_handle}

    return result


def gather_data_from_placer(placer_transform):
    """
    Gathers data from placer top node
    :param placer_transform:
    :return:
    """
    node = [x for x in cmds.listRelatives(placer_transform, children=True, fullPath=True)
            if x.split("|")[-1].startswith("Name__")][0]
    module_name = node.split("__")[-1]

    placers = {}
    for placer_str in "Arm_Hip", "Elbow_Knee", "Wrist_Ankle":
        node = [x for x in cmds.listRelatives(placer_transform, children=True, fullPath=True)
                if x.split("|")[-1].startswith(placer_str)][0]
        cns = node + "|" + cmds.listConnections(node + ".tx", type="pointConstraint")[0]
        placers[placer_str] = cmds.pointConstraint(cns, q=True, targetList=True)[0]

    # get connection info:
    plugged_name = None
    output = ""
    for x in cmds.listRelatives(placer_transform, children=True, fullPath=True):
        if x.endswith("Plug") and "Input" in x:
            cns = cmds.listConnections(x + ".tx", type="pointConstraint")[0]
            output = cmds.pointConstraint(cns, q=True, targetList=True)[0]

            plugged_placer = cmds.listRelatives(output, parent=True)[0]
            if plugged_placer != placer_transform:
                node = [z for z in cmds.listRelatives(plugged_placer, children=True, fullPath=True)
                        if z.split("|")[-1].startswith("Name__")][0]
                plugged_name = node.split("__")[-1]

    return dict(module_name=module_name,
                start_joint=placers["Arm_Hip"],
                mid_joint=placers["Elbow_Knee"],
                end_joint=placers["Wrist_Ankle"],
                input_module=plugged_name,
                input_plug=output.split("|")[-1] if plugged_name else None)


def set_data_to_placer(placer_transform, placer_data):
    """
    Sets data from placer top node
    :param placer_transform:
    :param placer_data:
    :return:
    """
    node = [x for x in cmds.listRelatives(placer_transform, children=True, fullPath=True)
            if x.split("|")[-1].startswith("Name__")][0]
    cmds.rename(node, "Name__{}".format(placer_data["module_name"]))

    cmds.pointConstraint(placer_data["start_joint"], placer_transform + "|Arm_Hip_Placer_Joint")
    cmds.pointConstraint(placer_data["mid_joint"], placer_transform + "|Elbow_Knee_Placer_Joint")
    cmds.pointConstraint(placer_data["end_joint"], placer_transform + "|Wrist_Ankle_Placer_Joint")


def load_placer_connection(placer_transform, placer_data):
    """
    This loads connection based on the placer data
    :param placer_transform:
    :param placer_data:
    :return:
    """
    # set connection info:
    if placer_data["input_module"]:
        connected_module = [x for x in cmds.ls(assemblies=True) if "Name__{}".format(placer_data["input_module"])
                            in cmds.listRelatives(x, children=True)][0]

        for x in cmds.listRelatives(placer_transform, children=True, fullPath=True):
            if x.endswith("Plug") and "Input" in x:

                constraint = cmds.listConnections(x, type="constraint", source=True, destination=False)
                if constraint:
                    cmds.delete(constraint)

                cmds.pointConstraint(connected_module + "|" + placer_data["input_plug"], x)


def get_data_for_module(scene_name=""):
    """
    Gathers data for other module
    :param scene_name:
    :return:
    """
    actual_scene_name = "{0}_Module_Group".format(scene_name)

    ikfk_switch_ctrl = [x for x in cmds.listRelatives(actual_scene_name, allDescendents=True, type="transform")
                        if x.endswith("IKFK_Switch_Ctrl")][0]

    ikfk_switch_attr = "ikfk_blend"

    main_ik_loc = [x for x in cmds.listRelatives(actual_scene_name, allDescendents=True, type="transform")
                   if x.endswith("IK_End_Loc")]

    fk_bones = cmds.listRelatives(scene_name + "_FK_Joints_Grp", allDescendents=True, type="joint")

    fk_end_bone = fk_bones[0]

    ik_bones = cmds.listRelatives(scene_name + "_IK_Joints_Grp", allDescendents=True, type="joint")

    ik_start_bone = ik_bones[-1]
    ik_end_bone = ik_bones[0]

    start_bone = ik_start_bone.replace("IK_", "")
    end_bone = ik_end_bone.replace("IK_", "")

    fk_start_joint = ik_start_bone.replace("IK_", "FK_")
    fk_start_joint_cns = cmds.listRelatives(fk_start_joint, children=True, type="parentConstraint")[0]
    fk_start_ctrl = cmds.parentConstraint(fk_start_joint_cns, q=True, targetList=True)[0]
    fk_start_ctrl_grp = cmds.listRelatives(fk_start_ctrl, parent=True)[0]

    return dict(
                ik_fk_switch_ctrl=ikfk_switch_ctrl,
                ik_fk_switch_attr=ikfk_switch_attr,
                main_ik_loc=main_ik_loc,
                fk_end_bone=fk_end_bone,
                start_bone=start_bone,
                end_bone=end_bone,
                ik_start_bone=ik_start_bone,
                ik_end_bone=ik_end_bone,
                fk_start_ctrl_grp=fk_start_ctrl_grp,
                )


def match_with_offset(obj1, obj2, offset_obj):
    """
    Match control with offset
    :param obj1:
    :param obj2:
    :param offset_obj:
    :return:
    """
    trans_mod.zero_out(obj1)

    obj_trs = cmds.createNode("transform")
    obj_trs_child = cmds.createNode("transform")
    cmds.parent(obj_trs_child, obj_trs)
    cmds.matchTransform(obj_trs, offset_obj)
    cmds.matchTransform(obj_trs_child, obj1)

    cmds.matchTransform(obj_trs, obj2)
    cmds.matchTransform(obj1, obj_trs_child)

    cmds.delete(obj_trs)


def ikfk_match(module_grp, ik_to_fk=True):
    """
    Matches ik to fk and fk to ik
    :param module_grp:
    :param ik_to_fk:
    :return:
    """
    module_name = module_grp.replace("_Module_Group", "")

    ik_main_node = "{}_IK_End_Loc".format(module_name)

    vector_ctrl = [x for x in cmds.listRelatives(module_grp, allDescendents=True) if x.endswith("Vector_IK_Ctrl")][0]

    cns = cmds.listConnections(ik_main_node + ".translateX")
    if not cns:
        limb_ik_ctrl = ik_main_node
        while not limb_ik_ctrl.endswith("_Ctrl"):
            limb_ik_ctrl = cmds.listRelatives(limb_ik_ctrl, parent=True)[0]
    else:
        limb_ik_ctrl = cmds.pointConstraint(cns, q=True, targetList=True)[0]

    fk_bones = cmds.listRelatives(module_name + "_FK_Joints_Grp", allDescendents=True, type="joint")[-3:]
    ik_bones = [x.replace("FK_", "IK_") for x in fk_bones]  # if feet are included we dont want them in here ^^^

    cns = cmds.listConnections(fk_bones[-1] + ".translateX")[0]
    start_fk_ctrl = cmds.parentConstraint(cns, q=True, targetList=True)[0]

    cns = cmds.listConnections(fk_bones[1] + ".translateX")[0]
    mid_fk_ctrl = cmds.parentConstraint(cns, q=True, targetList=True)[0]

    cns = cmds.listConnections(fk_bones[0] + ".translateX")[0]
    end_fk_ctrl = cmds.parentConstraint(cns, q=True, targetList=True)[0]

    # print vector_ctrl, limb_ik_ctrl, start_fk_ctrl, mid_fk_ctrl, end_fk_ctrl
    if ik_to_fk:
        match_with_offset(limb_ik_ctrl, end_fk_ctrl, ik_bones[0])  # this is in case there is offset between ik and fk

        cmds.matchTransform(vector_ctrl, mid_fk_ctrl)

        # Offset the pole vector out
        trs = cmds.createNode("transform")
        trs_child = cmds.createNode("transform")
        cmds.parent(trs_child, trs)
        cmds.setAttr(trs_child + ".tx", (abs(cmds.getAttr(fk_bones[0] + ".tx")) + abs(cmds.getAttr(fk_bones[1] + ".tx"))) / 1.5)
        cns = cmds.pointConstraint(start_fk_ctrl, end_fk_ctrl, trs)[0]
        cmds.setAttr(cns + "." + start_fk_ctrl + "W0", abs(cmds.getAttr(mid_fk_ctrl + ".tx")))
        cmds.setAttr(cns + "." + end_fk_ctrl + "W1", abs(cmds.getAttr(end_fk_ctrl + ".tx")))
        cmds.aimConstraint(mid_fk_ctrl, trs)

        cmds.matchTransform(vector_ctrl, trs_child)
        cmds.delete(trs)
    else:  # fk to IK:
        cmds.matchTransform(start_fk_ctrl, ik_bones[-1], position=False)
        cmds.matchTransform(mid_fk_ctrl, ik_bones[1], position=False)
        cmds.matchTransform(end_fk_ctrl, ik_bones[0], position=False)
