import maya.cmds as cmds

import maya.api.OpenMaya as apiOM

import rigging_framework.templates.common.utilities.transform as trans_mod
from rigging_framework.templates.common.utilities import utils
from rigging_framework.templates.common.utilities import colors
from rigging_framework.templates.common.utilities import rig_controls
from rigging_framework.templates.common.utilities import rigging_math
from rigging_framework.templates.common.utilities import constraints as cnst
from rigging_framework.templates.humanoid_rig.modules import ikfk_limb


def remove_secondary_joints(joint_chain, core_joint_names):
    """
    removes all the joints from the chain that don't have the core joint names in them and that are duplicates,
    physically deletes them from the scene and removes them from the joint_chain
    :param joint_chain:
    :param core_joint_names:
    :return:
    """
    filtered = []
    for joint in cmds.listRelatives(joint_chain[0], allDescendents=True):
        for core_joint in core_joint_names:
            if core_joint in joint and len(cmds.ls(joint)) == 1:
                filtered.append(joint)

    # deleting chain elements that are not core joints:
    for joint in cmds.listRelatives(joint_chain[0], allDescendents=True):
        if len(cmds.ls(joint)) > 1:
            for delete_me in cmds.ls(joint, long=True):
                if joint_chain[0] in delete_me:
                    cmds.delete(delete_me)

    return filtered


def switch_constraint(constraint, target_object):
    """
    Switches constraint to target object
    :param constraint:
    :param target_object:
    :return:
    """
    if cmds.nodeType(constraint) == "pointConstraint":
        constraint_type = "pointConstraint"
        targets = cmds.pointConstraint(constraint, q=True, targetList=True)
    elif cmds.nodeType(constraint) == "orientConstraint":
        constraint_type = "orientConstraint"
        targets = cmds.orientConstraint(constraint, q=True, targetList=True)
    else:
        cmds.error("Constraint type not known: {0}".format(cmds.nodeType(constraint)))
        return

    target_plugs = {}

    for i, target in enumerate(targets):
        conn = cmds.listConnections(constraint + "." + target + "W{0}".format(i),
                                    source=True, destination=False, plugs=True)
        if conn:
            target_plugs[target + "W{0}".format(i)] = conn[0]

    if constraint_type == "pointConstraint":
        cns = cmds.pointConstraint(targets, target_object, maintainOffset=True)[0]
    else:
        cns = cmds.orientConstraint(targets, target_object, maintainOffset=True)[0]

    for target_attr, plug in target_plugs.items():
        cmds.connectAttr(plug, cns + "." + target_attr)

    cmds.delete(constraint)


def build(name, start_joint, mid_joint, end_joint, scale=1.0, pole_vector_distance=1.0, aim_vector=(1, 0, 0),
          up_vector=(0, 1, 0), virtual_constraints=False):
    """
    Builds the module based on given parameters
    This module has to be build on almost straight limb - it will not straighten the limb...
    
    :param str name: Name of the module
    :param str start_joint: Upper arm or upper leg
    :param str mid_joint: Middle joint like knee or elbow
    :param str end_joint: End joint like ankle or hand 
    :param float scale: Scale of the controls
    :param float pole_vector_distance: Multiplier how far pole vector should land
    :param float[3] aim_vector: What is the aim vector for this system - where should it be aiming forward
    :param float[3] up_vector: Whats the up vector fir this system - what is up for it in the coordinate system
    :param bool virtual_constraints: Do we want to build constraints or leave that for later
    
    :return dictionary: 
        result = {"Main_Node": Main limb node - locator in the hand/ankle which can be used to attach things to,
                  "Controls": IKFK limb controls + tweaker control,
                  "Module_Group": Main module group,
                  "Gizmos_Grp": Gizmos - display elements (guide curves),
                  "Control_Groups": IKFK limb control groups and tweaker control group,
                  "IKFK_Switch_Ctrl": IKFK switch control name}
    
    """

    # calculate orientation based on up and aim:
    module_matrix = rigging_math.look_at(apiOM.MVector(aim_vector), apiOM.MVector(up_vector))
    module_transform = apiOM.MTransformationMatrix(module_matrix)

    # build module group and snap it to the start_joint
    module_group = cmds.createNode("transform", name="{0}_Tweak_Module_Group".format(name))
    cmds.xform(module_group, matrix=module_transform.asMatrix(), worldSpace=True)
    orientation = cmds.getAttr(module_group + ".rotate")[0]
    cmds.delete(cmds.pointConstraint(start_joint, module_group))

    if not virtual_constraints:
        cnst.initialize_constraints()

    # create duplicate chain for TWEAK
    tweak_joint_chain = utils.duplicate_unparent(start_joint, depth=3, prefix='TWEAK')
    tweak_joint_chain = remove_secondary_joints(tweak_joint_chain, [start_joint, mid_joint, end_joint])

    tweak_start_joint = "TWEAK_" + start_joint
    tweak_mid_joint = "TWEAK_" + mid_joint
    tweak_end_joint = "TWEAK_" + end_joint

    colors.set_color_from_str(tweak_start_joint, "green")

    tweak_joints_grp = utils.make_empty_group("{0}_TWEAK_Joints_Grp".format(name))
    cmds.delete(cmds.pointConstraint(start_joint, tweak_joints_grp))
    cmds.parent(tweak_start_joint, tweak_joints_grp)

    # create a tweak control for the knee:
    tweak_control = rig_controls.make_control(control_type="sphere", size=scale, name="{0}_Tweak_Ctrl".format(name),
                                              color="light_green")
    tweak_control_grp = trans_mod.make_space_for(tweak_control)
    cmds.xform(tweak_control_grp, rotation=orientation, worldSpace=True)

    cmds.delete(cmds.pointConstraint(tweak_mid_joint, tweak_control_grp))
    cnst.parent_constraint(tweak_control, tweak_mid_joint, maintain_offset=True)

    ikfk_limb_data = ikfk_limb.build(name, start_joint, mid_joint, end_joint, scale,
                                     pole_vector_distance, virtual_constraints)

    cmds.parent(tweak_control_grp, module_group)
    cmds.parent(ikfk_limb_data["Module_Group"], module_group)

    tweak_mid_joint_proxy = utils.duplicate_no_kids(tweak_mid_joint, rename=tweak_mid_joint + "_PROXY")
    tweak_start_joint_proxy = utils.duplicate_no_kids(tweak_mid_joint, rename=tweak_start_joint + "_PROXY")
    trans_mod.zero_out(tweak_start_joint_proxy)

    cnst.parent_constraint(tweak_mid_joint, tweak_control_grp)
    cnst.scale_constraint(tweak_mid_joint, tweak_control_grp)

    cnst.point_constraint(tweak_control, tweak_mid_joint_proxy)

    for attr in [".r", ".rx", ".ry", ".rz"]:
        cmds.setAttr(tweak_control + attr, lock=True)
        cmds.setAttr(tweak_control + attr, keyable=False)

    # this bit doesnt work too well - is it even needed? :
    world_up_node = cmds.createNode("transform", name="{0}_world_up_obj".format(name))
    trans_mod.parent_and_zero(world_up_node, tweak_mid_joint)

    # this has to be moved by a vector perpendicular to the pole Vector...
    pole_vector_position = cmds.xform(ikfk_limb_data["Controls"][0], q=True, translation=True, worldSpace=True)
    tweak_mid_position = cmds.xform(tweak_mid_joint, q=True, translation=True, worldSpace=True)
    tweak_end_position = cmds.xform(tweak_end_joint, q=True, translation=True, worldSpace=True)

    pole_vector_vector = apiOM.MVector(pole_vector_position) - apiOM.MVector(tweak_mid_position)

    # find a perpendicular vector:
    limb_vector = apiOM.MVector(tweak_end_position) - apiOM.MVector(tweak_mid_position)

    # now find perpendicular vector to those two vectors:
    cross_vector = apiOM.MVector(pole_vector_vector ^ limb_vector)
    cross_vector.normalize()

    # move the world up node by that perpendicular vector
    cmds.move(cross_vector.x * 50.0, cross_vector.y * 50.0, cross_vector.z * 50.0, world_up_node,
              relative=True, worldSpace=True)

    cmds.aimConstraint(tweak_mid_joint_proxy, tweak_start_joint_proxy, maintainOffset=True,
                       worldUpObject=world_up_node, worldUpType="object")
    cmds.aimConstraint(tweak_end_joint, tweak_mid_joint_proxy, maintainOffset=True,
                       worldUpObject=world_up_node, worldUpType="object")

    cmds.parent(tweak_joints_grp, module_group)
    cmds.setAttr(tweak_joints_grp + ".inheritsTransform", False)
    cmds.delete(cmds.pointConstraint(start_joint, tweak_joints_grp))

    # now switch all the connections from ikfk joints to tweak joints:
    for attr in ["t", "r", "s", "tx", "ty", "tz", "rx", "ry", "rz", "sx", "sy", "sz"]:
        for joint, tweak_joint in zip([start_joint, mid_joint, end_joint],
                                      [tweak_start_joint, tweak_mid_joint, tweak_end_joint]):
            connection = cmds.listConnections(joint + "." + attr, source=True, destination=False, plugs=True)
            # switch constraints:
            constraints = cmds.listRelatives(joint, children=True, type="constraint")
            if constraints:
                for constraint in constraints:
                    switch_constraint(constraint, tweak_joint)

            # switch connections:
            if connection:
                cmds.connectAttr(connection[0], tweak_joint + "." + attr)
                cmds.disconnectAttr(connection[0], joint + "." + attr)

    # now constraint target chain to middle tweak one (proxies):
    for joint, tweak_joint in zip([start_joint, mid_joint, end_joint],
                                  [tweak_start_joint_proxy, tweak_mid_joint_proxy, tweak_end_joint]):
        cnst.parent_constraint(tweak_joint, joint, maintain_offset=True)
        cnst.scale_constraint(tweak_joint, joint, maintain_offset=True)

    # replace point constraint on ik Handle:
    ik_handle = ikfk_limb_data["IK_Handle"]
    bad_point_constraint = cmds.listRelatives(ik_handle, children=True, type="pointConstraint")[0]

    targets = cmds.pointConstraint(bad_point_constraint, q=True, targetList=True)

    w0 = cmds.listConnections("{0}.{1}W0".format(bad_point_constraint, targets[0]),
                              source=True, destination=False, plugs=True)[0]
    w1 = cmds.listConnections("{0}.{1}W1".format(bad_point_constraint, targets[1]),
                              source=True, destination=False, plugs=True)[0]
    cmds.delete(bad_point_constraint)
    new_cns = cmds.pointConstraint(targets[0], tweak_start_joint, ik_handle)[0]
    cmds.connectAttr(w0, new_cns + ".{0}W0".format(targets[0]))
    cmds.connectAttr(w1, new_cns + ".{0}W1".format(tweak_start_joint))

    cmds.setAttr(tweak_joints_grp + ".visibility", False)  # hide tweak joint group
    cmds.scaleConstraint(module_group, tweak_joints_grp, maintainOffset=True)  # this is needed or system won't scale

    # now build remaining constraints:

    if not virtual_constraints:
        cnst.build_constraints()

    result = {"Main_Node": ikfk_limb_data["Main_Node"],
              "Controls": ikfk_limb_data["Controls"] + [tweak_control],
              "Module_Group": module_group,
              "Gizmos_Grp": ikfk_limb_data["Gizmos_Grp"],
              "Control_Groups": ikfk_limb_data["Control_Groups"] + [tweak_control_grp],
              "IKFK_Switch_Ctrl": ikfk_limb_data["IKFK_Switch_Ctrl"]}

    return result
