import maya.api.OpenMaya as apiOM
import maya.cmds as cmds

import rigging_framework.templates.common.utilities.transform as trans_mod
from rigging_framework.templates.common.utilities import utils
from rigging_framework.templates.common.utilities import colors
from rigging_framework.templates.common.utilities import joints
from rigging_framework.templates.common.utilities import driven_keys
from rigging_framework.templates.common.utilities import constraints as cnst
from rigging_framework.templates.common.utilities import rigging_math
from rigging_framework.templates.common.modules import fk_chain
from rigging_framework.templates.humanoid_rig.modules import ik_ribbon_spine


def build(name, spine_joints, ik_drives_fk=True, scale=1.0, up_vector=(0, 1, 0), aim_vector=(1, 0, 0),
          virtual_constraints=False):
    """
    Builds ikfk ribbon spine with a switch for biped
    
    :param str name: Name of the module
    :param list[str] spine_joints: List of spine joints
    :param bool ik_drives_fk: Default build is IK driving FK, but there is opposite option if needed
    :param float scale: Scale of the controls
    :param float[3] up_vector: Determines the up of the control and the ribbon
    :param float[3] aim_vector: Determines the forward of the control and the ribbon
    :param bool virtual_constraints: Whether or not constrains are going to be built right away or virtually
    
    result = {"Main_Node": Main COG control node
              "Controls": IK spine controls + FK spine controls
              "Module_Group": Main module group
              "Control_Groups": IK spine control groups + FK spine control groups,
              "Ribbon_Surface": Name of the ribbon surface node,
              "Ribbon_Tweak_Controls": Tweaker controls for the ribbon}
    """

    if not virtual_constraints:
        cnst.initialize_constraints()

    # calculate orientation based on up and aim:
    module_matrix = rigging_math.look_at(apiOM.MVector(aim_vector), apiOM.MVector(up_vector))
    module_transform = apiOM.MTransformationMatrix(module_matrix)

    # create main folders:
    module_group = cmds.createNode("transform", name="{0}_Module_Group".format(name))
    cmds.xform(module_group, matrix=module_transform.asMatrix(), worldSpace=True)

    cmds.addAttr(module_group, longName="moduleType", dataType="string")
    cmds.setAttr(module_group + ".moduleType", "IKFK_Spine", type="string")

    # create duplicate chain for FK
    fk_spine_joints = utils.duplicate_unparent(spine_joints[0], depth=len(spine_joints), prefix='FK',
                                               favor=spine_joints[1])

    colors.set_color_from_str(fk_spine_joints[0], "blue")

    # create duplicate chain for IK
    ik_spine_joints = utils.duplicate_unparent(spine_joints[0], depth=len(spine_joints), prefix='IK',
                                               favor=spine_joints[1])
    colors.set_color_from_str(ik_spine_joints[0], "red")

    # now lets xform all the spine joints so that we don't get into joint orients problems:
    joints.xform_joints(fk_spine_joints)
    joints.xform_joints(ik_spine_joints)
    joints.xform_joints(spine_joints)

    cmds.setAttr(fk_spine_joints[0] + ".visibility", 0)
    cmds.setAttr(ik_spine_joints[0] + ".visibility", 0)

    # build ik ribbon spine
    ik_spine_data = ik_ribbon_spine.build("{0}_Ik".format(name), ik_spine_joints, scale=scale, up_vector=up_vector,
                                          aim_vector=aim_vector, virtual_constraints=virtual_constraints)

    # build simple fk spine
    fk_chain_data = fk_chain.build(fk_spine_joints, name_or_none=None, add_suffix="_Fk_Ctrl",
                                   scale_constraint=True, scale=scale, virtual_constraints=virtual_constraints)

    spine_cog_ctrl = ik_spine_data["Main_Node"]
    fk_chain_top_grp = fk_chain_data["Module_Group"]
    cmds.parent(fk_chain_top_grp, spine_cog_ctrl)

    # add ik fk switch attribute on spine COG control
    cmds.addAttr(spine_cog_ctrl, longName="ikfk_blend", minValue=0, maxValue=1, keyable=True)

    # create ik and fk visibility curves and hook them up to controls
    fk_vis_crv = driven_keys.build_quick_curve("{0}_FK_Visibility_Curve".format(name), [(0, 0), (1, 1)])
    cmds.connectAttr(spine_cog_ctrl + ".ikfk_blend", fk_vis_crv + ".input")
    cmds.connectAttr(fk_vis_crv + ".output", fk_chain_top_grp + ".visibility")

    ik_vis_crv = driven_keys.build_quick_curve("{0}_IK_Visibility_Curve".format(name), [(0, 1), (1, 0)])
    cmds.connectAttr(spine_cog_ctrl + ".ikfk_blend", ik_vis_crv + ".input")

    for ik_control_grp in ik_spine_data["Controls_Grp"]:
        if ik_control_grp != trans_mod.get_parent(spine_cog_ctrl):
            cmds.connectAttr(ik_vis_crv + ".output", ik_control_grp + ".visibility")

    # now build the blending for the ikfk spine:
    for i, (joint, fk_joint, ik_joint) in enumerate(zip(spine_joints, fk_spine_joints, ik_spine_joints)):
        if i == 0:
            cns = cmds.parentConstraint(fk_joint, ik_joint, joint, maintainOffset=True)[0]
            cmds.connectAttr(fk_vis_crv + ".output", cns + ".{0}W0".format(fk_joint))
            cmds.connectAttr(ik_vis_crv + ".output", cns + ".{0}W1".format(ik_joint))
        else:
            cns = cmds.parentConstraint(fk_joint, ik_joint, joint)[0]
            cmds.connectAttr(fk_vis_crv + ".output", cns + ".{0}W0".format(fk_joint))
            cmds.connectAttr(ik_vis_crv + ".output", cns + ".{0}W1".format(ik_joint))

            blend_colors = cmds.createNode("blendColors")
            cmds.connectAttr(ik_joint + ".scale", blend_colors + ".color2")
            cmds.connectAttr(fk_joint + ".scale", blend_colors + ".color1")
            cmds.connectAttr(spine_cog_ctrl + ".ikfk_blend", blend_colors + ".blender")
            cmds.connectAttr(blend_colors + ".output", joint + ".scale")

    # now add ik drives fk functionality:
    if ik_drives_fk:
        cmds.addAttr(spine_cog_ctrl, longName="ik_drives_fk", attributeType="enum", enumName="Off:On", keyable=True)

        # override visibility of the ik controls so they always on in ik drives fk mode:
        for ik_control in ik_spine_data["Controls"][1:]:
            ik_control_grp = trans_mod.get_parent(ik_control)
            vis_input = cmds.listConnections(ik_control_grp + ".visibility", source=True, destination=False, plugs=True)[0]
            condition_node = cmds.createNode("condition")
            cmds.connectAttr(spine_cog_ctrl + ".ik_drives_fk", condition_node + ".firstTerm")
            cmds.connectAttr(vis_input, condition_node + ".colorIfTrueR")
            cmds.setAttr(condition_node + ".colorIfFalseR", 1)
            utils.disconnect_attr(ik_control_grp + ".visibility")
            cmds.connectAttr(condition_node + ".outColorR", ik_control_grp + ".visibility")

        # override ikfk-blend so its always 1
        outputs_from_ikfk_blend = cmds.listConnections(spine_cog_ctrl + ".ikfk_blend", source=False,
                                                       destination=True, plugs=True)
        condition_node = cmds.createNode("condition")
        cmds.connectAttr(spine_cog_ctrl + ".ik_drives_fk", condition_node + ".firstTerm")
        cmds.connectAttr(spine_cog_ctrl + ".ikfk_blend", condition_node + ".colorIfTrueR")
        cmds.setAttr(condition_node + ".colorIfFalseR", 1)
        for output in outputs_from_ikfk_blend:
            utils.disconnect_attr(output)
            cmds.connectAttr(condition_node + ".outColorR", output)

        # plug in rotation and translation of the ik joints to the fk groups
        for i, (ik_joint, fk_ctrl) in enumerate(zip(ik_spine_joints, fk_chain_data["Controls"])):
            fk_ctrl_grp = trans_mod.get_parent(fk_ctrl)
            pair_blend = cmds.createNode("pairBlend")

            if i == 0:
                # make offset group for first fk:
                offset_fk_ctrl_grp = cmds.createNode("transform", name=fk_ctrl_grp.replace("Grp", "Offset_Grp"))
                trans_mod.parent_and_zero(offset_fk_ctrl_grp, fk_chain_data["Module_Group"])
                cmds.parent(fk_ctrl_grp, offset_fk_ctrl_grp)
                fk_ctrl_grp = offset_fk_ctrl_grp

                cmds.connectAttr(ik_spine_data["Controls"][1] + ".translate", pair_blend + ".inTranslate2")
                cmds.setAttr(pair_blend + ".inTranslate1", *cmds.getAttr(fk_ctrl_grp + ".translate")[0])
                cmds.connectAttr(ik_spine_data["Controls"][1] + ".rotate", pair_blend + ".inRotate2")
                cmds.setAttr(pair_blend + ".inRotate1", *cmds.getAttr(fk_ctrl_grp + ".rotate")[0])
            else:
                cmds.connectAttr(ik_joint + ".translate", pair_blend + ".inTranslate2")
                cmds.setAttr(pair_blend + ".inTranslate1", *cmds.getAttr(fk_ctrl_grp + ".translate")[0])
                cmds.connectAttr(ik_joint + ".rotate", pair_blend + ".inRotate2")
                cmds.setAttr(pair_blend + ".inRotate1", *cmds.getAttr(fk_ctrl_grp + ".rotate")[0])

            cmds.connectAttr(spine_cog_ctrl + ".ik_drives_fk", pair_blend + ".weight")

            cmds.connectAttr(pair_blend + ".outTranslate", fk_ctrl_grp + ".translate")
            cmds.connectAttr(pair_blend + ".outRotate", fk_ctrl_grp + ".rotate")

    # cleanup a little:
    cmds.parent(ik_spine_data["Module_Group"], module_group)
    cmds.parent(ik_spine_joints[0], module_group)
    cmds.parent(fk_spine_joints[0], module_group)

    # build constraints:
    if not virtual_constraints:
        cnst.build_constraints()

    # return values
    result = {"Main_Node": spine_cog_ctrl,
              "Controls": ik_spine_data["Controls"] + fk_chain_data["Controls"],
              "Module_Group": module_group,
              "Control_Groups": [trans_mod.get_parent(x) for x in ik_spine_data["Controls"] + fk_chain_data["Controls"]],
              "Ribbon_Surface": ik_spine_data["Ribbon_Surface"],
              "Ribbon_Tweak_Controls": ik_spine_data["Ribbon_Tweak_Controls"]}

    return result


def gather_data_from_placer(placer_transform):
    """
    Gathers data from placer top node
    :param placer_transform:
    :return:
    """
    node = [x for x in cmds.listRelatives(placer_transform, children=True, fullPath=True)
            if x.split("|")[-1].startswith("Name__")][0]
    module_name = node.split("__")[-1]

    placers = {}
    for placer_str in "Spine_Start", "Spine_End":
        node = [x for x in cmds.listRelatives(placer_transform, children=True, fullPath=True)
                if x.split("|")[-1].startswith(placer_str)][0]
        cns = node + "|" + cmds.listConnections(node + ".tx", type="pointConstraint")[0]
        placers[placer_str] = cmds.pointConstraint(cns, q=True, targetList=True)[0]

    return dict(module_name=module_name,
                start_joint=placers["Spine_Start"],
                end_joint=placers["Spine_End"])


def set_data_to_placer(placer_transform, placer_data):
    """
    Sets data from placer top node
    :param placer_transform:
    :param placer_data:
    :return:
    """
    node = [x for x in cmds.listRelatives(placer_transform, children=True, fullPath=True)
            if x.split("|")[-1].startswith("Name__")][0]
    cmds.rename(node, "Name__{}".format(placer_data["module_name"]))

    cmds.pointConstraint(placer_data["start_joint"], placer_transform + "|Spine_Start_Placer_Joint")
    cmds.pointConstraint(placer_data["end_joint"], placer_transform + "|Spine_End_Placer_Joint")


def get_data_for_module(module_type="hips_rig", scene_name=""):
    """
    Gathers data for other module
    :param module_type:
    :param scene_name:
    :return:
    """
    actual_scene_name = "{0}_Module_Group".format(scene_name)

    if not cmds.objExists(actual_scene_name):
        raise RuntimeError("IKFK_Spine - cannot get data for non-existing node: {}".format(actual_scene_name))

    if module_type == "hips_rig" or module_type == "hand_rig":
        cog_control = [x for x in cmds.listRelatives(actual_scene_name, allDescendents=True, type="transform")
                       if x.endswith("Ik_COG_Ctrl")][0]

        return dict(cog_control=cog_control)

    if module_type == "shoulder_rig":
        cog_control = [x for x in cmds.listRelatives(actual_scene_name, allDescendents=True, type="transform")
                       if x.endswith("Ik_COG_Ctrl")][0]

        pelvis = [x for x in cmds.listRelatives(actual_scene_name, children=True, type="joint") if x.startswith("IK_")][0]
        ik_bones = cmds.listRelatives(pelvis, allDescendents=True, type="joint")
        chest_joint = ik_bones[0].replace("IK_", "")

        return dict(chest_joint=chest_joint,
                    cog_control=cog_control)


def get_center_joints(chain):
    """
    Finds center joint or joints
    :param chain:
    :return:
    """
    length = len(chain)
    if length % 2 != 0:
        return chain[length/2]
    else:
        return chain[length/2 - 1], chain[length/2]


def get_rotate_difference(a, b):
    """
    Gets rotational difference
    :param a:
    :param b:
    :return:
    """
    rot1 = cmds.xform(a, q=True, rotation=True, worldSpace=True)
    rot2 = cmds.xform(b, q=True, rotation=True, worldSpace=True)

    return [rot1[0]-rot2[0], rot1[1]-rot2[1], rot1[2]-rot2[2]]


def ikfk_match(module_grp, ik_to_fk=True):
    """
    Matches ik to fk and fk to ik
    :param module_grp:
    :param ik_to_fk:
    :return:
    """
    fk_main_joint = [x for x in cmds.listRelatives(module_grp, children=True) if x.startswith("FK")][0]
    fk_joints = cmds.listRelatives(fk_main_joint, allDescendents=True, type="joint") + [fk_main_joint]
    ik_joints = [x.replace("FK_", "IK_") for x in fk_joints]

    fk_controls = []
    for fk_joint in fk_joints:
        cns = cmds.listConnections(fk_joint + ".translateX")[0]
        fk_ctrl = cmds.parentConstraint(cns, q=True, targetList=True)[0]
        fk_controls.append(fk_ctrl)

    chest_ik_ctrl = [x for x in cmds.listRelatives(module_grp, allDescendents=True, type="transform")
                     if x.endswith("Chest_Ik_Ctrl")][0]
    hips_ik_ctrl = [x for x in cmds.listRelatives(module_grp, allDescendents=True, type="transform")
                    if x.endswith("Hip_Ik_Ctrl")][0]
    abdomen_ik_ctrl = [x for x in cmds.listRelatives(module_grp, allDescendents=True, type="transform")
                       if x.endswith("Abdomen_Ik_Ctrl")][0]

    if ik_to_fk:
        trans_mod.zero_out(chest_ik_ctrl)
        trans_mod.zero_out(hips_ik_ctrl)
        trans_mod.zero_out(abdomen_ik_ctrl)

        chest_trs = cmds.createNode("transform")
        chest_trs_child = cmds.createNode("transform")
        cmds.parent(chest_trs_child, chest_trs)
        cmds.matchTransform(chest_trs, ik_joints[0])
        cmds.matchTransform(chest_trs_child, chest_ik_ctrl)

        hips_trs = cmds.createNode("transform")
        hips_trs_child = cmds.createNode("transform")
        cmds.parent(hips_trs_child, hips_trs)
        cmds.matchTransform(hips_trs, ik_joints[-1])
        cmds.matchTransform(hips_trs_child, hips_ik_ctrl)

        cmds.matchTransform(chest_trs, fk_joints[0])
        cmds.matchTransform(hips_trs, fk_joints[-1])

        cmds.matchTransform(chest_ik_ctrl, chest_trs_child)
        cmds.matchTransform(hips_ik_ctrl, hips_trs_child)

        abd_trs = cmds.createNode("transform")
        abd_trs_child = cmds.createNode("transform")
        cmds.parent(abd_trs_child, abd_trs)
        cns = cmds.parentConstraint(get_center_joints(ik_joints), abd_trs)[0]
        cmds.setAttr(cns + ".interpType", 2)
        cmds.refresh()
        cmds.delete(cns)

        cmds.matchTransform(abd_trs_child, abdomen_ik_ctrl)

        center_jnts = get_center_joints(fk_joints)
        trs = cmds.createNode("transform")
        cns = cmds.parentConstraint(center_jnts, trs)[0]
        cmds.setAttr(cns + ".interpType", 2)
        cmds.refresh()
        cmds.delete(cns)

        cmds.matchTransform(abdomen_ik_ctrl, abd_trs_child)

        cmds.delete(chest_trs, hips_trs, abd_trs, trs)
    else:
        for fk_control, ik_joint in zip(reversed(fk_controls), reversed(ik_joints)):
            cmds.matchTransform(fk_control, ik_joint)
