import maya.cmds as cmds
from rigging_framework.templates.common.utilities import transform as trans_mod
from rigging_framework.templates.common.utilities import rig_controls


def build(name, cog_ctrl, chest_joint, shoulder_joint, ik_arm_joint, arm_fk_ctrl_grp, scale):
    """
    Builds a shoulder rig
    
    :param str name: Name of the module
    :param str cog_ctrl: Name of the COG control
    :param str chest_joint: Chest joint
    :param str shoulder_joint: Shoulder joint
    :param ik_arm_joint: IK arm joint (assumes we have IK chain and FK chain)
    :param str arm_fk_ctrl_grp: Arm FK control group
    :param float scale: Scale of rig controls
    
    :return str: Shoulder control
    """

    shoulder_joint_position = cmds.xform(shoulder_joint, query=True, translation=True, worldSpace=True)
    shoulder_joint_rotation = cmds.xform(shoulder_joint, query=True, rotation=True, worldSpace=True)

    component_rotation = (0, 180+45, 0)
    if "right" in shoulder_joint.lower() or shoulder_joint.lower().endswith("_r"):
        component_rotation = (0, 0+45, 0)

    shoulder_control = rig_controls.make_control("clavicle", size=scale, name="{0}_Main_Ctrl".format(name),
                                                 color="blue", rotate_comps=component_rotation,
                                                 position=shoulder_joint_position, rotation=shoulder_joint_rotation)
    shoulder_control_group = trans_mod.make_space_for(shoulder_control)
    cmds.parent(shoulder_control_group, cog_ctrl)

    # constraining in the end due to maya bug:
    cmds.parentConstraint(chest_joint, shoulder_control_group, maintainOffset=True)

    # now we need to hook up joints to the hip controls:
    cmds.parentConstraint(shoulder_control, shoulder_joint)
    cmds.parentConstraint(shoulder_control, ik_arm_joint, maintainOffset=True)

    # now we parent hip control fk group to the hip control
    cmds.parent(arm_fk_ctrl_grp, shoulder_control)

    cmds.parentConstraint(shoulder_control, shoulder_joint, maintainOffset=True)
    
    return shoulder_control


def gather_data_from_placer(placer_transform):
    """
    Gathers data from placers
    :param placer_transform:
    :return:
    """
    node = [x for x in cmds.listRelatives(placer_transform, children=True, fullPath=True)
            if x.split("|")[-1].startswith("Name__")][0]
    module_name = node.split("__")[-1]

    # get connection info:
    plugged_dict = {
        "IKFK_Arm": None,
        "IKFK_Spine": None,
    }

    for x in cmds.listRelatives(placer_transform, children=True, fullPath=True):
        if x.endswith("Plug") and "Input" in x:
            cns = cmds.listConnections(x + ".tx", type="pointConstraint")[0]
            output = cmds.pointConstraint(cns, q=True, targetList=True)[0]

            plugged_placer = cmds.listRelatives(output, parent=True)[0]
            if plugged_placer != placer_transform:
                node = [z for z in cmds.listRelatives(plugged_placer, children=True, fullPath=True)
                        if z.split("|")[-1].startswith("Name__")][0]
                plugged_name = node.split("__")[-1]
                for key in plugged_dict.keys():
                    if key in x:
                        plugged_dict[key] = [plugged_name, output.split("|")[-1]]

    # placers:
    placers = {}
    for placer_str in ["Shoulder"]:
        node = [x for x in cmds.listRelatives(placer_transform, children=True, fullPath=True)
                if x.split("|")[-1].startswith(placer_str)][0]
        cns = node + "|" + cmds.listConnections(node + ".tx", type="pointConstraint")[0]
        placers[placer_str] = cmds.pointConstraint(cns, q=True, targetList=True)[0]

    return dict(module_name=module_name,
                shoulder_joint=placers["Shoulder"],
                input_info_ikfk_arm=dict(module=plugged_dict["IKFK_Arm"][0], plug=plugged_dict["IKFK_Arm"][1]),
                input_info_ikfk_spine=dict(module=plugged_dict["IKFK_Spine"][0], plug=plugged_dict["IKFK_Spine"][1]))


def set_data_to_placer(placer_transform, placer_data):
    """
    Sets data to placers
    :param placer_transform:
    :param placer_data:
    :return:
    """
    node = [x for x in cmds.listRelatives(placer_transform, children=True, fullPath=True)
            if x.split("|")[-1].startswith("Name__")][0]
    cmds.rename(node, "Name__{}".format(placer_data["module_name"]))

    cmds.pointConstraint(placer_data["shoulder_joint"], placer_transform + "|Shoulder_Placer_Joint")


def load_placer_connection(placer_transform, placer_data):
    """
    Load placer connections
    :param placer_transform:
    :param placer_data:
    :return:
    """
    # set connection info:
    for input_info in ["input_info_ikfk_arm", "input_info_ikfk_spine"]:
        if placer_data[input_info]:
            connected_module = [x for x in cmds.ls(assemblies=True) if "Name__{}".format(placer_data[input_info]["module"])
                                in cmds.listRelatives(x, children=True)][0]

            for x in cmds.listRelatives(placer_transform, children=True, fullPath=True):
                if x.endswith("Plug") and "Input" in x and input_info.replace("input_info_", "") in x.lower():

                    constraint = cmds.listConnections(x, type="constraint", source=True, destination=False)
                    if constraint:
                        cmds.delete(constraint)

                    cmds.pointConstraint(connected_module + "|" + placer_data[input_info]["plug"], x)
