import os
import maya.cmds as cmds

import rigging_framework.templates.common.utilities.transform
from rigging_framework import global_variables as gvar
from rigging_framework.templates.common.utilities import utils as utils
from rigging_framework.templates.common.utilities import constraints


def load_angle_axis(prefix):
    """
    loads angle axis asset
    :param prefix:
    :return:
    """
    path = os.path.join(gvar.get("common_data_folder"), 'angle_axis.ma')
    cmds.warning("path={}".format(path))
    utils.import_and_add_prefix(path, prefix)
    return prefix+"_Angle_Axis_Grp"


def hook_up_angle_axis(angle_axis_group, parent_space, parent_joint, aim_joint):
    """
    hooks up angle axis and returns the result locator
    :param angle_axis_group:
    :param parent_space:
    :param parent_joint:
    :param aim_joint:
    :return:
    """
    angle_axis_root = angle_axis_group.replace("Grp", "Root")
    angle_axis_dir = angle_axis_group.replace("Grp", "Direction")
    angle_axis_joint = angle_axis_group.replace("Grp", "Joint")

    cmds.delete(cmds.parentConstraint(parent_joint, angle_axis_group))

    # needs to skip translate (rotation space locked)
    cmds.parentConstraint(parent_space, angle_axis_group, maintainOffset=True, skipTranslate=["x", "y", "z"])

    # needs to be point constraint to the parent joint... (translation locked to the shoulder)
    cmds.pointConstraint(parent_joint, angle_axis_group, maintainOffset=True)

    cmds.pointConstraint(parent_joint, angle_axis_root)
    cmds.delete(cmds.pointConstraint(aim_joint, angle_axis_dir))
    cmds.pointConstraint(aim_joint, angle_axis_joint)

    return angle_axis_group.replace("Grp", "Result")


def build_twisters():
    """
    builds twisters for upper parts of the leg:
    :return:
    """
    do_not_touch_grp = gvar.get("Do_Not_Touch_Grp")
    global_scale = gvar.get("global_scale")

    grp_twist_systems = cmds.createNode("transform", name="Twist_Systems_Grp")
    cmds.parent(grp_twist_systems, do_not_touch_grp)

    for side in "Left", "Right":
        hip_jnt = "{0}_Hip_Jnt".format(side)
        parent_space = cmds.listRelatives(hip_jnt, parent=True)[0]
        upper_knee_jnt = cmds.listRelatives(hip_jnt, children=True)[0]

        # create name and align twister joints:
        twister_jnt_01 = cmds.createNode("joint", name=hip_jnt.replace("Jnt", "Twister_01_Jnt"))
        twister_jnt_02 = cmds.createNode("joint", name=hip_jnt.replace("Jnt", "Twister_02_Jnt"))

        cmds.setAttr(twister_jnt_01 + ".radius", cmds.getAttr(hip_jnt + ".radius") * 1.5)
        cmds.setAttr(twister_jnt_02 + ".radius", cmds.getAttr(hip_jnt + ".radius") * 1.5)
        cmds.setAttr(twister_jnt_01 + ".displayLocalAxis", True)
        cmds.setAttr(twister_jnt_02 + ".displayLocalAxis", True)

        rigging_framework.templates.common.utilities.transform.parent_and_zero(twister_jnt_01, hip_jnt, joint_orient=True)
        rigging_framework.templates.common.utilities.transform.parent_and_zero(twister_jnt_02, hip_jnt, joint_orient=True)

        cmds.delete(cmds.parentConstraint(hip_jnt, twister_jnt_01))
        cmds.delete(cmds.parentConstraint(hip_jnt, twister_jnt_02))
        cmds.delete(cmds.pointConstraint(hip_jnt, upper_knee_jnt, twister_jnt_02))

        notwist_angle_axis_grp = load_angle_axis(hip_jnt.replace("_Jnt", "_NoTwist"))
        cmds.parent(notwist_angle_axis_grp, grp_twist_systems)  # create and organize them in a group

        # snap angle axis with the joint and parent constraint it with the hip joint:
        notwist_angle_axis_result = hook_up_angle_axis(notwist_angle_axis_grp, parent_space, hip_jnt, upper_knee_jnt)

        twist_angle_axis_grp = load_angle_axis(hip_jnt.replace("_Jnt", "_Twist"))
        cmds.parent(twist_angle_axis_grp, grp_twist_systems)  # create and organize them in a group

        # snap angle axis with the joint and parent constraint it with the hip joint:
        twist_angle_axis_result = hook_up_angle_axis(twist_angle_axis_grp, hip_jnt, hip_jnt, upper_knee_jnt)

        # map the twister joints:
        constraints.orient_constraint(notwist_angle_axis_result, twister_jnt_01, maintain_offset=True,
                                      use_mapping_file=True)

        constraints.orient_constraint([notwist_angle_axis_result, twist_angle_axis_result], twister_jnt_02,
                                      maintain_offset=True, interp_type=2, use_mapping_file=True)

        # make the twisters stretchy:
        constraints.point_constraint(hip_jnt, twister_jnt_01, use_mapping_file=True)
        constraints.point_constraint([hip_jnt, upper_knee_jnt], twister_jnt_02, use_mapping_file=True)


build_twisters()
