import os

import maya.api.OpenMaya as om
import maya.cmds as cmds
from rigging_framework.templates.common.utilities import utils
from rigging_framework.templates.common.utilities import colors
from rigging_framework.templates.common.utilities import joints
from rigging_framework.templates.common.utilities import rig_controls
from rigging_framework.templates.common.utilities import driven_keys
from rigging_framework.templates.common.modules import fk_chain
import rigging_framework.templates.common.utilities.transform as trans_mod


def build(name, shoulder_joint, upper_knee_joint, lower_knee_joint, ankle_joint, scale=1.0, orientation=(0, 0, 0)):
    """
    Builds the front IKFK quadrupedal leg based on given parameters
    
    :param str name: Name of the module
    :param str shoulder_joint: Name of the shoulder bone
    :param str upper_knee_joint: Name of the upper knee joint (assumes quadruped would have upper and lower knee)
    :param lower_knee_joint: Name of the lower knee joint (assumes quaruped would have upper and lower knee)
    :param joint ankle_joint: Name of the ankle joint
    :param float scale: Scale of the controls
    :param float[3] orientation: Orientation of the system (default 0, 0, 0)
    
    :return dict: 
        result = {"Main_Node": Main node - a locator to which you can attach footrig and other elements, follows IK and FK
                  "Controls": List of controls - lower knee pv, ik shoulder, fk controls
                  "Do_Not_Touch_Grp": IK system group
                  "Gizmos_Grp": Group containing visual gizmos - guides
                  "Controls_Grp": Main controls group
                  "IKFK_Switch_Ctrl": ikfk_switch_ctrl}
    """

    # create duplicate chain for FK
    fk_joint_chain = utils.duplicate_unparent(shoulder_joint, depth=4, prefix='FK')[0]
    colors.set_color_from_str(fk_joint_chain, "blue")

    fk_hip_joint = "FK_" + shoulder_joint
    fk_upper_knee_joint = "FK_" + upper_knee_joint
    fk_lower_knee_joint = "FK_" + lower_knee_joint
    fk_ankle_joint = "FK_" + ankle_joint

    fk_joints_grp = utils.make_empty_group("{0}_FK_Joints_Grp".format(name))
    cmds.parent(fk_hip_joint, fk_joints_grp)

    # build FK chain
    fk_data = fk_chain.build([fk_hip_joint, fk_upper_knee_joint, fk_lower_knee_joint, fk_ankle_joint],
                             name_or_none=None, replace_in_name=[("_Jnt", ""), ("FK_", "")], add_suffix="_FK_Ctrl",
                             scale=scale, scale_constraint=False)

    fk_controls = fk_data["Controls"]
    fk_control_grp = fk_data["Module_Group"]

    # create duplicate chain for IK
    ik_joint_chain = utils.duplicate_unparent(shoulder_joint, depth=4, prefix='IK')[0]
    colors.set_color_from_str(ik_joint_chain, "red")

    ik_shoulder_joint = "IK_" + shoulder_joint
    ik_upper_knee_joint = "IK_" + upper_knee_joint
    ik_lower_knee_joint = "IK_" + lower_knee_joint
    ik_ankle_joint = "IK_" + ankle_joint

    ik_joints_grp = utils.make_empty_group("{0}_IK_Joints_Grp".format(name))
    cmds.parent(ik_shoulder_joint, ik_joints_grp)

    # build upper and lower ik handles:
    upper_ik = cmds.ikHandle(startJoint=ik_shoulder_joint, endEffector=ik_upper_knee_joint, sol="ikSCsolver",
                             name=name+"_Upper_IKHandle")[0]

    lower_ik = cmds.ikHandle(startJoint=ik_upper_knee_joint, endEffector=ik_ankle_joint, sol="ikRPsolver",
                             name=name+"_Lower_IKHandle")[0]

    # build foot offset transform
    foot_offset_transform = cmds.createNode("transform", name="{0}_IK_Foot_Driver_Grp".format(name))
    utils.match_sec_to_first(lower_ik, foot_offset_transform)
    foot_offset_transform_grp = trans_mod.make_space_for(foot_offset_transform)
    cmds.pointConstraint(foot_offset_transform, lower_ik)

    # build ik foot locator
    foot_locator = cmds.spaceLocator(name=name+"_IK_Foot_Loc")[0]
    utils.match_sec_to_first(ankle_joint, foot_locator, rotation=False)

    # add auto shoulder attribute
    cmds.addAttr(foot_locator, longName="automatic_shoulder", attributeType="float", minValue=0, maxValue=1,
                 defaultValue=0.5, keyable=True)

    # build auto shoulder system:
    auto_shoulder_transform = cmds.createNode("transform", name="{}_AutoShoulder_System_Grp".format(name))
    utils.match_sec_to_first(upper_ik, auto_shoulder_transform, rotation=False)
    auto_shoulder_transform_grp = trans_mod.make_space_for(auto_shoulder_transform)
    cmds.pointConstraint(auto_shoulder_transform, upper_ik)

    md_auto_shoulder = cmds.createNode("multiplyDivide")
    cmds.connectAttr(foot_offset_transform + ".t", md_auto_shoulder + ".input1")

    # create driven curve so that it works better (like set range):
    auto_shoulder_curve = driven_keys.build_quick_curve("{}_auto_shoulder_crv".format(name), ((0, 0), (1, 0.4)))
    cmds.connectAttr(foot_locator + ".automatic_shoulder", auto_shoulder_curve + ".input")

    for axis in "XYZ":
        cmds.connectAttr(auto_shoulder_curve + ".output", md_auto_shoulder + ".input2{}".format(axis))

    cmds.connectAttr(md_auto_shoulder + ".output", auto_shoulder_transform + ".t")

    # now let's try to add stretchiness to the ik

    # add enums for the foot locator
    cmds.addAttr(foot_locator, longName="stretch", attributeType="enum", enumName="Off:On:SquashMode", keyable=True)

    # add global scale for the foot locator...
    cmds.addAttr(foot_locator, longName="global_scale", attributeType="float", minValue=0,
                 defaultValue=1, keyable=True)

    # add soft ik the foot locator...
    cmds.addAttr(foot_locator, longName="soft_ik", attributeType="float", minValue=0,
                 defaultValue=0, maxValue=50, keyable=True)

    # build measuring curves, default, current, and for 3 bones - up, mid, low:
    def_crv = utils.build_linear_curve(name="{}_def_len_crv".format(ik_shoulder_joint))
    cur_crv = utils.build_linear_curve(name="{}_cur_len_crv".format(ik_shoulder_joint))
    up_crv = utils.build_linear_curve(name="{}_len_crv".format(ik_upper_knee_joint))
    mid_crv = utils.build_linear_curve(name="{}_len_crv".format(ik_lower_knee_joint))
    low_crv = utils.build_linear_curve(name="{}_len_crv".format(ik_ankle_joint))

    # snap those curve cvs to right joint using clusters:
    cvs = cmds.ls('%s.cp[*]' % def_crv, fl=True)
    cmds.xform(cvs[0], translation=cmds.xform(ik_shoulder_joint, q=True, rp=True, ws=True), ws=True)
    cmds.xform(cvs[1], translation=cmds.xform(ik_ankle_joint, q=True, rp=True, ws=True), ws=True)

    cvs = cmds.ls('%s.cp[*]' % cur_crv, fl=True)
    clu1_handle_def = cmds.cluster(cvs[0], name="{}_start_clu".format(cur_crv))[1]
    clu2_handle_def = cmds.cluster(cvs[1], name="{}_end_clu".format(cur_crv))[1]
    cmds.pointConstraint(ik_shoulder_joint, clu1_handle_def)
    cmds.pointConstraint(foot_locator, clu2_handle_def)

    cvs = cmds.ls('%s.cp[*]' % up_crv, fl=True)
    clu1_handle_up = cmds.cluster(cvs[0], name="{}_start_clu".format(up_crv))[1]
    clu2_handle_up = cmds.cluster(cvs[1], name="{}_end_clu".format(up_crv))[1]
    cmds.delete(cmds.pointConstraint(ik_shoulder_joint, clu1_handle_up))
    cmds.delete(cmds.pointConstraint(ik_upper_knee_joint, clu2_handle_up))

    cvs = cmds.ls('%s.cp[*]' % mid_crv, fl=True)
    clu1_handle_mid = cmds.cluster(cvs[0], name="{}_start_clu".format(mid_crv))[1]
    clu2_handle_mid = cmds.cluster(cvs[1], name="{}_end_clu".format(mid_crv))[1]
    cmds.delete(cmds.pointConstraint(ik_upper_knee_joint, clu1_handle_mid))
    cmds.delete(cmds.pointConstraint(ik_lower_knee_joint, clu2_handle_mid))

    cvs = cmds.ls('%s.cp[*]' % low_crv, fl=True)
    clu1_handle_low = cmds.cluster(cvs[0], name="{}_start_clu".format(low_crv))[1]
    clu2_handle_low = cmds.cluster(cvs[1], name="{}_end_clu".format(low_crv))[1]
    cmds.delete(cmds.pointConstraint(ik_lower_knee_joint, clu1_handle_low))
    cmds.delete(cmds.pointConstraint(ik_ankle_joint, clu2_handle_low))

    # import ik system
    common_data_folder = os.path.dirname(os.path.realpath(__file__)).replace("modules", "data")
    # common_data_folder = gvar.get("common_data_folder")
    path = os.path.join(common_data_folder, 'ik_system_3.ma')
    utils.import_and_add_prefix(path, name)
    soft_ik_sys = name + "_soft_ik_system"

    # hookup measuring curves to curve infos
    cmds.connectAttr(cur_crv+".worldSpace[0]", name + "_limb_cur_len_crvInfo.inputCurve")
    cmds.connectAttr(def_crv+".worldSpace[0]", name + "_limb_def_len_crvInfo.inputCurve")
    cmds.connectAttr(up_crv+".worldSpace[0]", name + "_limb_up_len_crvInfo.inputCurve")
    cmds.connectAttr(mid_crv+".worldSpace[0]", name + "_limb_mid_len_crvInfo.inputCurve")
    cmds.connectAttr(low_crv+".worldSpace[0]", name + "_limb_low_len_crvInfo.inputCurve")

    # constraint lower ik between hip joint and foot locator
    cns = cmds.pointConstraint(foot_locator, foot_offset_transform)[0]
    cmds.pointConstraint(shoulder_joint, foot_offset_transform)

    # connect the constraint to soft ik system
    cmds.connectAttr(soft_ik_sys + ".ik_adjust_percent_rev", cns + "." + foot_locator + "W0")
    cmds.connectAttr(soft_ik_sys + ".ik_adjust_percent", cns + "." + shoulder_joint + "W1")

    def connect_joint_length(incoming_connection, joint_input):
        """
        connects joint length but if its negative then reverses it
        """
        if cmds.getAttr(joint_input) > 0:
            cmds.connectAttr(incoming_connection, joint_input)
        else:
            rev_md = cmds.createNode("multiplyDivide", name="md_reverse_{}_length".format(joint_input.split('.')[0]))
            cmds.setAttr(rev_md+".input2", -1, -1, -1)
            cmds.connectAttr(incoming_connection, rev_md+".input1X")
            cmds.connectAttr(rev_md+".outputX", joint_input)

    # connect soft ik outputs for 3 chains to their respective joints lengths
    connect_joint_length(name + "_limb_up_ik_length_md.outputX", ik_upper_knee_joint+joints.get_jnt_trans(ik_upper_knee_joint))
    connect_joint_length(name + "_limb_mid_ik_length_md.outputX", ik_lower_knee_joint+joints.get_jnt_trans(ik_lower_knee_joint))
    connect_joint_length(name + "_limb_low_ik_length_md.outputX", ik_ankle_joint + joints.get_jnt_trans(ik_ankle_joint))

    # hook up some more attributes to soft ik system
    cmds.connectAttr(foot_locator + ".stretch", name + "_soft_ik_core_squashing_mode_crv.input")
    cmds.connectAttr(foot_locator + ".stretch", name + "_soft_ik_core_stretchy_ik_crv.input")
    cmds.connectAttr(foot_locator + ".soft_ik", soft_ik_sys + ".soft_ik")
    cmds.connectAttr(foot_locator + ".global_scale", soft_ik_sys + ".soft_ik_scale")

    # organize common elements into scaling and non_scaling
    ik_system_grp = utils.make_empty_group("{}_IK_System_Grp".format(name))
    scaling_grp = utils.make_empty_group("{}_IK_Scaling_Grp".format(name), ik_system_grp)
    cmds.parent(def_crv, scaling_grp)
    cmds.parent(up_crv, mid_crv, low_crv, ik_system_grp)
    cmds.parent(soft_ik_sys, ik_system_grp)
    cmds.parent(cur_crv, ik_system_grp)
    cmds.parent(clu1_handle_def, clu1_handle_low, clu1_handle_mid, clu1_handle_up, ik_system_grp)
    cmds.parent(clu2_handle_def, clu2_handle_low, clu2_handle_mid, clu2_handle_up, ik_system_grp)
    cmds.parent(foot_offset_transform_grp, scaling_grp)  # this is essential otherwise rig wont scale
    cmds.parent(lower_ik, scaling_grp)
    cmds.parent(foot_locator, ik_system_grp)

    for axis in "xyz":
        cmds.connectAttr(foot_locator + ".global_scale", scaling_grp + ".s{}".format(axis))

    # in stretch mode we want default position of upper IK to stretch as well
    auto_shoulder_transform_grp_offset = utils.make_empty_group(auto_shoulder_transform.replace("Grp", "Offset_Grp"))
    utils.match_sec_to_first(ik_shoulder_joint, auto_shoulder_transform_grp_offset, rotation=False)
    cmds.parent(auto_shoulder_transform_grp, auto_shoulder_transform_grp_offset)
    cmds.parent(upper_ik, auto_shoulder_transform_grp_offset)

    cmds.parent(auto_shoulder_transform_grp_offset, scaling_grp)

    for axis in "xyz":
        cmds.connectAttr(soft_ik_sys + ".final_ik_ratio", auto_shoulder_transform_grp_offset + ".s{}".format(axis))

    # build ik shoulder controls
    shoulder_ik_control = rig_controls.make_control("clavicle", name="{0}_Shoulder_IK_Ctrl".format(name), size=scale*2,
                                                    color="red", rotation=orientation, rotate_comps=(90, 0, 0))
    hip_control_grp = trans_mod.make_space_for(shoulder_ik_control)
    utils.match_sec_to_first(ik_shoulder_joint, hip_control_grp, rotation=False)

    # now constraint stuff to shoulder control:
    cmds.pointConstraint(shoulder_ik_control, ik_shoulder_joint)
    
    # todo: investigate rotating the shoulder control issue:
    cmds.pointConstraint(shoulder_ik_control, auto_shoulder_transform_grp_offset, maintainOffset=True)
    cmds.pointConstraint(shoulder_ik_control, shoulder_joint)
    cmds.pointConstraint(shoulder_ik_control, foot_offset_transform_grp, maintainOffset=True)

    # hook up global scale
    groups = [ik_joints_grp, fk_joints_grp]
    for group in groups:
        for axis in "xyz":
            cmds.connectAttr(foot_locator + ".global_scale", "{0}.s{1}".format(group, axis))
        cmds.parent(group, ik_system_grp)

    # pole vector for lower knee
    lower_knee_pole_vec_ctrl = rig_controls.make_control("box", name="{0}_Lower_Knee_IK_Ctrl".format(name),
                                                         size=scale * 0.5, color="red", rotation=orientation)
    lower_knee_pole_vec_ctrl_grp = trans_mod.make_space_for(lower_knee_pole_vec_ctrl)
    cmds.delete(cmds.pointConstraint(ik_upper_knee_joint, ik_ankle_joint, lower_knee_pole_vec_ctrl_grp))

    limb_length = cmds.getAttr(name + "_limb_mid_ik_length_md.outputX") + \
                  cmds.getAttr(name + "_limb_low_ik_length_md.outputX")

    point1 = cmds.xform(ik_lower_knee_joint, query=True, rotatePivot=True, worldSpace=True)
    point2 = cmds.getAttr(lower_knee_pole_vec_ctrl_grp + ".t")[0]
    vector = om.MVector(point2) - om.MVector(point1)
    vector = om.MVector.normalize(vector)
    vector *= limb_length * -0.5
    cmds.move(vector.x, vector.y, vector.z, lower_knee_pole_vec_ctrl_grp, relative=True)
    cmds.poleVectorConstraint(lower_knee_pole_vec_ctrl, lower_ik)

    lower_ik_guide = utils.build_linear_curve(name="{}_Lower_Knee_Guide_Crv".format(name))
    cmds.setAttr(lower_ik_guide + ".overrideEnabled", True)
    cmds.setAttr(lower_ik_guide + ".overrideDisplayType", 1)

    # snap those curve cvs to right joint using clusters:
    cvs = cmds.ls('%s.cp[*]' % lower_ik_guide, fl=True)
    cmds.xform(cvs[0], translation=cmds.xform(ik_lower_knee_joint, q=True, rp=True, ws=True), ws=True)
    cmds.xform(cvs[1], translation=cmds.xform(lower_knee_pole_vec_ctrl, q=True, rp=True, ws=True), ws=True)

    clu1_lower_handle_def = cmds.cluster(cvs[0], name="{}_start_clu".format(lower_ik_guide))[1]
    clu2_lower_handle_def = cmds.cluster(cvs[1], name="{}_end_clu".format(lower_ik_guide))[1]
    cmds.pointConstraint(ik_lower_knee_joint, clu1_lower_handle_def)
    cmds.pointConstraint(lower_knee_pole_vec_ctrl, clu2_lower_handle_def)

    # clean up those guide curves:
    cmds.parent(clu1_lower_handle_def, clu2_lower_handle_def,
                ik_system_grp)

    gizmo_group = utils.make_empty_group("{}_Gizmo_Grp".format(name))
    cmds.parent(lower_ik_guide, gizmo_group)

    # parent fk chain to hip control
    cmds.parent(fk_control_grp, shoulder_ik_control)

    # now build ikfk switching and visibility switching
    ikfk_switch_ctrl = rig_controls.make_control("diamond", name="{0}_IKFK_Switch_Ctrl".format(name),
                                                 size=scale * 0.2, color="red", rotation=orientation,
                                                 rotate_comps=(90, 0, 0))

    ikfk_switch_ctrl_grp = trans_mod.make_space_for(ikfk_switch_ctrl)
    utils.match_sec_to_first(ankle_joint, ikfk_switch_ctrl_grp, rotation=False)
    cmds.pointConstraint(ankle_joint, ikfk_switch_ctrl_grp)
    cmds.addAttr(ikfk_switch_ctrl, longName="ikfk_blend", minValue=0, maxValue=1, keyable=True)

    fk_vis_crv = driven_keys.build_quick_curve("{0}_FK_Visibility_Curve".format(name), [(0, 0), (1, 1)])
    cmds.connectAttr(ikfk_switch_ctrl + ".ikfk_blend", fk_vis_crv + ".input")
    cmds.connectAttr(fk_vis_crv + ".output", fk_control_grp + ".visibility")


    ik_vis_crv = driven_keys.build_quick_curve("{0}_IK_Visibility_Curve".format(name), [(0, 1), (1, 0)])
    cmds.connectAttr(ikfk_switch_ctrl + ".ikfk_blend", ik_vis_crv + ".input")
    cmds.connectAttr(ik_vis_crv + ".output", lower_knee_pole_vec_ctrl_grp + ".visibility")
    cmds.connectAttr(ik_vis_crv + ".output", gizmo_group + ".visibility")
    cmds.connectAttr(ik_vis_crv + ".output", foot_locator + ".visibility")

    # build blendColors or pairBlends to blend ik and fk...
    for i, joint in enumerate([shoulder_joint, upper_knee_joint, lower_knee_joint, ankle_joint]):
        fk_joint = "FK_" + joint
        ik_joint = "IK_" + joint
        pair_blend = cmds.createNode("pairBlend")
        blend_colors = cmds.createNode("blendColors")
        cmds.connectAttr(ik_joint + ".translate", pair_blend + ".inTranslate1")
        cmds.connectAttr(fk_joint + ".translate", pair_blend + ".inTranslate2")
        cmds.connectAttr(ik_joint + ".rotate", pair_blend + ".inRotate1")
        cmds.connectAttr(fk_joint + ".rotate", pair_blend + ".inRotate2")
        cmds.connectAttr(ik_joint + ".scale", blend_colors + ".color2")
        cmds.connectAttr(fk_joint + ".scale", blend_colors + ".color1")
        cmds.connectAttr(ikfk_switch_ctrl + ".ikfk_blend", pair_blend + ".weight")
        cmds.connectAttr(ikfk_switch_ctrl + ".ikfk_blend", blend_colors + ".blender")

        # we don't want to plug anything to the first joint translation or rotation as its in a different space
        if i != 0:
            cmds.connectAttr(pair_blend + ".outTranslate", joint + ".translate")
            cmds.connectAttr(pair_blend + ".outRotate", joint + ".rotate")
        else:
            # for top joint we use orientConstraint otherwise we will have issues with skeleton not locked in rotate
            orient_cns = cmds.orientConstraint(fk_joint, ik_joint, joint, maintainOffset=True)[0]
            cmds.connectAttr(fk_vis_crv + ".output", "{0}.{1}W0".format(orient_cns, fk_joint))
            cmds.connectAttr(ik_vis_crv + ".output", "{0}.{1}W1".format(orient_cns, ik_joint))

        cmds.connectAttr(blend_colors + ".output", joint + ".scale")

    # add ik mode attribute:
    cmds.addAttr(ikfk_switch_ctrl, attributeType="enum", longName="ik_mode",
                 enumName="Regular:Stretchy:Squash", keyable=True)
    cmds.connectAttr(ikfk_switch_ctrl + ".ik_mode", foot_locator + ".stretch")

    # clean up - create system control group
    control_group = utils.make_empty_group("{}_Ctrl_Grp".format(name))
    cmds.parent(hip_control_grp, control_group)
    cmds.parent(lower_knee_pole_vec_ctrl_grp, control_group)
    cmds.parent(ikfk_switch_ctrl_grp, control_group)

    result = {"Main_Node": foot_locator,
              "Controls": [lower_knee_pole_vec_ctrl, shoulder_ik_control] + fk_controls,
              "Do_Not_Touch_Grp": ik_system_grp,
              "Gizmos_Grp": gizmo_group,
              "Controls_Grp": control_group,
              "IKFK_Switch_Ctrl": ikfk_switch_ctrl}

    return result
