import maya.cmds as cmds

import rigging_framework.templates.common.utilities.transform as trans_mod
from rigging_framework.templates.common.utilities import utils
from rigging_framework.templates.common.utilities import colors
from rigging_framework.templates.common.utilities import rig_controls
from rigging_framework.templates.common.utilities import constraints as cnst
from rigging_framework.templates.common.modules import fk_chain
from rigging_framework.templates.common.utilities import driven_keys
from rigging_framework.templates.quadruped_rig.modules import quad_ik_ribbon_spine

import imp
imp.reload(quad_ik_ribbon_spine)


def build(name, tail_joints, orientation=(0, 0, 0), scale=1.0, virtual_constraints=False):
    """
    Builds ikfk tail
    
    :param str name: Module name
    :param list[str] tail_joints: List of tail joints
    :param float[3] orientation: orientation of the rig (default 0 0 0)
    :param float scale: scale of the controls
    :param virtual_constraints: whether or not constraining will be executed in this script or saved for later
    
    :return dict:
        result = {"Main_Node": main_tail_ctrl,
                  "Controls": [main_tail_ctrl] + fk_tail_data["Controls"] + ik_tail_data["Controls"],
                  "Do_Not_Touch_Grp": do_not_touch_grp,
                  "Scaling_Grp": scaling_jnt_grp,
                  "Controls_Grp": all_controls_grp}
    """
    
    if not virtual_constraints:
        cnst.initialize_constraints()

    # create duplicate chain for FK
    fk_tail_joints = utils.duplicate_unparent(tail_joints[0], depth=len(tail_joints), prefix='FK',
                                              favor=tail_joints[1])
    for fk_joint in fk_tail_joints:
        colors.set_color_from_str(fk_joint, "blue")

    # create duplicate chain for IK
    ik_tail_joints = utils.duplicate_unparent(tail_joints[0], depth=len(tail_joints), prefix='IK',
                                              favor=tail_joints[1])
    for ik_joint in ik_tail_joints:
        colors.set_color_from_str(ik_joint, "red")

    # build fk chain for the tail
    fk_tail_data = fk_chain.build(fk_tail_joints, name_or_none=name, add_suffix="_Fk_Ctrl", scale=scale,
                                  scale_constraint=False, virtual_constraints=virtual_constraints)

    # make ik tail (Fix naming issues)
    ik_tail_data = quad_ik_ribbon_spine.build(name="IK_{0}".format(name), spine_joints=ik_tail_joints, scale=scale * 1.2,
                                              control_names=("Start", "Mid", "End"), main_cog_control=False,
                                              world_orientated_controls=False, middle_control_constrained=False,
                                              virtual_constraints=virtual_constraints)

    # make main tail control with driven key attributes (use main fk control)
    main_fk_control = fk_tail_data["Controls"][0]

    cmds.addAttr(main_fk_control, longName="Curl", attributeType="float", minValue=-1.0, maxValue=1.0, keyable=True)
    cmds.addAttr(main_fk_control, longName="Sway", attributeType="float", minValue=-1.0, maxValue=1.0, keyable=True)
    cmds.addAttr(main_fk_control, longName="Twist", attributeType="float", minValue=-1.0, maxValue=1.0, keyable=True)

    fk_multiplier = cmds.createNode("multiplyDivide")

    cmds.connectAttr(main_fk_control + ".Twist", fk_multiplier + ".input1X")
    cmds.connectAttr(main_fk_control + ".Sway", fk_multiplier + ".input1Y")
    cmds.connectAttr(main_fk_control + ".Curl", fk_multiplier + ".input1Z")

    cmds.setAttr(fk_multiplier + ".input2", 45, 45, 45)

    # create offset groups for all the fk controls:
    for i, fk_control in enumerate(fk_tail_data["Controls"]):
        sdk_group = cmds.createNode("transform", name=fk_control + "_SDK_Grp")
        cmds.delete(cmds.parentConstraint(fk_control, sdk_group))
        fk_control_grp = trans_mod.get_parent(fk_control)
        cmds.parent(sdk_group, fk_control_grp)
        cmds.parent(fk_control, sdk_group)
        cmds.connectAttr(fk_multiplier + ".output", sdk_group + ".rotate")

    # make main control for the tail with ikfk switch:
    main_tail_ctrl = rig_controls.make_control("clavicle", name="{0}_Main_Ctrl".format(name),
                                               position=trans_mod.get_exact_position(main_fk_control), color="yellow",
                                               rotation=orientation, rotate_comps=(90, 0, 0))

    main_tail_ctrl_grp = trans_mod.make_space_for(main_tail_ctrl)

    cmds.addAttr(main_tail_ctrl, longName="ikfk_blend", attributeType="float",
                 minValue=0.0, maxValue=1.0, keyable=True)

    cmds.parent(main_fk_control + "_Grp", main_tail_ctrl)

    # add ikfk blending for visibility:
    fk_curve = driven_keys.build_quick_curve("{0}_FK_sdk".format(name), [(0, 1), (1, 0)])
    ik_curve = driven_keys.build_quick_curve("{0}_IK_sdk".format(name), [(0, 0), (1, 1)])
    cmds.connectAttr(main_tail_ctrl + ".ikfk_blend", fk_curve + ".input")
    cmds.connectAttr(main_tail_ctrl + ".ikfk_blend", ik_curve + ".input")

    cmds.connectAttr(fk_curve + ".output", main_fk_control + "_Grp.visibility")

    for ik_control in ik_tail_data["Controls"]:
        cmds.connectAttr(ik_curve + ".output", ik_control + "_Grp.visibility")

    # store data and cleanup
    do_not_touch_grp = cmds.createNode("transform", name="{0}_Do_Not_Touch_Grp".format(name))
    all_controls_grp = cmds.createNode("transform", name="{0}_Controls_Grp".format(name))

    cmds.parent(main_tail_ctrl_grp, all_controls_grp)

    for ik_control in ik_tail_data["Controls"]:
        cmds.parent(trans_mod.get_parent(ik_control), main_tail_ctrl)

    cmds.parent(ik_tail_data["Do_Not_Touch_Grp"], do_not_touch_grp)

    # scaling joints group:
    scaling_jnt_grp = cmds.createNode("transform", name="{0}_Scaling_Jnt".format(name))
    cmds.parent(scaling_jnt_grp, ik_tail_data["Do_Not_Touch_Grp"])
    cmds.parent(fk_tail_joints[0], scaling_jnt_grp)
    cmds.parent(ik_tail_joints[0], scaling_jnt_grp)
    cmds.parent(ik_tail_data["Scaling_Groups"], scaling_jnt_grp)

    # add ikfk blending for following:
    main_parent_cns = cmds.parentConstraint(fk_tail_joints[0], tail_joints[0], maintainOffset=True)[0]
    cmds.parentConstraint(ik_tail_joints[0], tail_joints[0], maintainOffset=True)
    cmds.setAttr(main_parent_cns + ".interpType", 2)

    cmds.connectAttr(ik_curve + ".output", "{0}.{1}W1".format(main_parent_cns, ik_tail_joints[0]))
    cmds.connectAttr(fk_curve + ".output", "{0}.{1}W0".format(main_parent_cns, fk_tail_joints[0]))

    for tail_joint, fk_tail_joint, ik_tail_joint in zip(tail_joints[1:], fk_tail_joints[1:], ik_tail_joints[1:]):
        pair_blend = cmds.createNode("pairBlend")
        cmds.connectAttr(fk_curve + ".output", pair_blend + ".weight")
        cmds.connectAttr(ik_tail_joint + ".translate", pair_blend + ".inTranslate1")
        cmds.connectAttr(ik_tail_joint + ".rotate", pair_blend + ".inRotate1")
        cmds.connectAttr(fk_tail_joint + ".translate", pair_blend + ".inTranslate2")
        cmds.connectAttr(fk_tail_joint + ".rotate", pair_blend + ".inRotate2")
        cmds.connectAttr(pair_blend + ".outTranslate", tail_joint + ".translate")
        cmds.connectAttr(pair_blend + ".outRotate", tail_joint + ".rotate")

    # finalize build:

    result = {"Main_Node": main_tail_ctrl,
              "Controls": [main_tail_ctrl] + fk_tail_data["Controls"] + ik_tail_data["Controls"],
              "Do_Not_Touch_Grp": do_not_touch_grp,
              "Scaling_Grp": scaling_jnt_grp,
              "Controls_Grp": all_controls_grp}

    if not virtual_constraints:
        cnst.build_constraints()

    return result
