import maya.api.OpenMaya as om
import maya.cmds as cmds
from rigging_framework.templates.common.utilities import follicles
from rigging_framework.templates.common.utilities import joints
from rigging_framework.templates.common.utilities import rig_controls
from rigging_framework.templates.common.utilities import skin_clusters
from rigging_framework.templates.common.modules import ribbon
from rigging_framework.templates.common.utilities import constraints as cnst
import rigging_framework.templates.common.utilities.transform as trans_mod


def build(name, spine_joints, control_names=("Hip", "Abdomen", "Chest"), main_cog_control=True,
          world_orientated_controls=True, middle_control_constrained=True, scale=1.0, orientation=(0, 0, 0),
          virtual_constraints=False):

    """
    Builds ribbon spine for quadruped rig

    :param str name: Name of the module
    :param list[str] spine_joints: Joints starting from the hip, ending with the chest
    :param float scale: Scale of controls
    :param list[str] control_names: Names of 3 ik controls
    :param str main_cog_control: Whether to build main cog control
    :param bool world_orientated_controls: Whether we want the controls to be world orientated or object
    :param bool middle_control_constrained: Whether we want middle control to be constrained between outer ones
    :param float[3] orientation: Default orientation of the ik ribbon spine (0, 0, 0) default
    :param bool virtual_constraints: Whether constrains are going to be built right away or virtually

    :return:
         result = {"Main_Node": COG control - main rig node
                   "Controls": Controls - (COG), Hip, Abdomen, Chest,
                   "Do_Not_Touch_Grp": Rig elements that are supposed to be hidden,
                   "Scaling_Groups": Groups requiring individual scaling,
                   "Controls_Grp": Offset groups for controls above
    """

    if not virtual_constraints:
        cnst.initialize_constraints()

    # create main folders:
    do_not_touch_grp = cmds.createNode("transform", name="{0}_Do_Not_Touch_Grp".format(name))

    ribbon_data = ribbon.build(spine_joints, "{0}_Ribbon".format(name), attach_objects=True, sticky=False,
                               make_controls=False, scale=1)

    ribbon_sfc = ribbon_data["Ribbon_Surface"]

    # get hip and chest joint
    hip_joint = spine_joints[0]
    chest_joint = spine_joints[-1]

    # get mid spine position
    spine_mid_follicle = follicles.make_follicle(ribbon_sfc, [0.5, 0.5], "spine_mid_follicle")
    mid_spine_position = cmds.xform(spine_mid_follicle, query=True, translation=True, worldSpace=True)
    cmds.delete(spine_mid_follicle)

    # create three big ribbon controlling joints:
    ribbon_control_joints = dict()
    ribbon_control_joints["Hip"] = joints.create_joint("{0}_{1}_Ribbon_Jnt".format(name, control_names[0]),
                                                       position=cmds.xform(hip_joint, query=True, translation=True,
                                                                           worldSpace=True))

    ribbon_control_joints["Chest"] = joints.create_joint("{0}_{1}_Ribbon_Jnt".format(name, control_names[2]),
                                                         position=cmds.xform(chest_joint, query=True, translation=True,
                                                                             worldSpace=True))

    ribbon_control_joints["Abdomen"] = joints.create_joint("{0}_{1}_Ribbon_Jnt".format(name, control_names[1]),
                                                           position=mid_spine_position)

    # constraint chest and hips to their respective ribbon joints (first delete existing constraints)
    cmds.delete(cmds.listRelatives(hip_joint, type="constraint"))
    cmds.delete(cmds.listRelatives(chest_joint, type="constraint"))

    cnst.parent_constraint(ribbon_control_joints["Hip"], hip_joint, maintain_offset=True)
    cnst.parent_constraint(ribbon_control_joints["Chest"], chest_joint, maintain_offset=True)

    # bind ribbon to the ribbon joints:
    skin_clusters.skin_without_bindpose(list(ribbon_control_joints.values()) + [ribbon_sfc])

    # determine the axis for the spine:
    spine_vector = om.MVector(cmds.xform(chest_joint, q=True, translation=True, worldSpace=True)) - \
                   om.MVector(cmds.xform(hip_joint, q=True, translation=True, worldSpace=True))

    ribbon_controls = dict()
    ribbon_control_groups = dict()

    # create spine control hierarchy
    for key in ribbon_control_joints:
        control_position = cmds.xform(ribbon_control_joints[key], query=True, translation=True, worldSpace=True)
        spine_ctrl = rig_controls.make_control("square_X", size=scale, name="{0}_{1}_Ik_Ctrl".format(name, key),
                                               color="yellow", position=control_position, rotation=orientation)

        # if controls are not supposed to be world orientated then get their orientations from the given spine joints
        if not world_orientated_controls:

            if key == "Hip":
                closest_spine_joint = spine_joints[0]
            elif key == "Chest":
                closest_spine_joint = spine_joints[-1]
            else:
                closest_spine_joint = spine_joints[int(len(spine_joints)/2 - 1)]

            cmds.delete(cmds.orientConstraint(closest_spine_joint, spine_ctrl))

        spine_ctrl_grp = trans_mod.make_space_for(spine_ctrl)
        cnst.parent_constraint(spine_ctrl, ribbon_control_joints[key], maintain_offset=True)
        ribbon_controls[key] = spine_ctrl
        ribbon_control_groups[key] = spine_ctrl_grp

    # parent abdomen control between two outer ones:
    if middle_control_constrained:
        cnst.parent_constraint(masters=(ribbon_controls["Hip"], ribbon_controls["Chest"]),
                               slave=ribbon_control_groups["Abdomen"], maintain_offset=True,
                               interp_type=2)

    # cleanup - parent all the ribbon joints to do not touch group:
    cmds.parent(ribbon_data["Module_Group"], do_not_touch_grp)

    # create scaling group for ribbon joints:
    ribbon_joints_scaling_grp = cmds.createNode("transform", name="{0}_Ribbon_Joints_Grp".format(name))
    cmds.parent(list(ribbon_control_joints.values()), ribbon_joints_scaling_grp)

    cmds.parent(ribbon_joints_scaling_grp, do_not_touch_grp)

    # creating the COG arrow if that option was specified:
    if main_cog_control:
        cog_ctrl = rig_controls.make_control("arrow", size=scale * 0.5, name="{0}_COG_Ctrl".format(name),
                                             color="yellow", translate_comps=(0, scale * 2.0, 0), rotation=orientation,
                                             rotate_comps=(90, 0, 90),
                                             position=cmds.xform(hip_joint, q=True, translation=True, worldSpace=True))

        cog_ctrl_grp = trans_mod.make_space_for(cog_ctrl)

        cmds.parent(list(ribbon_control_groups.values()), cog_ctrl)

        if not virtual_constraints:
            cnst.build_constraints()

        result = {"Main_Node": cog_ctrl,
                  "Controls": [cog_ctrl, ribbon_controls["Hip"], ribbon_controls["Abdomen"], ribbon_controls["Chest"]],
                  "Do_Not_Touch_Grp": do_not_touch_grp,
                  "Scaling_Groups": [ribbon_joints_scaling_grp],
                  "Controls_Grp": [cog_ctrl_grp] + list(ribbon_control_groups.values())}

        return result

    if not virtual_constraints:
        cnst.build_constraints()

    # returning data without main COG control otherwise
    result = {"Main_Node": ribbon_controls["Hip"],
              "Controls": [ribbon_controls["Hip"], ribbon_controls["Abdomen"], ribbon_controls["Chest"]],
              "Do_Not_Touch_Grp": do_not_touch_grp,
              "Scaling_Groups": [ribbon_joints_scaling_grp],
              "Controls_Grp": ribbon_control_groups.values()}

    return result
