import maya.cmds as cmds

import rigging_framework.templates.common.utilities.transform
from rigging_framework.templates.common.utilities import utils
from rigging_framework.templates.common.utilities import colors
from rigging_framework.templates.common.utilities import joints
from rigging_framework.templates.common.utilities import driven_keys
from rigging_framework.templates.common.modules import fk_chain
from rigging_framework.templates.quadruped_rig.modules import quad_ik_ribbon_spine

import imp
imp.reload(quad_ik_ribbon_spine)


def build(name, spine_joints, ik_drives_fk=True, scale=1.0, orientation=(0, 0, 0), skip_joints_missing_str=()):
    """
    Builds ikfk ribbon spine with a switch for quadruped
    
    :param str name: Name of the module
    :param list[str] spine_joints: List of spine joints
    :param bool ik_drives_fk: By default IK drives FK spine - but it could be opposite
    :param float scale: Scale of the controls
    :param float[3] orientation: Default orientation of IK controls (0, 0, 0) default
    :param skip_joints_missing_str: When duplicating chains to create IK and FK chains - deletes the joints that do not contain this string
    
    :return dict: 
        result = {"Main_Node": COG Control for the spine - Main Node,
                  "Controls": IK Spine controls and FK controls,
                  "Do_Not_Touch_Grp": Group containing hidden rig elements,
                  "Scaling_Groups": Group requiring individual scaling,
                  "Controls_Grp": Control offset groups for controls above}
    """
    
    # create duplicate chain for FK
    fk_spine_joints = utils.duplicate_unparent(spine_joints[0], depth=len(spine_joints), prefix='FK',
                                               favor=spine_joints[1])

    if skip_joints_missing_str:
        joints_to_delete = [x for x in cmds.listRelatives(fk_spine_joints[0], allDescendents=True, fullPath=True)
                            if not utils.is_any_in(skip_joints_missing_str, x.split("|")[-1])]
        cmds.delete(joints_to_delete)
        fk_spine_joints = utils.sort_by_depth(list(set(fk_spine_joints) - set([x.split("|")[-1] for x in joints_to_delete])))

    colors.set_color_from_str(fk_spine_joints[0], "blue")

    # create duplicate chain for IK
    ik_spine_joints = utils.duplicate_unparent(spine_joints[0], depth=len(spine_joints), prefix='IK',
                                               favor=spine_joints[1])

    if skip_joints_missing_str:
        joints_to_delete = [x for x in cmds.listRelatives(ik_spine_joints[0], allDescendents=True, fullPath=True)
                            if not utils.is_any_in(skip_joints_missing_str, x.split("|")[-1])]
        cmds.delete(joints_to_delete)
        ik_spine_joints = utils.sort_by_depth(list(set(ik_spine_joints) - set([x.split("|")[-1] for x in joints_to_delete])))

    colors.set_color_from_str(ik_spine_joints[0], "red")

    # now lets xform all the spine joints so that we don't get into joint orients problems:
    joints.xform_joints(fk_spine_joints)
    joints.xform_joints(ik_spine_joints)
    joints.xform_joints(spine_joints)

    ik_spine_data = quad_ik_ribbon_spine.build("{0}_Ik".format(name), ik_spine_joints, scale=scale,
                                               orientation=orientation)

    fk_chain_data = fk_chain.build(fk_spine_joints, name_or_none=None, replace_in_name=[("Jnt", "Fk_Ctrl")],
                                   scale_constraint=True)

    spine_cog_ctrl = ik_spine_data["Main_Node"]
    fk_chain_top_grp = fk_chain_data["Module_Group"]
    cmds.parent(fk_chain_top_grp, spine_cog_ctrl)

    # add ik fk switch attribute on spine COG control
    cmds.addAttr(spine_cog_ctrl, longName="ikfk_blend", minValue=0, maxValue=1, keyable=True)

    # create ik and fk visibility curves and hook them up to controls
    fk_vis_crv = driven_keys.build_quick_curve("{0}_FK_Visibility_Curve".format(name), [(0, 0), (1, 1)])
    cmds.connectAttr(spine_cog_ctrl + ".ikfk_blend", fk_vis_crv + ".input")
    cmds.connectAttr(fk_vis_crv + ".output", fk_chain_top_grp + ".visibility")

    ik_vis_crv = driven_keys.build_quick_curve("{0}_IK_Visibility_Curve".format(name), [(0, 1), (1, 0)])
    cmds.connectAttr(spine_cog_ctrl + ".ikfk_blend", ik_vis_crv + ".input")

    for ik_control_grp in ik_spine_data["Controls_Grp"]:
        if ik_control_grp != rigging_framework.templates.common.utilities.transform.get_parent(spine_cog_ctrl):
            cmds.connectAttr(ik_vis_crv + ".output", ik_control_grp + ".visibility")

    # now build the blending for the ikfk spine:
    for i, (joint, fk_joint, ik_joint) in enumerate(zip(spine_joints, fk_spine_joints, ik_spine_joints)):
        if i == 0:
            cns = cmds.parentConstraint(fk_joint, ik_joint, joint, maintainOffset=True)[0]
            cmds.connectAttr(fk_vis_crv + ".output", cns + ".{0}W0".format(fk_joint))
            cmds.connectAttr(ik_vis_crv + ".output", cns + ".{0}W1".format(ik_joint))
        else:
            cns = cmds.parentConstraint(fk_joint, ik_joint, joint)[0]
            cmds.connectAttr(fk_vis_crv + ".output", cns + ".{0}W0".format(fk_joint))
            cmds.connectAttr(ik_vis_crv + ".output", cns + ".{0}W1".format(ik_joint))

            blend_colors = cmds.createNode("blendColors")
            cmds.connectAttr(ik_joint + ".scale", blend_colors + ".color2")
            cmds.connectAttr(fk_joint + ".scale", blend_colors + ".color1")
            cmds.connectAttr(spine_cog_ctrl + ".ikfk_blend", blend_colors + ".blender")
            cmds.connectAttr(blend_colors + ".output", joint + ".scale")

    # now add ik drives fk functionality:
    if ik_drives_fk:
        cmds.addAttr(spine_cog_ctrl, longName="ik_drives_fk", attributeType="enum", enumName="Off:On", keyable=True)

        # override visibility of the ik controls so they always on in ik drives fk mode:
        for ik_control in ik_spine_data["Controls"][1:]:
            ik_control_grp = rigging_framework.templates.common.utilities.transform.get_parent(ik_control)
            vis_input = cmds.listConnections(ik_control_grp + ".visibility", source=True, destination=False, plugs=True)[0]
            condition_node = cmds.createNode("condition")
            cmds.connectAttr(spine_cog_ctrl + ".ik_drives_fk", condition_node + ".firstTerm")
            cmds.connectAttr(vis_input, condition_node + ".colorIfTrueR")
            cmds.setAttr(condition_node + ".colorIfFalseR", 1)
            utils.disconnect_attr(ik_control_grp + ".visibility")
            cmds.connectAttr(condition_node + ".outColorR", ik_control_grp + ".visibility")

        # override ikfk-blend so its always 1
        outputs_from_ikfk_blend = cmds.listConnections(spine_cog_ctrl + ".ikfk_blend", source=False,
                                                       destination=True, plugs=True)
        condition_node = cmds.createNode("condition")
        cmds.connectAttr(spine_cog_ctrl + ".ik_drives_fk", condition_node + ".firstTerm")
        cmds.connectAttr(spine_cog_ctrl + ".ikfk_blend", condition_node + ".colorIfTrueR")
        cmds.setAttr(condition_node + ".colorIfFalseR", 1)
        for output in outputs_from_ikfk_blend:
            utils.disconnect_attr(output)
            cmds.connectAttr(condition_node + ".outColorR", output)

        # plug in rotation and translation of the ik joints to the fk groups
        for i, (ik_joint, fk_ctrl) in enumerate(zip(ik_spine_joints, fk_chain_data["Controls"])):
            fk_ctrl_grp = rigging_framework.templates.common.utilities.transform.get_parent(fk_ctrl)
            pair_blend = cmds.createNode("pairBlend")

            if i == 0:
                # make offset group for first fk:
                offset_fk_ctrl_grp = cmds.createNode("transform", name=fk_ctrl_grp.replace("Grp", "Offset_Grp"))
                rigging_framework.templates.common.utilities.transform.parent_and_zero(offset_fk_ctrl_grp, spine_cog_ctrl)
                cmds.parent(fk_ctrl_grp, offset_fk_ctrl_grp)
                fk_ctrl_grp = offset_fk_ctrl_grp

                cmds.connectAttr(ik_spine_data["Controls"][1] + ".translate", pair_blend + ".inTranslate2")
                cmds.setAttr(pair_blend + ".inTranslate1", *cmds.getAttr(fk_ctrl_grp + ".translate")[0])
                cmds.connectAttr(ik_spine_data["Controls"][1] + ".rotate", pair_blend + ".inRotate2")
                cmds.setAttr(pair_blend + ".inRotate1", *cmds.getAttr(fk_ctrl_grp + ".rotate")[0])
            else:
                cmds.connectAttr(ik_joint + ".translate", pair_blend + ".inTranslate2")
                cmds.setAttr(pair_blend + ".inTranslate1", *cmds.getAttr(fk_ctrl_grp + ".translate")[0])
                cmds.connectAttr(ik_joint + ".rotate", pair_blend + ".inRotate2")
                cmds.setAttr(pair_blend + ".inRotate1", *cmds.getAttr(fk_ctrl_grp + ".rotate")[0])

            cmds.connectAttr(spine_cog_ctrl + ".ik_drives_fk", pair_blend + ".weight")

            cmds.connectAttr(pair_blend + ".outTranslate", fk_ctrl_grp + ".translate")
            cmds.connectAttr(pair_blend + ".outRotate", fk_ctrl_grp + ".rotate")

    # cleanup a little:
    do_not_touch_grp = cmds.createNode("transform", name="{0}_Do_Not_Touch_Grp".format(name))
    cmds.parent(ik_spine_data["Do_Not_Touch_Grp"], do_not_touch_grp)
    cmds.parent(ik_spine_joints[0], do_not_touch_grp)
    cmds.parent(fk_spine_joints[0], do_not_touch_grp)
    # return values

    result = {"Main_Node": spine_cog_ctrl,
              "Controls": ik_spine_data["Controls"] + fk_chain_data["Controls"],
              "Do_Not_Touch_Grp": do_not_touch_grp,
              "Scaling_Groups": ik_spine_data["Scaling_Groups"],
              "Controls_Grp": rigging_framework.templates.common.utilities.transform.get_parent(spine_cog_ctrl)}

    return result
