import maya.api.OpenMaya as om
import maya.cmds as cmds
from rigging_framework.templates.common.utilities import follicles
from rigging_framework.templates.common.utilities import utils
from rigging_framework.templates.common.utilities import joints
from rigging_framework.templates.common.utilities import rig_controls
from rigging_framework.templates.common.utilities import skin_clusters
from rigging_framework.templates.common.modules import ribbon
import rigging_framework.templates.common.utilities.transform as trans_mod


def build(name, neck_joints, scale=1.0, neck_has_orientation=False, head_has_orientation=True, orientation=(0, 0, 0)):

    """
    Builds quad ribbon neck and head system
    
    :param str name: Name of the module created
    :param list[str] neck_joints: Neck joints starting from the hip, ending with the chest
    :param float scale: Scale of controls
    :param bool neck_has_orientation: Whether the neck controls will be orientated like the rig orientation or follow the joints
    :param bool head_has_orientation: Whether the head control will be orientated like the rig orientation or follow the joint
    :param float[3] orientation: What is the default orientation for this rig
    
    :return dict:
            result = {"Main_Node": ribbon_controls["Neck_Base"],
                      "Controls": [ribbon_controls["Neck_Base"], ribbon_controls["Neck_Mid"], ribbon_controls["Head"]],
                      "Do_Not_Touch_Grp": do_not_touch_grp,
                      "Scaling_Groups": [ribbon_joints_scaling_grp],
                      "Controls_Grp": ribbon_control_groups["Neck_Base"]}
    """
    
    # create main folders:
    do_not_touch_grp = cmds.createNode("transform", name="{0}_Do_Not_Touch_Grp".format(name))

    ribbon_data = ribbon.build(neck_joints, "{0}_Ribbon".format(name), attach_objects=True, sticky=False,
                               make_controls=False, scale=1)

    ribbon_sfc = ribbon_data["Ribbon_Surface"]

    # get hip and chest joint
    neck_base_joint = neck_joints[0]
    head_joint = neck_joints[-1]

    # get mid spine position
    neck_mid_follicle = follicles.make_follicle(ribbon_sfc, [0.5, 0.5], "neck_mid_follicle")
    mid_neck_position = cmds.xform(neck_mid_follicle, query=True, translation=True, worldSpace=True)
    cmds.delete(neck_mid_follicle)

    # create three big ribbon controlling joints:
    ribbon_control_joints = dict()
    ribbon_control_joints["Neck_Base"] = joints.create_joint("{0}_Neck_Base_Ribbon_Jnt".format(name),
                                                             position=cmds.xform(neck_base_joint, query=True,
                                                                                 translation=True, worldSpace=True))

    ribbon_control_joints["Neck_Mid"] = joints.create_joint("{0}_Neck_Mid_Ribbon_Jnt".format(name),
                                                            position=mid_neck_position)

    ribbon_control_joints["Head"] = joints.create_joint("{0}_Head_Ribbon_Jnt".format(name),
                                                        position=cmds.xform(head_joint, query=True,
                                                                            translation=True, worldSpace=True))

    # constraint chest and hips to their respective ribbon joints (first delete existing constraints)
    cmds.delete(cmds.listRelatives(neck_base_joint, type="constraint"))
    cmds.delete(cmds.listRelatives(head_joint, type="constraint"))

    cmds.parentConstraint(ribbon_control_joints["Neck_Base"], neck_base_joint, maintainOffset=True)
    cmds.parentConstraint(ribbon_control_joints["Head"], head_joint, maintainOffset=True)

    # bind ribbon to the ribbon joints:
    skin_clusters.skin_without_bindpose(list(ribbon_control_joints.values()) + [ribbon_sfc])

    # determine the axis for the neck:
    neck_vector = om.MVector(cmds.xform(head_joint, q=True, translation=True, worldSpace=True)) - \
                  om.MVector(cmds.xform(neck_base_joint, q=True, translation=True, worldSpace=True))

    ribbon_controls = dict()
    ribbon_control_groups = dict()

    # create neck control hierarchy
    for i, key in enumerate(ribbon_control_joints.keys()):
        control_position = cmds.xform(ribbon_control_joints[key], query=True, translation=True, worldSpace=True)
        if neck_has_orientation:
            neck_ctrl = rig_controls.make_control("square_X", size=scale, name="{0}_{1}_Ik_Ctrl".format(name, key),
                                                  color="yellow", position=control_position, rotation=orientation)
        else:
            # using neck joints to get orientation of the controls:
            neck_joint_index = 0
            if i == 1:
                neck_joint_index = int(len(neck_joints) / 2)
            elif i == 2:
                neck_joint_index = len(neck_joints) - 1

            control_rotation = cmds.xform(neck_joints[neck_joint_index], query=True, rotation=True, worldSpace=True)

            # build head control - box (if head world axis off then set control rotation back to 0,0,0
            if i == 2:
                if head_has_orientation:
                    control_rotation = orientation
                neck_ctrl = rig_controls.make_control("box", size=scale, name="{0}_{1}_Ik_Ctrl".format(name, key),
                                                      color="yellow", position=control_position,
                                                      rotation=control_rotation)
            # build neck controls - squares
            else:
                neck_ctrl = rig_controls.make_control("square_X", size=scale, name="{0}_{1}_Ik_Ctrl".format(name, key),
                                                      color="yellow", position=control_position,
                                                      rotation=control_rotation)

        neck_ctrl_grp = trans_mod.make_space_for(neck_ctrl)
        cmds.parentConstraint(neck_ctrl, ribbon_control_joints[key], maintainOffset=True)
        ribbon_controls[key] = neck_ctrl
        ribbon_control_groups[key] = neck_ctrl_grp

    # parent neck mid control and head control under the neck base:
    cmds.parent(ribbon_control_groups["Head"], ribbon_controls["Neck_Base"])
    cmds.parent(ribbon_control_groups["Neck_Mid"], ribbon_controls["Neck_Base"])

    # parent constraint neck_mid control between two outer ones:
    neck_constraint = cmds.parentConstraint(ribbon_controls["Neck_Base"], ribbon_controls["Head"],
                                            ribbon_control_groups["Neck_Mid"], maintainOffset=True)[0]

    cmds.setAttr("{0}.interpType".format(neck_constraint), 2)

    # add following ratio on the middle control:
    cmds.addAttr(ribbon_controls["Neck_Mid"], longName="follow_head", attributeType="float", defaultValue=0.5,
                 maxValue=1.0, minValue=0.0, keyable=True)

    reverse_neck_follow = cmds.createNode("reverse")
    cmds.connectAttr(ribbon_controls["Neck_Mid"] + ".follow_head", reverse_neck_follow + ".inputX")
    cmds.connectAttr(reverse_neck_follow + ".outputX", neck_constraint + ".{0}W0".format(ribbon_controls["Neck_Base"]))
    cmds.connectAttr(ribbon_controls["Neck_Mid"] + ".follow_head", neck_constraint + ".{0}W1".format(ribbon_controls["Head"]))

    # add neck mode:
    cmds.addAttr(ribbon_controls["Neck_Base"], longName="neck_mode", attributeType="enum", enumName="ik:fk",
                 keyable=True)

    cond_neck_mode_t = cmds.createNode("condition")
    cond_neck_mode_r = cmds.createNode("condition")
    cmds.connectAttr(ribbon_controls["Neck_Base"] + ".neck_mode", cond_neck_mode_t + ".firstTerm")
    cmds.connectAttr(ribbon_controls["Neck_Base"] + ".neck_mode", cond_neck_mode_r + ".firstTerm")

    cmds.setAttr(cond_neck_mode_t + ".colorIfFalse", *cmds.getAttr(ribbon_control_groups["Neck_Mid"] + ".translate")[0])
    cmds.setAttr(cond_neck_mode_r + ".colorIfFalse", *cmds.getAttr(ribbon_control_groups["Neck_Mid"] + ".rotate")[0])

    cmds.connectAttr(neck_constraint + ".constraintTranslate", cond_neck_mode_t + ".colorIfTrue")
    cmds.connectAttr(neck_constraint + ".constraintRotate", cond_neck_mode_r + ".colorIfTrue")

    for axis in "XYZ":
        utils.disconnect_attr(ribbon_control_groups["Neck_Mid"] + ".translate{0}".format(axis))
        utils.disconnect_attr(ribbon_control_groups["Neck_Mid"] + ".rotate{0}".format(axis))

    cmds.connectAttr(cond_neck_mode_t + ".outColor", ribbon_control_groups["Neck_Mid"] + ".translate")
    cmds.connectAttr(cond_neck_mode_r + ".outColor", ribbon_control_groups["Neck_Mid"] + ".rotate")

    # build parent Constraint for FK mode (only enabled in that mode)
    cmds.setAttr(ribbon_controls["Neck_Base"] + ".neck_mode", 1)

    fk_constraint = cmds.parentConstraint(ribbon_controls["Neck_Mid"], ribbon_control_groups["Head"],
                                          maintainOffset=True)[0]

    fk_condition_t = cmds.createNode("condition")
    fk_condition_r = cmds.createNode("condition")

    cmds.connectAttr(ribbon_controls["Neck_Base"] + ".neck_mode", fk_condition_t + ".firstTerm")
    cmds.connectAttr(ribbon_controls["Neck_Base"] + ".neck_mode", fk_condition_r + ".firstTerm")

    cmds.connectAttr(fk_constraint + ".constraintTranslate", fk_condition_t + ".colorIfFalse")
    cmds.connectAttr(fk_constraint + ".constraintRotate", fk_condition_r + ".colorIfFalse")

    cmds.setAttr(fk_condition_t + ".colorIfTrue", *cmds.getAttr(ribbon_control_groups["Head"] + ".translate")[0])
    cmds.setAttr(fk_condition_r + ".colorIfTrue", *cmds.getAttr(ribbon_control_groups["Head"] + ".rotate")[0])

    for axis in "XYZ":
        utils.disconnect_attr(ribbon_control_groups["Head"] + ".translate{0}".format(axis))
        utils.disconnect_attr(ribbon_control_groups["Head"] + ".rotate{0}".format(axis))

    cmds.connectAttr(fk_condition_t + ".outColor", ribbon_control_groups["Head"] + ".translate")
    cmds.connectAttr(fk_condition_r + ".outColor", ribbon_control_groups["Head"] + ".rotate")

    # cleanup - parent all the ribbon joints to do not touch group:
    cmds.parent(ribbon_data["Module_Group"], do_not_touch_grp)

    # create scaling group for ribbon joints:
    ribbon_joints_scaling_grp = cmds.createNode("transform", name="{0}_Ribbon_Joints_Grp".format(name))
    cmds.parent(list(ribbon_control_joints.values()), ribbon_joints_scaling_grp)
    cmds.parent(ribbon_joints_scaling_grp, do_not_touch_grp)

    result = {"Main_Node": ribbon_controls["Neck_Base"],
              "Controls": [ribbon_controls["Neck_Base"], ribbon_controls["Neck_Mid"], ribbon_controls["Head"]],
              "Do_Not_Touch_Grp": do_not_touch_grp,
              "Scaling_Groups": [ribbon_joints_scaling_grp],
              "Controls_Grp": ribbon_control_groups["Neck_Base"]}

    return result
